%%%-------------------------------------------------------------------
%%% File    : berkeley_db_global_transaction_manager.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%    A simple, **non-replicated** Global Transaction Manger that 
%%%    meets the requirements of berkeley_db_port_coordinator
%%%    when distributed transactions are enabled.
%%%
%%%    In our model of distributed transactions the GTM 
%%%    is a passive store of authoratative state.
%%%
%%% APIs    
%%%
%%% get_new_distributed_txn_guid()
%%%    Returns a 128-byte binary, suitable for use with BDB's txn_prepare call.
%%%    This uses a berkeley_db_sequence_server to generate 64-bit 
%%%    unique ids. Currently the rest of the 128-byte guid is all 0's.
%%%    If we ever support multiple GTMs then each GTM will need a unique 
%%%    ID of its own, and that ID will also need to be placed in the guid.
%%%
%%% commit(DistTxnGUID)
%%%    If and only if the transaction has not been recorded as aborted 
%%%    this writes 'committed' to stable storage, and if that succeeds,
%%%    it returns ok, and the caller should tell the participants to do 
%%%    local commit.
%%%
%%%    If this returns {error, {GUID, already_aborted}} then the
%%%    caller should tell all participants to do local abort.
%%%
%%%    If this returns any other error or times-out then the
%%%    transaction MAY still have been committed.  The caller must NOT
%%%    not tell partiticpants to abort (or commit), and should call
%%%    resolve().
%%%
%%% abort(DistTxnGUID)
%%%    If and only if the transaction has not been recorded as committed 
%%%    this writes 'aborted' to stable storage, and if that succeeds,
%%%    it returns ok, and the caller should tell the participants to do 
%%%    local aborted.
%%%
%%%    If this returns {error, {GUID, already_committed}} then the
%%%    caller should tell all participants to do local commit, and
%%%    should probably ask a human for help, as this should never
%%%    happen.
%%%
%%%    If this returns any other error or times-out then the
%%%    transaction MAY still have been aborted.  The caller must NOT
%%%    not tell partiticpants to abort (or commit), and should call
%%%    resolve().
%%%
%%% resolve(DistTxnGUID or ListDistTxnGUIDs)
%%%    If successful, returns {ok, {GUID, committed|aborted}}.
%%%
%%%    IMPORTANT: If no record exists in the database for this
%%%    transaction , we use the 'presumed abort' optimization.  So
%%%    calling 'resolve' must implicitly abort such a transaction.
%%%    This is necessary because of a race condition exists: It is
%%%    possible that a participant failed and recovered, and is asking
%%%    for a decision without the transaction's coordinator being
%%%    aware of the fact.  So the transaction coordinator may later
%%%    calling commit() on the same transaction.  To guard against
%%%    this, the the first call to resolve() on a txn for which we
%%%    have no record on disk will write 'aborted' to disk before it
%%%    returns to the caller.  This causes any subsequent call to
%%%    commit() to fail with {error, already_aborted}.
%%%
%%%    If resolve() returns an error or times-out then the result is
%%%    implicitly UNKNOWN and the caller must NOT locally abort (or
%%%    commit).  They should retry the call to resolve until it
%%%    succeeds (and ask a human for help).
%%%
%%%    This is why it is ESSENTIAL that the GTM is highly-available.
%%%    Another alternative is for all participant port_coordinators to
%%%    locally record the decision for the most recent N distributed
%%%    transactions, and the {Node,Name} of each participant.  Then if
%%%    the GTM is not available, any participant may ask any other for
%%%    the decision (and we are only blocked if no-one knows what that
%%%    decision is).  However, is much easier to just use BDB-HA for
%%%    the GTM, and an appropriately high number of replicas.
%%%
%%% TODO: 
%%% - It is ESSENTIAL that a GTM is highly available. 
%%%   Therefore this needs to use berkeley_db_replication_group.
%%% 
%%% - Adding (extremely conservative) purging of old state records based 
%%%   on timestamps.
%%%
%%% - Add retry-limits for deadlock handling (not done now as this will 
%%%   change completely when replication is used -- deadlock retries 
%%%   will be automatic then)
%%%-------------------------------------------------------------------
-module(berkeley_db_global_transaction_manager).

-behaviour(gen_server).

-include("berkeley_db_drv.hrl").  %% required for DB_XIDDATASIZE
-define(BDB,     berkeley_db_drv).
-define(BDB_PC,  berkeley_db_port_coordinator).
-define(BDB_SEQ, berkeley_db_sequence_server).

%% API
-export([start_link/1, 
         start_link/2,
         stop/0,
         get_new_distributed_txn_guid/0, get_new_distributed_txn_guid/1,
         commit/1, commit/2,
         abort/1, abort/2,
         resolve/1, resolve/2
        ]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% This is read-only state (I would normally keep it in the process dictionary, but as we have so little mutable state,
%% I keep these for speed).
-record(state, {bdbPortHandle, db}).

%% These are the terms (plus an erlang:now timestamp) that are stored on disk.
%% Obviously we only need a single byte.
-define(COMMITTED_IN_BDB_RECORD, c).
-define(ABORTED_IN_BDB_RECORD,   a).

-define(PHYSICAL_DATABASE_FILENAME, "global_transaction_state.db").


%%-define(INFO_LOGGER,    error_logger:info_msg).
%%-define(WARNING_LOGGER, error_logger:warning_msg).
%%-define(ERROR_LOGGER,   error_logger:error_msg).
-define(INFO_LOGGER,    logger).
-define(WARNING_LOGGER, logger).
-define(ERROR_LOGGER,   logger).

logger(String) ->
    logger(String, []).
logger(String, Args) ->
    case process_info(self(), registered_name) of
        {registered_name, ProcessName} ->
            io:format("~p: " ++ String, [ProcessName | Args]);
        [] ->
            %% This requires that we only create support processes with proc_lib:spawn_link
            io:format("~p: support process ~p: " ++ String, [hd(get('$ancestors')) | [self() | Args]])
    end.



%%====================================================================
%% API
%%====================================================================

-define(THROW_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_gtm_error, Error})
           end).

%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link(EnvDir) ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, {EnvDir, []}, []).

start_link(EnvDir, PortCoordinatorConfigProplist) ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, {EnvDir, PortCoordinatorConfigProplist}, []).

stop() ->
    gen_server:cast(?MODULE, stop).

get_new_distributed_txn_guid() ->
    get_new_distributed_txn_guid(node()).

get_new_distributed_txn_guid(Node) ->
    ?THROW_ERRORS( gen_server:call({?MODULE, Node}, get_new_distributed_txn_guid) ).
    
commit(DistributedTxnGUID) ->
    commit(node(), DistributedTxnGUID).
    
commit(Node, DistributedTxnGUID) ->
    ?THROW_ERRORS( gen_server:call({?MODULE, Node}, {commit, DistributedTxnGUID}) ).

abort(DistributedTxnGUID) ->
    abort(node(), DistributedTxnGUID).

abort(Node, DistributedTxnGUID) ->
    ?THROW_ERRORS( gen_server:call({?MODULE, Node}, {abort, DistributedTxnGUID}) ).

resolve(DistributedTxnGUID) ->
    resolve(node(), DistributedTxnGUID).

resolve(Node, DistributedTxnGUID) ->
    ?THROW_ERRORS( gen_server:call({?MODULE, Node}, {resolve, DistributedTxnGUID}) ).


%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%
%% Args is a 'proplist' of tagged tuples
%%
%% Required
%%    {env_dir, "directory path"}                          
%%
%% Optional
%%    See defaults in comments at start.
%%
%%--------------------------------------------------------------------
init({EnvDir, PortCoordinatorConfigProplist}) ->

    ?INFO_LOGGER("starting (~p)~n", [self()]),
    process_flag(trap_exit, true),

    {ok, PortCoordPid} = ?BDB_PC:start_link(global_transaction_manager_port_coordinator,
                                            EnvDir, 
                                            PortCoordinatorConfigProplist),    

    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(PortCoordPid),

    %% We will use multiple databases in the same physical file, to keep OS file descriptors to a minimum.
    DB = ?BDB_PC:get_db_handle(BdbPortHandle, 
                               bdb_DB_BTREE, 
                               {?PHYSICAL_DATABASE_FILENAME, "dist_txn_states"}),

    {ok, SeqServerPid} = ?BDB_SEQ:start_link(global_transaction_manager_sequence_server,
                                             global_transaction_manager_port_coordinator, 
                                             {?PHYSICAL_DATABASE_FILENAME, "sequences"},
                                             <<"next_dist_txn_guid">>,
                                             50), % PersistEveryNConsumedValues

    put(port_coordinator_pid, PortCoordPid),
    put(sequence_server_pid, SeqServerPid),

    {ok, #state{bdbPortHandle=BdbPortHandle, db=DB}}.


%%--------------------------------------------------------------------
%% Function: handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call(get_new_distributed_txn_guid, _From, State) ->

    TxnCounter = ?BDB_SEQ:consume_one_value(global_transaction_manager_sequence_server),

    DistributedTxnGUID = <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>>,
    ?INFO_LOGGER("get_new_distributed_txn_guid: returning ~w~n", [TxnCounter]),
    {reply, {ok, DistributedTxnGUID}, State};


handle_call({commit, <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID}, _From, State) ->
    {reply, 
     do_commit_or_abort(?COMMITTED_IN_BDB_RECORD, TxnCounter, DistributedTxnGUID, State),
     State};


handle_call({abort, <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID}, _From, State) ->
    %% TODO: we don't _need_ to write an abort record, even on explicit abort 
    %% (we could just use the presumed-abort feature).
    %% Currently we only write it because it is less code and we rarely explicitly abort, 
    %% (we have to read the record anyway, to see if it has already been explicitly 
    %% committed).
    {reply, 
     do_commit_or_abort(?ABORTED_IN_BDB_RECORD, TxnCounter, DistributedTxnGUID, State),
     State};


handle_call({resolve, <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID}, _From, State) when is_binary(DistributedTxnGUID) ->    
    Reply = resolve_one_txn(TxnCounter, DistributedTxnGUID, State),
    %%?INFO_LOGGER("resolve: for ~w returning ~p~n", [TxnCounter, Reply]),
    {reply, Reply, State};


handle_call({resolve, ListDistributedTxnGUIDs}, _From, State) when is_list(ListDistributedTxnGUIDs) ->    
    Reply = {ok, lists:map(fun(<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID) ->
                                   resolve_one_txn(TxnCounter, DistributedTxnGUID, State)
                           end,
                           ListDistributedTxnGUIDs)},
    %%?INFO_LOGGER("resolve (multiple): returning ~p~n", [Reply]),
    {reply, Reply, State}.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State}.

%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%
%% If port_coordinator or sequence-server die we will get EXIT messages.
%% Currently we choose not to handle them (we will die too).
%%--------------------------------------------------------------------
handle_info(Info, State) ->
    ?ERROR_LOGGER("Unexpected msg ~p~n", [Info]),
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(Reason, State) ->

    ?INFO_LOGGER("terminate: (reason: ~p)~n", [Reason]),

    %% Shut down support processes cleanly, and wait for them to exit (or 
    %% they will receive exit signals when this process exits after terminate returns).
    %% Important: the support processes might *already* have died
    %% and we might already have eaten the EXIT message.
    %% Fortunately erlang:monitor does what we want -- we will get a DOWN message
    %% even if the process is not running when we try to monitor it.

    SeqServerPid = get(sequence_server_pid),
    SeqServerMonitorRef = erlang:monitor(process, SeqServerPid),
    ok = ?BDB_SEQ:stop(SeqServerPid),

    %% Wait for it to exit as it is a client of our port_coordinator, and will be killed when 
    %% the port_coordinator shuts down.
    ?INFO_LOGGER("terminate: waiting for global_transaction_manager_sequence_server ~p to exit~n", [SeqServerPid]),
    receive {'DOWN', SeqServerMonitorRef, process, _, _} -> ok end,
    ?INFO_LOGGER("terminate: got exit for global_transaction_manager_sequence_server~n"),

    PortCoordPid = get(port_coordinator_pid),
    PortCoordMonitorRef = erlang:monitor(process, PortCoordPid),

    %% Note: even if the port_coordinator is still running, this call to release_bdb_port_handle will unlink us 
    %% so we won't get an EXIT message (another reason why we need to monitor it to know when it has finished terminating).
    ok = ?BDB_PC:release_bdb_port_handle(State#state.bdbPortHandle),

    ok = ?BDB_PC:stop(PortCoordPid),

    ?INFO_LOGGER("terminate: waiting for global_transaction_manager_port_coordinator ~p to exit~n", [PortCoordPid]),
    receive {'DOWN', PortCoordMonitorRef, process, _, _} -> ok end,

    ?INFO_LOGGER("terminate: success~n"),
    ok.


%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%% Helper function
%%
%% NewTxnStateInBdbRecord must be one of 
%%    ?COMMITTED_IN_BDB_RECORD
%%    ?ABORTED_IN_BDB_RECORD
%%
%% This function will ONLY write NewTxnState if there is no existing record for 
%% that transaction.  It will never overwrite any existing state
%% (will return {error, {already_committed|already_aborted, DistTxnGUID}}
%%
%% The non-overwrite guarantee so allows 'resolve' to do a separate read first, 
%% (outside of this function), without having to worry about an explicit commit happening 
%% between the read and calling this function to write an abort record 
%% (it does the latter if no record is found).
%%
do_commit_or_abort(NewTxnStateForBdbRecord, TxnCounter, DistributedTxnGUID, State) ->

    ?INFO_LOGGER("do_commit_or_abort: called with '~w' for txn ~w~n", [NewTxnStateForBdbRecord, TxnCounter]),

    Txn    = ?BDB:txn_begin(State#state.bdbPortHandle, void, []),
    Cursor = ?BDB:db_cursor(State#state.bdbPortHandle, State#state.db, Txn, []),

    Key = <<TxnCounter:64>>,

    %% Important: we take a read-modify-write lock to prevent race conditions while we 
    %% examine any existing record for this transaction.
    Reply = 
        try ?BDB:c_get(State#state.bdbPortHandle, Cursor, Key, <<>>, [bdb_DB_SET, bdb_DB_RMW]) of 

            {Key, ValueBinary} -> 
                case binary_to_term(ValueBinary) of

                    {?ABORTED_IN_BDB_RECORD, _} ->
                        case NewTxnStateForBdbRecord of
                            ?COMMITTED_IN_BDB_RECORD ->
                                %% already aborted (and we want to commit), so we return an error
                                {error, {DistributedTxnGUID, already_aborted}};
                            ?ABORTED_IN_BDB_RECORD ->
                                %% already aborted (and we want to abort), so we pretend success
                                {ok, {DistributedTxnGUID, aborted}} 
                        end;

                    {?COMMITTED_IN_BDB_RECORD, _} -> 
                        case NewTxnStateForBdbRecord of
                            ?COMMITTED_IN_BDB_RECORD ->
                                %% already committed (and we want to commit), so we pretend success
                                {ok, {DistributedTxnGUID, committed}};
                            ?ABORTED_IN_BDB_RECORD ->
                                %% already committed (and we want to abort), so we return an error
                                {error, {DistributedTxnGUID, already_committed}}
                        end
                end
        catch
            {bdb_error, bdb_DB_NOTFOUND} -> 
                ?INFO_LOGGER("do_commit_or_abort: writing '~w' for txn ~w~n", [NewTxnStateForBdbRecord, TxnCounter]),
                ok = ?BDB:c_put(State#state.bdbPortHandle, 
                                Cursor, 
                                Key, 
                                term_to_binary({NewTxnStateForBdbRecord, erlang:now()}), 
                                bdb_DB_KEYFIRST),

                case NewTxnStateForBdbRecord of
                    ?COMMITTED_IN_BDB_RECORD ->
                        {ok, {DistributedTxnGUID, committed}};
                    ?ABORTED_IN_BDB_RECORD ->
                        {ok, {DistributedTxnGUID, aborted}}
                end;

            {bdb_error, bdb_DB_LOCK_DEADLOCK} ->
                %% TODO: Temporary: infinite retries on deadlock (will change when replication is implemented)
                ?INFO_LOGGER("do_commit_or_abort: deadlock for txn ~w: Retrying~n", [TxnCounter]),
                ok = ?BDB:c_close(State#state.bdbPortHandle, Cursor),
                ok = ?BDB:txn_abort(State#state.bdbPortHandle, Txn),
                do_commit_or_abort(NewTxnStateForBdbRecord, TxnCounter, DistributedTxnGUID, State)

        end,    

    ok = ?BDB:c_close(State#state.bdbPortHandle, Cursor),
    ok = ?BDB:txn_commit(State#state.bdbPortHandle, Txn, []),
    Reply.


resolve_one_txn(TxnCounter, DistributedTxnGUID, State) ->
    %% If 'resolve' finds no record for a transaction it 
    %% MUST write an explicit abort record before returning.
    %%
    %% If that write is necessary we DON'T need to do it in the same
    %% transaction as the original read.  If a commit happens between 
    %% the two then the write will fail, as do_commit_or_abort will 
    %% ONLY write something if there is no existing record (and if a 
    %% record does exist which matches the 'new' state -- in this case 
    %% an 'aborted' record, then do_commit_or_abort just returns as 
    %% though it succeeded).
    Key = <<TxnCounter:64>>,
    try ?BDB:db_get(State#state.bdbPortHandle, State#state.db, void, Key, <<>>, []) of 

        {Key, ValueBinary} -> 
            case binary_to_term(ValueBinary) of
                {?ABORTED_IN_BDB_RECORD, _} ->
                    {ok, {DistributedTxnGUID, aborted}};

                {?COMMITTED_IN_BDB_RECORD, _} -> 
                    {ok, {DistributedTxnGUID, committed}}
            end

    catch
        {bdb_error, bdb_DB_NOTFOUND} -> 
            ?INFO_LOGGER("resolve_one_txn: no record found for txn ~w so writing explicit 'aborted' record~n", [TxnCounter]),
            do_commit_or_abort(?ABORTED_IN_BDB_RECORD, TxnCounter, DistributedTxnGUID, State);

        {bdb_error, bdb_DB_LOCK_DEADLOCK} ->
            %% TODO: Temporary: infinite retries on deadlock (will change when replication is implemented)
            ?INFO_LOGGER("resolve_one_txn: deadlock while reading txn ~w: Retrying~n", [TxnCounter]),
            resolve_one_txn(TxnCounter, DistributedTxnGUID, State)

    end.
