
%%% File    : berkeley_db_helpers.erl
%%% Author  : Chris Newcombe <>
%%% Description : 
%%% Created :  7 Sep 2006 by Chris Newcombe <>

-module(berkeley_db_helpers).


-export([do_txn/2,             % throws on error
         do_txn/3,
         do_txn_no_throw/3     % mainly for use by servers
        ]).


% do_txn calls group_valmap_cleanup, which needs #bdb_port_handle.group_valmap_id
-include("berkeley_db_drv.hrl").

-define(BDB,    berkeley_db_drv).
-define(BDB_PC, berkeley_db_port_coordinator).



%%-define(INFO_LOGGER,    error_logger:info_msg).
%%-define(WARNING_LOGGER, error_logger:warning_msg).
%%-define(ERROR_LOGGER,   error_logger:error_msg).
-define(INFO_LOGGER,    logger).
-define(WARNING_LOGGER, logger).
-define(ERROR_LOGGER,   logger).



%%% Function: do_txn()
%%%
%%% Description: 
%%%
%%%   This is a convenience wrapper.
%%%   It Executes a TxnFun, and optionally retry some number of times if a BDB deadlock occurs.
%%%   The TxnFun is executed by the caller process -- i.e. no new process is spawned.
%%%
%%%   TxnFun must be  
%%%
%%%      {Module, Function, ClientDataTerm}
%%%
%%%   where the following exported function must exist
%%%
%%%      Module:Function(BdbPortHandle, ClientDataTerm) -> ResultTerm
%%%
%%%   Note that TxnFun is (unfortunately) NOT a normal fun (closure) or even a new-style 'fun M:F/A'.
%%%   This is done to keep a uniform interface with berkeley_db_replication_group:do_txn_on_master 
%%%   (which in fact is implemented in terms of this function).
%%%
%%%   [Sidebar: do_txn_on_master cannot use normal funs as TxnFun might be shipped to another host
%%%   (to execute on the master), and the code for TxnFun on that other host might be older or newer than 
%%%   the code that creates TxnFun on the client.  The rules for fun() code upgrades are complicated,
%%%   and relatively undocumented.  It appears that the MD5 signatures of the modules must match, which 
%%%   would prevent staggered upgrades. Using the Module:Function form to specify a TxnFun means that the 
%%%   client code can be a different version than the version on the server.]
%%%
%%%   IMPORTANT: TxnFun must **NOT** call ?BDB_PC:release_bdb_port_handle() on BdbPortHandle
%%%   This should be obvious as the TxnFun is passed the handle, and does not obtain it itself.
%%%
%%% Returns    
%%%
%%%   ResultOfTxnFun
%%%
%%% or throws one of...
%%%
%%%   {txn_fun_error, ThrownException}                   (e.g. if TxnFun called throw)
%%%   {txn_fun_error, {exit, Reason}}                    (e.g. if TxnFun called exit(Reason))
%%%   {txn_fun_error, {error, Reason}}                   (e.g. if TxnFun hit a badmatch/badarg/function_clause/case_clause etc)
%%%   {txn_fun_error, {bdb_error, bdb_DB_LOCK_DEADLOCK}}  (if TxnFun hit BDB deadlocks and exhausted NumDeadlockRetries)
%%%   {txn_fun_error, {return_with_open_resources, 
%%%                    NumResources, 
%%%                    TxnFunResult}}                    (if TxnFun returned with an open cursor/txn)
%%%
%%% or an internal error (these are examples, *not* an exhaustive list)...
%%%
%%%   {bdb_helper_error, Reason}
%%%   {bdb_helper_error, {acquire_bdb_port_handle, Reason}}
%%%   {bdb_helper_error, {group_valmap_cleanup, Reason}}
%%%   {bdb_helper_error, {release_bdb_port_handle, Reason, TxnFunResult}}
%%%
%%% Example of a simple TxnFun
%%%
%%%         my_txn(BdbPortHandle, {MyDbName, MyKey, MyData}) -> 
%%%
%%%                 %% we almost always need at least one DB handle (which we must NOT close)
%%%                 {ok, DB} = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, MyDbName),
%%%          
%%%                 %% now do some work and auto-commit it...
%%%                 {ok, 0}  = ?BDB:db_put(BdbPort, DB, void, MyKey, MyData, []),
%%%          
%%%                 %% and/or do a multi-part transaction and explicitly commit it...
%%%                 {ok, Txn} = ?BDB:txn_begin(BdbPort, void, []),
%%%                 {ok, 0}   = ?BDB:db_put(BdbPort, DB, Txn, MyKey, MyData, []),
%%%                 {ok, 0}   = ?BDB:db_put(BdbPort, DB, Txn, <<"some other key">>, MyData, []),
%%%                 {ok, 0}   = ?BDB:txn_commit(BdbPort, Txn, []),
%%%         end
%%% 
%%% 


%% This is a convenience function for applications that use a single default port-coordinator.
do_txn(TxnFun, NumDeadlockRetries) ->
    do_txn(berkeley_db_port_coordinator, TxnFun, NumDeadlockRetries).


do_txn(PortCoord, TxnFun, NumDeadlockRetries) ->
    case do_txn_no_throw(PortCoord, TxnFun, NumDeadlockRetries) of
        {ok, Result}    -> Result;
        {error, Reason} -> throw(Reason)
    end.


%% PortCoord must be an atom (name of a locally-registered process) or a *local* pid 
%% (as we call acquire_bdb_port_handle and the resulting handle can only be used locally)
%%
%% This returns one of
%%
%%   {ok,    Result}
%%   {error, {txn_fun_error, Reason}}
%%   {error, {bdb_helper_error, Reason}}
%%
%% We want a 'no throw' variant as it may be called by a server that wishes to return 
%% a term to be rethrown on the client side.
%%
%% This *does* catch and convert **ALL** TxnFun throws, exits, and 'erlang errors' (e.g. badmatch)
%% and converts them to appropriate error tuples.
%%
%% This also catches and converts throws from 'internal' code to error tuples (tagged accordingly).
%%
%% HOWEVER, this does NOT catch/convert exits and 'erlang errors' from internal code (as they may be bugs).
%% i.e. In those cases we _do_ want to kill the calling server and have its supervisor restart it.

do_txn_no_throw(PortCoord, {Module, Function, ClientData}, NumDeadlockRetries) when is_atom(PortCoord) ->
    %% This variant is a minor optimization, to avoid repeated local lookups of the process name.
    case whereis(PortCoord) of 
        undefined -> {error, {bdb_helper_error, {noproc, PortCoord}}};
        Pid       -> do_txn_no_throw(Pid, {Module, Function, ClientData}, NumDeadlockRetries)
    end;

%% TODO: this should catch & convert gen_server:call exits if PortCoord nolonger exists (but see the comments after the call to release_bdb_port_handle)
do_txn_no_throw(PortCoord, {Module, Function, ClientData}, NumDeadlockRetries) ->

    try ?BDB_PC:acquire_bdb_port_handle(PortCoord) of

        BdbPortHandle ->

            %% We need a sub-function in order to retry if there is a BDB deadlock
            %% This must return either {ok, TxnFunResult} or {error, SomeError}.
            %% It should not throw.
            RetryWrapperFun = 

                fun(_NumRetriesLeft, _) when _NumRetriesLeft < 0 ->
                        {error, {txn_fun_error, {bdb_error, bdb_DB_LOCK_DEADLOCK}}};

                   (NumRetriesLeft, ThisFun) ->
                        try 
                            {ok, Module:Function(BdbPortHandle, ClientData)}

                        catch 
                            %% If we got a BDB deadlock then we may wish to retry.
                            %% Otherwise convert the error to something that will be
                            %% rethrown by the client.

                            {bdb_error, bdb_DB_LOCK_DEADLOCK} ->

                                %% Before we retry we need to close any cursors, abort any txns

                                try ?BDB:group_valmap_cleanup(BdbPortHandle, 
                                                              BdbPortHandle#bdb_port_handle.group_valmap_id) of

                                    NumResourcesCleaned ->
                                        ?INFO_LOGGER("TxnFor got deadlock. Cleaned ~w resources and retrying ~w more times~n", [NumResourcesCleaned, NumRetriesLeft]),
                                        ThisFun(NumRetriesLeft - 1, ThisFun)

                                catch
                                    GroupCleanupReason -> 
                                        {error, {bdb_helper_error, {group_valmap_cleanup, GroupCleanupReason}}}

                                    %% We don't catch errors and exits here as they would mean a bug in system code not caller code
                                end;

                            TxnFunReason       -> {error, {txn_fun_error, TxnFunReason}};

                            %% caller could provide bad/undefined Module or Function, or TxnFun cause other errors/exits
                            exit:TxnFunReason  -> {error, {txn_fun_error, TxnFunReason}};
                            error:TxnFunReason -> {error, {txn_fun_error, TxnFunReason}}
                        end
                end,

            %% Actually run TxnFun
            RetryWrapperFunResult = RetryWrapperFun(NumDeadlockRetries, RetryWrapperFun),

            %% We must always cleanup, even if TxnFun has returned an error

            try ?BDB_PC:release_bdb_port_handle(BdbPortHandle) of

                ok -> RetryWrapperFunResult   % final success (this is returned to caller)

            catch
                {bdb_port_coordinator_error, {release_bdb_port_handle, BdbPortHandle, still_owned_resources, NumResources}} ->

                    case RetryWrapperFunResult of 

                        {ok, TxnFunResult} ->
                            %% TxnFun returned normally, but accidentally left resources open
                            %% We obviously consider this to be an error (so will throw on the client),
                            %% but we do return TxnFun's return value to help with diagnostics
                            {error, {txn_fun_error, {return_with_open_resources, NumResources, TxnFunResult}}};

                        _ ->
                            %% Here TxnFun failed for some reason, so it's not suprising that 
                            %% it left resources open.  We ignore the latter, and return the real TxnFun error.
                            RetryWrapperFunResult
                    end;

                ReleaseReason -> 
                    %% Some other throw exception. We don't expect this (we are more likely to get an exit signal 
                    %% from the port_coordinator crashing or the port beint closed) but convert it to an error tuple 
                    %% anyway, as this function is 'no throw'.
                    {error, {bdb_helper_error, {release_bdb_port_handle, ReleaseReason}, RetryWrapperFunResult}}

                %% TODO:
                %% If release_bdb_port_handle() failed to contact the port_coordinator (to tell it to unlink the caller)
                %% (e.g. if the port_coordiantor happened to crash after TxnFun returned but before we call 
                %% release_bdb_port_handle) then OTP (gen_server:call) will kill the caller process.
                %% In that case we want to ignore the exit and just return the result of TxnFun, as the caller doesn't care 
                %% (at the moment) that the port coordinator died after TxnFun completed.
                %%
                %% HOWEVER, we cannot simply catch exits here and ignore them, because release_bdb_port_handle() 
                %% _also_ calls group_valmap_cleanup in the context of the caller process, and that function 
                %% might exit (if the port has been slammed shut).  In that case it is NOT safe to return TxnFun's result, 
                %% as if TxnFun left resources open (e.g. did not commit the txn) then it's result is invalid 
                %% (e.g. the transaction will be rolled back, either in the C port stop() function, or in BDB recovery).
                %% So this needs some thinking about.

                %% We don't catch other exits or errors here as they would mean a bug in system code not caller code
            end

    catch 
        %% acquire_bdb_port_handle failed
        AcquireReason -> {error, {bdb_helper_error, {acquire_bdb_port_handle, AcquireReason}}}
    end.


%%
%% Private functions
%%

%% logger(String) ->
%%     logger(String, []).
logger(String, Args) ->
    case process_info(self(), registered_name) of
        {registered_name, ProcessName} ->
            io:format("~w: " ++ String, [ProcessName | Args]);
        [] ->
            case get('$ancestors') of
                undefined ->
                    io:format("~w: " ++ String, [self() | Args]);
                Ancestors ->
                    %% This requires that we only create support processes with proc_lib:spawn_link
                    io:format("~w: support process ~w: " ++ String, [hd(Ancestors) | [self() | Args]])
            end
    end.
