%%%-------------------------------------------------------------------
%%% File    : berkeley_db_port_coordinator.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%%   A gen_server that owns a single instance of an EDTK-generated
%%%   berkeley_db port.
%%%
%%%   This server does *not* abstract the BDB API.
%%%   Instead, it allows multiple Erlang processes to share 
%%%   a single BDB port in a safe way.
%%%   
%%%   Pros of this approach:
%%%        - For linked-in drivers, commands and replies go 
%%%          directly to/from the port, without any intermediate server.
%%%          (In pipe-mode, the server forwards the replies to the 
%%%           correct client processes).
%%%
%%%        - Less code, as no need to abstract the BDB API.
%%%
%%%   Cons:
%%%        - Clients must obey some (fairly simple) restrictions.
%%%
%%%        - The full BDB API is exposed, which can cause havoc.(e.g. closing a 
%%%          transaction that has a cursor open causes an environment panic).
%%%
%%%        - driver_caller and driver_send_term are not supported in pipe-mode.
%%%          So this server recieves the messages, and we have to arrange 
%%%          for the client's pid to be included in the message, so the server
%%%          can forward the message to the client.
%%%
%%%   The server also offers some important conveniences
%%%        - multiple registered instances are supported (one per BDB environment)
%%%        - simple configuration with good defaults for linked & pipe modes
%%%        - convenient replication, checkpointing, deadlock-detection, memp_trickle, etc.
%%%
%%%   --- API ---
%%%
%%%   init()
%%%
%%%        - Opens a single port instance and DBENV during init()
%%%          (and runs recovery, uses DB_REGISTER if desired, etc).
%%%
%%%        - Traps exits.  If an exit is received from a client, 
%%%          handle_info will call ?BDB:group_valmap_cleanup(DeadClientGroupValmapId).
%%% 
%%%         If replication is enabled then a gen_event process is created with the same name as the port coordinator
%%%         process name, plus a suffix.
%%%         Clients can retrieve that name with get_replication_gen_event_proc_name/0,1.
%%%         (Internally, the server must call dynamic_replication_gen_event_proc_name/0 to get the same information.)
%%%     
%%%
%%%   acquire_bdb_port_handle()
%%%
%%%        returns   BdbPortHandle   (includes the DBENV handle as member 'dbenv')
%%%      or throws   {bdb_port_coordinator_error, Reason}
%%%
%%%      The BdbPortHandle contains a unique GroupValmapId that will be used to 
%%%      track resources (e.g. transactions, cursors) opened by the caller, and clean them 
%%%      up if the caller dies.
%%%
%%%      The BdbPortHandle is bound to the pid of the caller, and should not be passed to 
%%%      other processes.
%%%
%%%      Side-effects: ***Links the calling process to the server (if successful)***
%%%
%%%      Client may then:
%%%        - Directly hit the returned port handle via functions in 
%%%          berkeley_db_drv.erl (the reply will be sent to the client).
%%%        - Open its own transactions and cursors (**but see below**)
%%%
%%%      Client MUST
%%%        - When opening transactions and cursors, client **MUST** 
%%%          specify the UniqueGroupValmapId returned from acquire_bdb_port_handle().
%%%          This allows the server can clean up if the client exits.
%%%
%%%      Client must NOT
%%%        - Pass the BdbPortHandle to any other processes (it won't work anyway).
%%%
%%%        - Open any Db handles itself.  It must use get_db_handle().
%%%
%%%          Note that get_db_handle does not support opening (creation) of databases within a 
%%%          transaction (because there is a race condition -- other clients may attempt to use 
%%%          that DB handle before the original client commits their transaction -- which BDB 
%%%          prevents, and worse, the original client may abort the transaction and leave the 
%%%          server with a cached but invalid DB handle, which is then handed out to other clients).
%%%          
%%%          This means that if a client *must* create a DB handle, it must do it itself.  
%%%          However, this is potentially dangerous:
%%%
%%%          - DB handles cost about 600 bytes of heap, plus a BDB mutex and a locker (and a valmap slot).
%%%            (Fortunately, they don't each require an OS file descriptor.)
%%%            So if many Erlang processes do this, there may be resource problems.
%%%
%%%          - If a client dies while holding a DB handle that it created/opened (i.e. group_valmap_cleanup sees a DB handle),
%%%            and if there happen to be any open txn handles (not even in-use at that instant, just open)
%%%            then the OS process (i.e. probably the entire Erlang VM) will be hard aborted.
%%%            This cannot happen if clients stick to using get_db_handle (as the port_coordinator owns 
%%%            all of the handles).
%%%
%%%        - Close any Db handles, or the DBENV handle.
%%%
%%%        - Remove or truncate databases, or remove the environment.
%%%
%%%        - Pass it's GroupValmapId to any other process.
%%%          In general, processes and GroupValmapIds should have a 
%%%          one-to-one relationship).
%%%
%%%        - Allow ANY concurrent operations to use the same 
%%%          transaction or cursor handle.  This means not specifying 
%%%          the same handle to multiple concurrent non_block operations,
%%%          and NOT passing transaction or cursor handles to other 
%%%          processes.
%%%
%%%
%%%   get_db_handle(DBType, DBName)   DBType is one of bdb_DB_BTREE, bdb_DB_HASH, DB_QUEUE etc.
%%%                                   DBName is a Filename or {Filename, DatabaseNameWithinFile}
%%%
%%%          returns   DB
%%%         or throws  {bdb_port_coordinator_error, Reason}
%%%
%%%        This returns a copy of a shared Db handle for a database.
%%%        If the database is not open it is opened, and if it does 
%%%        not exist it is created.  The gen_server caches these and
%%%        reuses them.  
%%%
%%%        Note: when creating the database it is NOT possible for the caller to provide 
%%%        a Txn handle to be passed to db_open, because that txn may later be rolled back
%%%        by the caller (and we will have cached an invalid handle).  Also there would be a 
%%%        condition: other clients may request that handle before the original caller 
%%%        commits the txn.  BDB returns an error if a DB handle is used from another txn 
%%%        when the txn that created that DB has not yet been committed.
%%%        If the caller really needs to create a database as part of a transaction then 
%%%        it should create the database handle itself.  
%%%        (Each DB handle requires about 600 bytes of heap plus a locker and a mutex
%%%        so this approach should be applied with caution -- e.g. after the transaction 
%%%        commits, consider closing the private handle and asking the port_coordinator 
%%%        for a cached/cacheable handle.)
%%%
%%%        If the database is already open then the type must match
%%%        and the config list must be 'compatible' (not contradict
%%%        the existing config list)
%%%
%%%      Client must NOT close Db handles, or truncate or delete databases.
%%%
%%%  
%%%   close_unused_database(DBName, UserTxn)   DBName is a Filename or {Filename, DatabaseNameWithinFile}
%%%                                             UserTxn is required (cannot be 'void')
%%%
%%%          returns   ok
%%%         or throws  {bdb_port_coordinator_error, Reason}
%%%
%%%        This API allows a database to closed in preparation for calling env_dbremove() or env_dbrename() 
%%%        (i.e. before physically deleting or renaming the database).
%%%
%%%        IMPORTANT: This is an **EXTREMELY DANGEROUS** API, with stringent preconditions for safe use.
%%%        Incorrect use will both
%%%              - crash Berkeley DB very badly (undefined behaviour)
%%%              - **almost certainly corrupt the transaction log and make the entire BDB environment UNRECOVERABLE**.
%%%
%%%        Also, you should think strongly before using this API if you are using replication as:
%%%
%%%              - DB handles closed on the master are NOT automatically closed on replicas.
%%%
%%%              - At the time of writing there is a Berkeley DB bug (#15076 or TAR 5815135.992)
%%%                that will cause replication to hang if a database is physically removed on the 
%%%                master but there are DB handles open for that database on replicas.
%%%
%%%        The pre-conditions for this API can ONLY be achieved by conventions implemented and enforced by
%%%        the application (for a suggested convention, see below).
%%%
%%%        The port coordinator does NOT enforce these preconditions, or even detect if they have been
%%%        violated.  The port_coordinator does NOT track which clients have requested a given DB 
%%%        handle, as this would be both expensive (it is a very common operation) and pointless (clients 
%%%        can open their own private handles to the database without knowledge or monitoring by the 
%%%        port coordinator).
%%%
%%%        The pre-conditions are:
%%%
%%%        1. DBName must have been opened by at least one client using get_db_handle().
%%%           (Note that this is done as a side-effect of the coordination scheme described below.)
%%%
%%%        2a. No client may be using (or later begin to use) the shared DB handle for the passed database
%%%           (i.e. the handle returned by get_db_handle() & co).
%%%
%%%        2b. Subtle point: 2a implies that this DB handle CANNOT EVER be used in distributed transaction. 
%%%            This is because the port_coordinator may be told to do incremental cleanup of distributed txns
%%%            - so a 2PC Txn handle may outlive the other valmap slot entries owned by that group
%%%            (basically, the port_coordinator takes ownership of the txn).
%%%            So, if you do need to combine close_unused_database() with (hopefully few) distributed transactions, 
%%%            then processes performing the distributed transactions should open their own DB handles for those transactions.
%%%            HOWEVER, doing so is playing **Russian Roulette**, as if the client crashes with an unresolved prepared txn
%%%            the OS process will be HARD-ABORTED, as it is not safe to close a DB handle that is involved 
%%%            in an unresolved txn.
%%%            Note that most applications do not need to call close_unused_database, in which case, 
%%%            using get_db_handle() with txn_prepare is quite safe (as the port_coordinator owns 
%%%            all of the DB handles and guarantees that the DB handles will stay open longer than an unresolved txns,
%%%            (or, if the port_coordinator cannot resolve all prepared txns before shutting down, then it (the port_coordinator)
%%%            will make the final round of db_close calls safe -- i.e. prevent them from logging via a Sleepycat hack 
%%%            - but it cannot do that for the db_close inside close_unused_database, as the Sleepycat hack manipulates a global flag).
%%%
%%%        3. [Related, but not a true precondition of close_unused_database]
%%%
%%%           As the goal of calling close_unused_database() is in preparation for 
%%%           immediately calling env_dbremove() or env_dbrename() it must be noted that 
%%%           those BDB APIs alsl require that there are **ZERO** open database handles for the 
%%%           database in question.
%%%
%%%           This function (close_unused_database) only closes the handle owned by the port_coordinator.
%%%           So to safely use those BDB APIs the application is also responsible for:
%%%
%%%                - preventing ANY application process from holding or opening a private DB handle 
%%%                  for DBName, until after it has been renamed/removed.
%%%
%%%                - preventing ANY application process from calling get_db_handle() & co. 
%%%                  on DBName, until after it has been renamed/removed.
%%%
%%%        Applications can enforce this preconditions above by ensuring that all operations that use 
%%%        a database which might ever be passed to close_unused_database() first obtain and hold a shared (read) lock 
%%%        on some other object outside of this database.
%%%
%%%        IMPORTANT: the operations should obtain the lock **before** they call get_db_handle(), or db_open(),
%%%        or before they use a privately-cached DB handle obtained from those calls.
%%%        The operat8on should only release that lock after it is certain that it will not refer to the DB handle again 
%%%        (without re-acquiring the read lock first).
%%%
%%%        An example of the 'other object' that must be locked might a 'meta-data' record in a different database 
%%%        -- note that it must be in an entirely different physical file if you want to remove a file-level database.
%%%
%%%        Then the application code that calls close_unused_database() should obtain an exclusive lock on the 
%%%        'meta-data' object (thus waiting for any existing operations that use the database have finished, 
%%%        and also preventing any new operations from starting to using the database)
%%%
%%%
%%%   release_bdb_port_handle()
%%%
%%%      returns   ok
%%%
%%%      or throw  {bdb_port_coordinator_error, {release_bdb_port_handle, BdbPortHandle, still_owned_resources, NumResourcesCleanedUp}}
%%%
%%%      Side-effects: Unlinks the client process from the server.
%%%
%%%      (If the server is shutdown or learns of a serious error 
%%%       -- e.g. DB_RUNRECOVERY -- then it will kill all of it's active 
%%%       clients -- i.e. any that have called acquire_bdb_port_handle() but not yet 
%%%       called release_bdb_port_handle).
%%%
%%%      Client should
%%%        - Have closed all of it's cursors and transactions
%%%          before calling this function.
%%%          To err on the side of caution, the server calls 
%%%          group_valmap_cleanup(), just in case the client did not close all 
%%%          of it's txn and cursor handles (hence the error return value).
%%%          But it is very bad practice to rely on this. 
%%%
%%%      Client must NOT
%%%        - Use the port instance again after calling this function.
%%%
%%%
%%%   get_replication_gen_event_proc_name()
%%%
%%%      Returns the locally-registered name of a gen_event process that 
%%%      will publish replication events (if replication is enabled).
%%%     
%%%   Additional server responsibilities:
%%%
%%%   - If the BDB port is opened in 'pipe' mode then the port driver 
%%%     cannot use driver_caller/driver_send_term to send replies
%%%     directly to the Erlang process that sent each command.
%%%     Instead, the replies come to this server as it is the 
%%%     port's connected process.  Therefore we must forward the 
%%%     replies to the correct clients. (The client's pid is included 
%%%     in the reply, so we don't have to do any book-keeping).
%%%
%%%     Important: It is critical that we garbage-collect binaries 
%%%     aggresively.  For this reason, if pipe-mode is used 
%%%     the port_coordinator is spawned using {fullsweep_after, 0}
%%%     so the collector will consider reducing reference counts of 
%%%     binaries whenever the process is garbage collected.
%%%     Still there is the concern that the port_coordinator won't 
%%%     be garbage-collected very often, as it's State does not 
%%%     change much (only when clients register/unregister, and 
%%%     when DB handles are first opened).  In pipe-mode we might 
%%%     want to add explicit calls to the collector on a timer.
%%%     However, the last paragraph of the following mailining-list 
%%%     post suggests that we shouldn't have to do this.
%%%         
%%%        From: Bjorn Gustavsson  
%%%        Date: 20 Oct 2003 14:02:08 +0200 
%%%
%%%        Garbage collections of binaries is a tricky issue.
%%%         
%%%        To clarify how binaries are removed: It is done by a combination of
%%%        garbage collection and reference counters. Binaries are represented in
%%%        each process as small objects on the heap. Those objects are also part
%%%        of a linked list (one list per process). After each garbage collection,
%%%        the linked list is scanned to which objects were not moved (thus, no longer
%%%        referenced). For each unreferenced object, the reference count in the binary
%%%        itself is decreased.
%%%         
%%%        As a consequence of the above, if a binary is referenced from the
%%%        old generation heap, it cannot be removed until a new major collection
%%%        (a.k.a. fullsweep collection) is done. How often a major collection occurs
%%%        depends on the application.
%%%         
%%%        If a process is known to have large binaries but small amounts of live
%%%        heap data, the process can be started using the spawn_opt/4, passing in the
%%%        '{fullsweep_after,0}' option. (See the documentation for spawn_opt/4 for
%%%        further details.) That will make every GC a major collection (fullsweep),
%%%        cleaning away any binaries that are no longer referenced.
%%%         
%%%        Note that a major collection copies all data that is is still alive; if you
%%%        have large amounts of live data on the heap doing a major collection every time
%%%        will typically be slower than the default generational GC (which only does a
%%%        major collection when there is no more place on the old generation heap).
%%%         
%%%        BTW, when a process aquires a new binary, it gets "penalty points" so that it
%%%        will be garbage collected sooner. But a binary that has survived one GC will
%%%        not generate further "penalty points".
%%%         
%%%
%%% TODO
%%%
%%% - in kill_all_support_procs_and_clients_and_close_environment,
%%%   recalculate the timout when waiting for group_valmap_cleanup_non_block 
%%%   and db_close_non_block to reply, so that we don't wait for 
%%%   a maximum of N * timeout.
%%%
%%% - consider adding a defragmenter process using DB->compact
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_port_coordinator).

-behaviour(gen_server).

%% Basic API (server is registered locally using this module's name)
-export([start_link/1, 
         start_link/2,
         stop/0, 
         acquire_bdb_port_handle/0,
         get_replication_gen_event_proc_name/0]).

%% Variants of the above that also take a process name (as their first argument)
%% which will be registered locally.
%% These are necessary if using multiple BDB environments in the same VM 
%% (each environment has its own own port_coordinator)
-export([start_link/3,
         stop/1, 
         acquire_bdb_port_handle/1,
         get_replication_gen_event_proc_name/1]).

%% APIs that take a bdb_port_handle (the port_coordinator process name is obtained from that handle)
-export([get_db_handle/3, 
         get_db_handle_with_db_config/4,
         close_unused_database/2,
         release_bdb_port_handle/1]).


%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).


%%
%% Internal declarations
%%

-define(REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX, "_replication_gen_event").

%% Support processes are checkpointer, cache_trickler, deadlock_detector, replication_gen_event, ...
%% Any that access the BDB environment are spawned in a 'logically suspended' state, and told to start
%% after the environment has been opened.
-record(support_proc, {name, pid, start_fun, stop_fun}).

-record(state, {bdb_port_handle,          % includes dbenv
                db_handle_cache,          % a dict, mapping filenames to open DB handles 
                next_group_valmap_id,
                client_group_valmap_ids,  % a dict, mapping client pids to valmap group ids
                support_procs}).          % list of support_proc records

-define(get_erlang_port(State),            (State#state.bdb_port_handle)#bdb_port_handle.erlang_port).
-define(get_server_group_valmap_id(State), (State#state.bdb_port_handle)#bdb_port_handle.group_valmap_id).
-define(get_proc_name(State),              (State#state.bdb_port_handle)#bdb_port_handle.port_coordinator_name).

%% In addition to the state record, we use the process dictionary to cache some **immutable** state 
%% (created by init).  These are just parts of the initial server config that we need fast access to.
%%
%%   file_permissions
%%   default_db_config_proplist


-include("berkeley_db_drv.hrl").
-define(BDB,   berkeley_db_drv).
-define(BDB_U, berkeley_db_utilities).

%% This is for inet:gethostname() and inet:gethostbyname() which are used in handle_replication_option_before_env_open/2.
%% -include("inet.hrl").

%%-define(INFO_LOGGER,    error_logger:info_msg).
%%-define(WARNING_LOGGER, error_logger:warning_msg).
%%-define(ERROR_LOGGER,   error_logger:error_msg).
-define(INFO_LOGGER,    logger).
-define(WARNING_LOGGER, logger).
-define(ERROR_LOGGER,   logger).


%%
%% Config defaults
%%

-define(DEFAULT_PORT_START_FUNCTION, start).

%% env_dir is required, so there is no default.

%% If no explicit override of 'env_open_flags' is specified 
%% then we use a set of 'base' flags (see below) PLUS one of:
%%
%%   DB_PRIVATE  if port_start_function = 'start'
%%   DB_REGISTER if port_start_function = 'start_pipe'
%%
%% PLUS (if replication is not disabled)
%%
%%   DB_INIT_REP
%%
%% Note that we *always* specify DB_RECOVER. 
%%
%% We assume that we are NOT sharing the BDB environment with other OS processes (although it would work in pipe-mode),
%% so DB_PRIVATE is ok.
%%
%% DB_PRIVATE is prefered as it is significantly faster than DB_REGISTER, due to use of thread mutexes instead of process mutexes.
%% But DB_PRIVATE does not support DB_REGISTER.
%%
%% DB_REGISTER is very useful to prevent a new server from being started while an old one is still shutting down 
%% (which could cause data-corruption). That can happen much more easily if we are in pipe-mode, as ?BDB:shutdown() 
%% returns to Erlang before pipe-main has finished closing the environment.

-define(DEFAULT_ENV_OPEN_BASE_FLAGS,  [bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN]).

-define(DEFAULT_CONFIG_PROPLIST, 
        [{port_start_function,                                ?DEFAULT_PORT_START_FUNCTION}, 
         {default_receive_reply_timeout_milliseconds,         ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},
         {env_open_receive_reply_timeout_milliseconds,        20000},                                     % recovery can take significant time
         {checkpoint_receive_reply_timeout_milliseconds,      2 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},    % also used for memp_trickle
         {file_permissions,                                   8#644},                                     % user read/write, group read, world read
         {env_cache_size_mb_and_fragments,                    {50, 1}},                
         {env_set_flags,                                      [bdb_DB_AUTO_COMMIT, bdb_DB_LOG_AUTOREMOVE, bdb_DB_REGION_INIT, bdb_DB_TIME_NOTGRANTED]},
         {env_verbose_types,                                  [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
         {default_db_config_proplist,                         []},
         {max_num_concurrent_client_procs,                    2 * ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS},  % adjust to prevent saturation of BDB (note: not all client procs start explicit txns)
         {min_limit_concurrent_transactions,                  ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS},      % a BDB-enforced *lower* limit (real limit may be up to approx 3x this number)
         {{initial_num_valmap_slots, bdb_cursor},             ?BDB_INITIAL_NUM_VALMAP_CURSOR_SLOTS},
         {{initial_num_valmap_slots, bdb_db_txn},             ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS},
         {{initial_num_valmap_slots, bdb_db},                 ?BDB_INITIAL_NUM_VALMAP_DB_SLOTS},
         {auto_deadlock_detect,                               bdb_DB_LOCK_DEFAULT},          % or 'disabled'
         {deadlock_detector_milliseconds_and_policy,          {5000, bdb_DB_LOCK_DEFAULT}},  % or 'disabled' (**see edtk-1.2/TODO: currently required to prevent rare threadpool starvation**)
         {default_global_txn_timeout_microseconds,            5000 * 1000},                  % or 'disabled' (**see edtk-1.2/TODO: currently required to prevent rare threadpool starvation**)
         {default_global_lock_timeout_microseconds,           disabled},
         {max_locks,                                          1000},
         {max_lockers,                                        1000},
         {max_locked_objects,                                 1000},
         %%{max_mutexes,                                      ????},                  % OBTAIN FROM db_stat AFTER TESTING (default sizing is poor: SR: 14691) 
         {log_region_size_bytes,                              65000},                 % may need to increase if very many databases are created
         {log_file_size_bytes,                                10485760},              % 10MB
         {log_buffer_size_bytes,                              32768},                 % increasing may improve performance of long-running or large txns, or highly-concurrent applications
         {checkpoint_seconds_and_min_kb,                      {30, 100}},             % or 'disabled'  (see also checkpoint_receive_reply_timeout_milliseconds)
         {cache_trickle_seconds_and_percent,                  {15, 80}},              % or 'disabled'  (see also checkpoint_receive_reply_timeout_milliseconds)
         {cache_throttle_io_sleep_microseconds_and_write_ops, disabled},              % disabled is same as {0, anything}
         {replication,                                        disabled},              % see below for sub-options
         {distributed_transactions,                           disabled},              % see below for sub-options

         %% On stack size for Berkeley DB, Keith Bostic (VP Product Development, Sleepycat) says 
         %%
         %%  "We do not document a smallest, or even a "safe", stack size value for any release of Berkeley DB.
         %%   Berkeley DB puts relatively large objects on the stack, on the order of 1 or 2KB 
         %%   (for examples, a routine might have 2 1KB buffers to hold pathnames for a file system rename call).
         %%   Berkeley DB does not go deeply recursive.
         %%   If I had to pick a number out of the air, I'd guess that 8KB would be risky, 16KB would probably work, 
         %%   and 32KB would be safe.  I have no experimental data to prove that, however."
         %%
         %% The EDTK threadpool implementation uses neglible stack.
         {{thread_count_stacksize_qlenlimit, workers},      {10, 32*1024, 1000}},
         {{thread_count_stacksize_qlenlimit, valmap_stop},  {10, 32*1024, 1000}},   % if replication is enabled, these threads block for acks
         {{thread_count_stacksize_qlenlimit, consume_wait},  {0, 32*1024,    0}},
         {{thread_count_stacksize_qlenlimit, housekeeping},  {3, 32*1024,   10}}
        ]).


%% env_set_flags:
%%  - DB handle operations for which no explicit transaction handle was specified, and which modify databases in the database environment, will be automatically enclosed within a transaction. 
%%    (Note: this does NOT apply to cursors; non-transactional cursors must NOT perform write operations in a transactional environment.)
%%  - Log files should be auto-removed asap (only needed for catastrophic recovery, which is too expensive).
%%  - Regions (most significantly, the page-cache) should be page-faulted into memory when created/joined.
%%  - Lock timeouts (e.g. used by berkeley_db_msg_queue.erl, and when using lock_vec to unblock a db_get(CONSUME_WAIT) 
%%    should cause DB_LOCK_NOTGRANTED, not DB_LOCK_DEADLOCK (so we can distinguish between the two)       
%%
%% min_limit_concurrent_transactions : 
%% We might want this to be more than the number of db_txn valmap slots because many operations use auto-commit transactions, 
%% which don't use a valmap slot (but do use an internal transaction).
%% However, BDB typically allows up to 3x this number of transactions; this only sets a lower limit, not a hard upper limit.
%% So the default (#valmap slots) might be fine.
%%
%% default_db_config_proplist:
%% A list containing any of the following (they are all optional):
%%   
%%    [{page_size,                      NumBytes},
%%     {set_flags,                      Flags},    (should include bdb_DB_CHKSUM if an explicit page-size is chosen)
%%     {set_btree_min_keys_per_page,    MinKey},
%%     {set_fixed_record_size_bytes,    NumBytes},
%%     {set_fixed_record_pad_byte,      ByteValue},
%%     {set_queue_extent_size_db_pages, NumDBPages}]
%%
%% replication :
%% This is disabled by default. 
%% If is enabled by the caller then it has a whole set of possible sub-options.
%% Multiple 'replication' entries may appear in the main config list. 
%% Each 'replication' entry must be a list of sub-options, or the atom 'disabled'.
%% If any replication entry is 'disabled' then any others must also be 'disabled' (i.e. you cannot provide conflicting config).
%% The sub-lists from all of the entries are concatenated (in order) and then treated as a normal proplist.

-define(REPLICATION_REQUIRED_KEYS, [this_hostname_port, 
                                    this_site_priority,
                                    peer_hostnames_ports
                                   ]).

%% If replication is explicitly enabled then these defaults are used for some keys (hence they are not listed in REPLICATION_REQUIRED_KEYS)
-define(DEFAULT_REPLICATION_CONFIG_IF_ENABLED, [{initial_role_and_threadcount,  {bdb_DB_REP_FULL_ELECTION, 2}},
                                                {ack_policy,                     bdb_DB_REPMGR_ACKS_QUORUM},
                                                {limit_transfer_per_call_bytes,         1024 * 1024}, % 1 MB
                                                {{timeout, acks_microseconds},               500000}, % 0.5 seconds
                                                {{timeout, election_milliseconds},             3000}, % 3   seconds
                                                {{timeout, election_retry_milliseconds},       1000}, % 1   second
                                                {{timeout, connection_retry_milliseconds},     3000}  % 3   seconds
                                               ]).

%% Format for the other options:
%%
%%       {replication, [{this_hostname_port,   {"hostname", port},
%%                      {this_site_priority,   N},
%%                      {peer_hostnames_ports, [{"hostname", port}, ... ],
%%  optional:           {event_handlers,       [{ ModuleName|{ModuleName,Id}, Args}]}     % (see below)
%%                     ]}
%%
%% If replication is enabled then it spawns a gen_fsm (owned by the port coordinator) with a locally-registered 
%% public name, which available via the get_replication_gen_event_proc_name/0,1 API.
%% Any process may add/deleted handlers from this gen_event at any time.
%%
%% To avoid missing events during replication start up, handlers may be registered before repmgr_start is called 
%% by listing them in the optional event_handlers argument.  
%% This is only a convenience, not a requirement.
%% Note that event_handlers is currently processed from the server process, calling gen_event:add_handler, 
%% -- i.e. NOT calling add_sup_handler.  So the process that calls start_link will NOT currently
%% receive any {gen_event_EXIT,Handler,Reason} messages.  I think that's the behaviour we want.
%% If necessary, we could register these from the client process (in start_link) and use add_sup_handler, 
%% so the registering process would get those messages.
%%
%%
%% distributed_transactions :
%% 
%% This is disabled by default.  If enabled then the config works like that for 'replication'
%% -- i.e. multiple lists of sub-options are allowed and there are some defaults.

-define(DISTRIBUTED_TRANSACTIONS_REQUIRED_KEYS, [global_transaction_manager_node]).

-define(DEFAULT_DISTRIBUTED_TRANSACTIONS_CONFIG_IF_ENABLED, []).
%%-define(DEFAULT_DISTRIBUTED_TRANSACTIONS_CONFIG_IF_ENABLED, [{incremental_resolution_timeout_milliseconds, 5000}]).

%% Format for the other options:
%%
%%       {distributed_transactions, [{global_transaction_manager_node, NodeName},
%%                                  ]}
%%
%% If distributed_transactions is enabled (and replication is disabled) then init() runs 
%% txn_recover immediately after opening the environment, and if there are any prepared but unresovled 
%% transactions it attempts to ask the GTM whether they should be committed or aborted.
%%
%% If *both* distributed_transactions AND replication are enabled then port_coordinator does NOT run 
%% txn_recover immediately.  Instead txn_recover is always run by a node when it receives A DB_REP_MASTER
%% event.
%%
%% Note: as of BDB v4.5, it is illegal to closee DB handles if there is a prepared but unresolved transaction
%% -- the log will be left in an unrecoverable state.  Calling txn_discard does not help (and is illegal in this context).
%% So if the port is closed while there are unresolved prepared txns, the driver uses a patch from Sleepycat
%% (patch.14785]) to prevent DB:close from logging the close.  
%% The port_coordinators own clean-shutdown code alos has to obey this 
%% rule - see the call to ?BDB:make_db_close_safe_if_there_are_unresolved_transactions(). 
%% Sleepycat plan to fix this problem in a future release (probably in a different way to the mechanism used 
%% by this patch, so the driver and port-coordinator code will have to change then).


%% Some timeouts for shutdown
-define(TERMIMATE_GROUP_VALMAP_CLEANUP_TIMEOUT_MILLISECONDS, 5000).
-define(TERMIMATE_DB_CLOSE_TIMEOUT_MILLISECONDS,             5000).


-define(STRINGIZE(Arg), ??Arg).


%%====================================================================
%% API
%%====================================================================

-define(THROW_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_port_coordinator_error, Error})
           end).

%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%--------------------------------------------------------------------

start_link(EnvDir) ->
    start_link(?MODULE, EnvDir, []).

start_link(EnvDir, ConfigProplist) ->
    start_link(?MODULE, EnvDir, ConfigProplist).

stop() ->
    gen_server:cast(?MODULE, stop).

acquire_bdb_port_handle() ->
    ?THROW_ERRORS( gen_server:call(?MODULE, acquire_bdb_port_handle) ).

%% TODO: An alternative would be to store dynamic_replication_gen_event_proc_name() in the process dictionary 
%% (during init) and fetch it with a gen_server:call()
get_replication_gen_event_proc_name() ->
    list_to_atom(?STRINGIZE(?MODULE) ++ ?REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX).


%% These variants of the above are used when we want to use multiple port_coordinators 
%% (IMPORTANT: each port_coordiantor MUST use a DIFFERENT BDB ENV.
%% TODO: we could use a global ETS table to track which port coord is using 
%% which (absolute path) EnvDir.  However, most systems will only use one, so 
%% the extra complexity is not worth it (yet).

start_link(ServerRegisterName, EnvDir, ConfigProplist) ->
    %% See comments at top of file.  In 'pipe' mode it's essential that we use {fullsweep_after, 0}
    %% as the port_coordinator will receive ALL results from the BDB driver.
    SpawnOptions = 
        case proplists:get_value(port_start_function, ConfigProplist, ?DEFAULT_PORT_START_FUNCTION) of
            start      -> [];
            start_pipe -> [{fullsweep_after, 0}]
        end,    
    gen_server:start_link({local, ServerRegisterName}, ?MODULE, [{env_dir, EnvDir} | ConfigProplist], [{spawn_opt, SpawnOptions}]).

stop(ServerRegisterName) ->
    gen_server:cast(ServerRegisterName, stop).

acquire_bdb_port_handle(ServerRegisterName) ->
    ?THROW_ERRORS( gen_server:call(ServerRegisterName, acquire_bdb_port_handle) ).

get_replication_gen_event_proc_name(ServerRegisterName) ->
    list_to_atom(atom_to_list(ServerRegisterName) ++ ?REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX).


%%
%% APIs that require a BdbPortHandle
%%

get_db_handle(BdbPortHandle, DBType, DBName) when is_record(BdbPortHandle, bdb_port_handle) ->
    get_db_handle_with_db_config(BdbPortHandle, DBType, DBName, []).

%% DBConfigProplist is for configuring the DB handle before db_open is called
%% (The default option list is simply empty.)
%%
%%    {set_fixed_record_size_bytes,      NumBytes}
%%    {set_fixed_record_pad_byte,       ByteValue}
%%    {set_queue_extent_size_db_pages,  NumDBPages}
%%
get_db_handle_with_db_config(BdbPortHandle, DBType, DBName, DBConfigProplist) when is_record(BdbPortHandle, bdb_port_handle) ->
    ?THROW_ERRORS( gen_server:call(BdbPortHandle#bdb_port_handle.port_coordinator_pid, {get_db_handle, {DBType, DBName, DBConfigProplist}}) ).


%% **EXTREMELY DANGEROUS API** 
%% See documentation for stringent preconditions to make this 

close_unused_database(BdbPortHandle, DBName) when is_record(BdbPortHandle, bdb_port_handle) ->

    ?THROW_ERRORS( gen_server:call(BdbPortHandle#bdb_port_handle.port_coordinator_pid, {close_unused_database, DBName}) ).


release_bdb_port_handle(BdbPortHandle) ->

    %% We need to call group_valmap_cleanup and we want it to be a serious error 
    %% (throw) if the caller still owns any resources.
    %% Therefore the call to group_valmap_cleanup must be blocking, as it needs the return value.
    %% We cannot block the port_coordinator server process for this call, as group_valmap_cleanup
    %% might take considerable time.
    %% Rather than spawning a temporary process on the server to do it (and using gen_server:reply)
    %% it seems reasonable and fair to make the client call it.

    NumResourcesCleanedUp = ?BDB:group_valmap_cleanup(BdbPortHandle, BdbPortHandle#bdb_port_handle.group_valmap_id),

    %% We *always* tell the server to forget about this client (as cleanup has been done).
    %% We don't wrap this in a ?THROW_ERRORS() invocation because if NumResourcesCleanedUp is not 0 then 
    %% that is a more important error.

    %% This is a call as the client might want to terminate immediately after release_bdb_port_handle returns
    %%(so if we used a cast the EXIT signal would race the cast).
    ServerResult = gen_server:call(BdbPortHandle#bdb_port_handle.port_coordinator_pid, release_bdb_port_handle),

    %% Now throw if the client leaked any resources, or (if it didn't leak resources), if the server command returned an error.
    case NumResourcesCleanedUp of
        0 -> ?THROW_ERRORS(ServerResult);
        _ -> throw({bdb_port_coordinator_error, {release_bdb_port_handle, BdbPortHandle, still_owned_resources, NumResourcesCleanedUp}})            
    end.



%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%
%% Args is a 'proplist' of tagged tuples
%%
%% Required
%%    {env_dir, "directory path"}                          
%%
%% Optional
%%    See defaults in comments at start.
%%
%%--------------------------------------------------------------------
init(SpecifiedConfigProplist) ->

    ?INFO_LOGGER("starting (~p)~n", [self()]),
    process_flag(trap_exit, true),

    %% This is the only required arg (no default)
    EnvDir = case proplists:lookup(env_dir, SpecifiedConfigProplist) of 
                 {env_dir, Value} -> Value;
                 none             -> throw({missing_required_config_option, env_dir})
             end,

    %% We need this to construct the default env_open_flags
    PortStartFunc = proplists:get_value(port_start_function, SpecifiedConfigProplist, ?DEFAULT_PORT_START_FUNCTION),

    %% Some options (replication and distributed_transactions) have lists of sub-options, and we want to treat
    %% multiple occurences of those parameters as cummulative (comcatenated) proplists in their own right.  
    %% So preprocess those first, otherwise get_dominant_proplist_values will discard all but the first occurance.
    %% Also, we cannot have a 'disabled' parameter and a sublist.  Either all occurences are 'disabled', or none of them are.
    PropList2 = ?BDB_U:normalize_sub_proplists_for_key(
                   replication, 
                   SpecifiedConfigProplist,
                   ?DEFAULT_REPLICATION_CONFIG_IF_ENABLED),

    PropList3 = ?BDB_U:normalize_sub_proplists_for_key(
                   distributed_transactions, 
                   PropList2,
                   ?DEFAULT_DISTRIBUTED_TRANSACTIONS_CONFIG_IF_ENABLED),

    %% Append the default config to the list specified by the caller, and extract all of the first occurences (the final config).
    ConfigProplist 
        = ?BDB_U:get_dominant_proplist_values(
             [PropList3,
              [{env_open_flags,
                case PortStartFunc of 
                    start      -> 
                        ?INFO_LOGGER("init: Using DB_PRIVATE as port_start_function = 'start'~n"),
                        [bdb_DB_PRIVATE];
                    start_pipe ->
                        ?INFO_LOGGER("init: Using DB_REGISTER as port_start_function = 'start_pipe'~n"),
                        [bdb_DB_REGISTER]
                end
                ++ case proplists:lookup(replication, PropList3) of
                       none                    -> [];
                       {replication, disabled} -> [];
                       {replication, _}        -> [bdb_DB_INIT_REP]
                   end
                ++ ?DEFAULT_ENV_OPEN_BASE_FLAGS}],
              ?DEFAULT_CONFIG_PROPLIST]),

    %% Store the final config in the process dictionary, so that handle_init_option_before_env_open 
    %% can use it if necessary, and so that it will be logged if the server crashes.
    %% (Also the main code may check this if needed, e.g. to see if we are in linked-in or pipe-mode.
    %% If all commands start timing-out (deadlock) and we are in pipe-mode we can simply kill and 
    %% restart the server.  But in linked-in mode we will need to restart the entire Erlang VM.)
    put(config_proplist, ConfigProplist),

    %% Display the final config.
    lists:foreach(fun({Key, Value}) -> ?INFO_LOGGER("init: ~51w = ~200p~n", [Key, Value]) end, lists:keysort(1, ConfigProplist)),

    %% Actually open the BDB port instance
    DefaultBdbPortHandle = ?BDB:PortStartFunc(),

    %% Add the port_coordinator details, and possibly override receive_reply_timeout
    BdbPortHandle = DefaultBdbPortHandle#bdb_port_handle{port_coordinator_name = get_port_coordinator_proc_name(),
                                                         port_coordinator_pid  = self(),
                                                         receive_reply_timeout = proplists:get_value(default_receive_reply_timeout_milliseconds, 
                                                                                                     ConfigProplist)}, 

    DriverSidePortNumber = ?BDB:get_port_number(BdbPortHandle),
    ok = ?BDB:set_port_label(BdbPortHandle, atom_to_list(get_port_coordinator_proc_name())),

    {VersionMajor, VersionMinor, VersionPatch, VersionText} = ?BDB:db_version(BdbPortHandle),

    ?INFO_LOGGER("init: Successfully opened port ~w (on driver side: ~w). Berkeley DB version = ~w.~w.~w ~p~n", 
                 [BdbPortHandle#bdb_port_handle.erlang_port, DriverSidePortNumber, VersionMajor, VersionMinor, VersionPatch, binary_to_list(VersionText)]),

    %% Validate and apply the config options before openening the BDB environment.
    %% The config handler clauses have the option of adjusting the server State record.
    %% They can also move or add options to a list of AfterEnvOpenConfigOptions.

    {State1, AfterEnvOpenConfigOptions}
        = lists:foldl(fun(ConfigArg, State) -> handle_init_option_before_env_open(ConfigArg, State) end, 
                      {#state{bdb_port_handle         = BdbPortHandle, 
                              db_handle_cache         = dict:new(), 
                              next_group_valmap_id    = BdbPortHandle#bdb_port_handle.group_valmap_id + 1,  % will be 1
                              client_group_valmap_ids = dict:new(), 
                              support_procs           = []}, 
                       []}, % AfterEnvOpenConfigOptions
                      ConfigProplist),

    ?INFO_LOGGER("init: Opening environment (and running recovery)...~n"),
    ok = ?BDB:env_open(BdbPortHandle#bdb_port_handle{receive_reply_timeout 
                                                     = proplists:get_value(env_open_receive_reply_timeout_milliseconds, 
                                                                           get(config_proplist))},
                       EnvDir, 
                       proplists:get_value(env_open_flags, ConfigProplist), 
                       get(file_permissions)),
    ?INFO_LOGGER("init: Recovery finished.~n"),

    %% Arguably, we ought to call do_distributed_transaction_recovery EVERY time, 
    %% as a site may have been configured for distributed transactions before and left 
    %% unresolve txns, even though it is not configured for distributed txns now.
    %% However,
    %%  - this is both unlikely and a serious operator error (not a normal use-case)
    %%  - we cannot resolve any txns without config pointing to the global_transaction_manager
    %%  - it is safe, as BDB will reject txn_begin until all unresolved txns have been resolved.
    case proplists:lookup(distributed_transactions, ConfigProplist) of
        {distributed_transactions, disabled} ->
            ok;
        {distributed_transactions, _} ->
            ?INFO_LOGGER("init: Calling do_distributed_transaction_recovery~n"),
            do_distributed_transaction_recovery(BdbPortHandle, bdb_DB_FIRST),
            ?INFO_LOGGER("init: do_distributed_transaction_recovery finished.~n")
    end,

    %% Now do all config that requires the environment to be open
    State2 = lists:foldl(fun(ConfigArg, State) -> handle_init_option_after_env_open(ConfigArg, State) end, 
                         State1,
                         AfterEnvOpenConfigOptions),

    %% Now we allow our support processes to begin work.
    lists:foreach(fun(SupportProc) -> StartFun = SupportProc#support_proc.start_fun,
                                      StartFun(SupportProc) 
                  end, 
                  State2#state.support_procs),

    {ok, State2}.


%% Handlers for config options
handle_init_option_before_env_open({env_cache_size_mb_and_fragments, {CacheSizeMB, Fragments}}, {State, _} = StateAndPostponedConfigOptions) ->
    ok = ?BDB:env_set_cachesize(State#state.bdb_port_handle, 
                               CacheSizeMB div 1024,                 % gigabytes
                               (CacheSizeMB rem 1024) * 1024 * 1024, % bytes
                               Fragments),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_open_flags, _}, StateAndPostponedConfigOptions) ->
    %% Handled by env_open
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({default_receive_reply_timeout_milliseconds, _}, StateAndPostponedConfigOptions) ->
    %% Handled by init()
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_open_receive_reply_timeout_milliseconds, _}, StateAndPostponedConfigOptions) ->
    %% Handled by init()
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({checkpoint_receive_reply_timeout_milliseconds, _}, StateAndPostponedConfigOptions) ->
    %% Handled when we start the checkpointer process
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_verbose_types, VerboseTypeList}, {State, _} = StateAndPostponedConfigOptions) ->
    lists:foreach(
      fun(bdb_DB_VERB_RECOVERY)    -> ok = ?BDB:env_set_verbose(State#state.bdb_port_handle, bdb_DB_VERB_RECOVERY, 1);
         (bdb_DB_VERB_REGISTER)    -> ok = ?BDB:env_set_verbose(State#state.bdb_port_handle, bdb_DB_VERB_REGISTER, 1);
         (bdb_DB_VERB_REPLICATION) -> ok = ?BDB:env_set_verbose(State#state.bdb_port_handle, bdb_DB_VERB_REPLICATION, 1)
      end,
      VerboseTypeList),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_set_flags, Flags}, {State, _} = StateAndPostponedConfigOptions) when is_list(Flags) ->
    ok = ?BDB:env_set_flags(State#state.bdb_port_handle, Flags, 1),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({file_permissions, FilePermissions}, StateAndPostponedConfigOptions) when is_integer(FilePermissions) ->
    put(file_permissions, FilePermissions),        % Store for later use (an immutable cache)
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({default_db_config_proplist, Proplist}, StateAndPostponedConfigOptions) when is_list(Proplist) ->
    true = ?BDB_U:is_proplist(Proplist),           % Check well-formedness (our own type assertion function)
    put(default_db_config_proplist, Proplist),     % Store for later use (an immutable cache)
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_num_concurrent_client_procs, N}, StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    put(max_num_concurrent_client_procs, N),       % Store for later use (an immutable cache)
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({min_limit_concurrent_transactions, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    %% This can only be set before opening the environment.  
    %% The actual limit is likely to be 2 to 3 times the number set here - hence the 'minimum' tag.
    ok = ?BDB:set_num_valmap_slots(State#state.bdb_port_handle, bdb_db_txn, N),
    ok = ?BDB:env_set_tx_max(State#state.bdb_port_handle, N),    
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({{initial_num_valmap_slots, ValmapType}, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:set_num_valmap_slots(State#state.bdb_port_handle, ValmapType, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({auto_deadlock_detect, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({auto_deadlock_detect, Policy}, {State, _} = StateAndPostponedConfigOptions) when is_atom(Policy) ->
    ok = ?BDB:set_lk_detect(State#state.bdb_port_handle, Policy),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({default_global_lock_timeout_microseconds, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({default_global_lock_timeout_microseconds, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    ok = ?BDB:env_set_timeout(State#state.bdb_port_handle, N, bdb_DB_SET_LOCK_TIMEOUT),  
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({default_global_txn_timeout_microseconds, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({default_global_txn_timeout_microseconds, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    ok = ?BDB:env_set_timeout(State#state.bdb_port_handle, N, bdb_DB_SET_TXN_TIMEOUT),  
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_locks, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:set_lk_max_locks(State#state.bdb_port_handle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_lockers, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:set_lk_max_lockers(State#state.bdb_port_handle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_locked_objects, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:set_lk_max_objects(State#state.bdb_port_handle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_mutexes, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:mutex_set_max(State#state.bdb_port_handle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({log_region_size_bytes, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:set_lg_regionmax(State#state.bdb_port_handle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({log_file_size_bytes, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:set_lg_max(State#state.bdb_port_handle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({log_buffer_size_bytes, N}, {State, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ok = ?BDB:set_lg_bsize(State#state.bdb_port_handle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({{thread_count_stacksize_qlenlimit, workers}, {Threads, StackSizeBytes, QLenLimit}}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Threads), Threads >= 0, is_integer(StackSizeBytes), StackSizeBytes >= 0,is_integer(QLenLimit), QLenLimit >= 0 ->
    ok = ?BDB:set_threadpool_params(State#state.bdb_port_handle, 0, Threads, StackSizeBytes, QLenLimit),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({{thread_count_stacksize_qlenlimit, valmap_stop}, {Threads, StackSizeBytes, QLenLimit}}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Threads), Threads >= 0, is_integer(StackSizeBytes), StackSizeBytes >= 0,is_integer(QLenLimit), QLenLimit >= 0 ->
    ok = ?BDB:set_threadpool_params(State#state.bdb_port_handle, 1, Threads, StackSizeBytes, QLenLimit),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({{thread_count_stacksize_qlenlimit, consume_wait}, {Threads, StackSizeBytes, QLenLimit}}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Threads), Threads >= 0, is_integer(StackSizeBytes), StackSizeBytes >= 0,is_integer(QLenLimit), QLenLimit >= 0 ->
    ok = ?BDB:set_threadpool_params(State#state.bdb_port_handle, 2, Threads, StackSizeBytes, QLenLimit),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({{thread_count_stacksize_qlenlimit, housekeeping}, {Threads, StackSizeBytes, QLenLimit}}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Threads), Threads >= 0, is_integer(StackSizeBytes), StackSizeBytes >= 0, is_integer(QLenLimit), QLenLimit >= 0 ->
    ok = ?BDB:set_threadpool_params(State#state.bdb_port_handle, 3, Threads, StackSizeBytes, QLenLimit),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({checkpoint_seconds_and_min_kb, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({checkpoint_seconds_and_min_kb, {Seconds, MinKBytes}}, {State, PostponedConfigOptions}) when is_integer(Seconds), is_integer(MinKBytes), Seconds >= 0, MinKBytes >= 0 ->
    %% Note that the support processes should NOT call acquire_bdb_port_handle() [as all client processes must] as that would
    %% register them as clients of the server, and terminate would kill them, rather than shutting them
    %% down cleanly.  So we pass our instance of bdb_port_handle, and hack the issued_to_pid member in process startup.
    CheckpointReceiveReplyTimeout = proplists:get_value(checkpoint_receive_reply_timeout_milliseconds, get(config_proplist)),
    {State#state{support_procs = [#support_proc{name      = checkpointer, 
                                                pid       = proc_lib:spawn_link(
                                                              fun() -> 
                                                                      checkpointer_process(
                                                                        State#state.bdb_port_handle, 
                                                                        Seconds, 
                                                                        MinKBytes,
                                                                        CheckpointReceiveReplyTimeout)
                                                              end),
                                                start_fun = fun(SupportProc) -> SupportProc#support_proc.pid ! go end,
                                                stop_fun  = fun(SupportProc) -> SupportProc#support_proc.pid ! stop end}
                                  | State#state.support_procs]},
     PostponedConfigOptions};

handle_init_option_before_env_open({deadlock_detector_milliseconds_and_policy, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({deadlock_detector_milliseconds_and_policy, {Milliseconds, Policy}}, {State, PostponedConfigOptions}) when is_integer(Milliseconds), is_atom(Policy), Milliseconds >= 0 ->
    {State#state{support_procs = [#support_proc{name      = deadlock_detector, 
                                                pid       = proc_lib:spawn_link(
                                                              fun() -> 
                                                                      deadlock_detector_process(
                                                                        State#state.bdb_port_handle, 
                                                                        Milliseconds, 
                                                                        Policy) 
                                                              end),
                                                start_fun = fun(SupportProc) -> SupportProc#support_proc.pid ! go end,
                                                stop_fun  = fun(SupportProc) -> SupportProc#support_proc.pid ! stop end}
                                  | State#state.support_procs]},
     PostponedConfigOptions};

handle_init_option_before_env_open({cache_trickle_seconds_and_percent, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({cache_trickle_seconds_and_percent, {Seconds, Percent}}, {State, PostponedConfigOptions}) when is_integer(Seconds), Seconds >= 0, is_integer(Percent), Percent > 0, Percent =< 100 ->
    CheckpointReceiveReplyTimeout = proplists:get_value(checkpoint_receive_reply_timeout_milliseconds, get(config_proplist)),
    {State#state{support_procs = [#support_proc{name      = cache_trickler, 
                                                pid       = proc_lib:spawn_link(
                                                              fun() -> 
                                                                      cache_trickler_process(
                                                                        State#state.bdb_port_handle, 
                                                                        Seconds, 
                                                                        Percent,
                                                                        CheckpointReceiveReplyTimeout)
                                                              end),
                                                start_fun = fun(SupportProc) -> SupportProc#support_proc.pid ! go end,
                                                stop_fun  = fun(SupportProc) -> SupportProc#support_proc.pid ! stop end}
                                  | State#state.support_procs]},
     PostponedConfigOptions};

handle_init_option_before_env_open({cache_throttle_io_sleep_microseconds_and_write_ops, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({cache_throttle_io_sleep_microseconds_and_write_ops, {Microseconds, SequentialWriteOps}}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Microseconds), Microseconds >= 0, is_integer(SequentialWriteOps), SequentialWriteOps >= 0 ->
    ok = ?BDB:set_mp_max_write(State#state.bdb_port_handle, Microseconds, SequentialWriteOps),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({replication, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({replication, ReplicationConfigList}, {State, PostponedConfigOptions}) ->

    lists:foreach(fun(RequiredKey) -> case proplists:lookup(RequiredKey, ReplicationConfigList) of 
                                          {RequiredKey, _} -> ok;
                                          none             -> throw({missing_required_replication_config, RequiredKey, ReplicationConfigList})
                                      end
                  end,
                  ?REPLICATION_REQUIRED_KEYS),

    %% Clients will probably want to be notified of replication events, so spawn a gen_fsm with a well-known name.
    %% We CANNOT call get_replication_gen_event_proc_name() here as we don't know which variant to call 
    %% (i.e. this port_coordinator might not be registered as ?MODULE). We need to get our dynamic process name.
    GenEventProcessName = dynamic_replication_gen_event_proc_name(),
    {ok, GenEventPid} = gen_event:start_link({local, GenEventProcessName}),
    ?INFO_LOGGER("Started ~w (~w)~n", [GenEventProcessName, GenEventPid]),

    %% Replication has a bunch of sub-options -- process them.
    lists:foldl(fun(ConfigArg, StateAndPostponedConfigOptions) -> 
                        handle_replication_option_before_env_open(ConfigArg, StateAndPostponedConfigOptions) 
                end, 
                {State#state{support_procs = [#support_proc{name      = replication_gen_event, 
                                                            pid       = GenEventPid,
                                                            start_fun = fun(_SupportProc) -> ok end,
                                                            stop_fun  = fun(SupportProc) -> gen_event:stop(SupportProc#support_proc.pid) end}
                                              | State#state.support_procs]}, 
                 PostponedConfigOptions},
                ReplicationConfigList);

handle_init_option_before_env_open({distributed_transactions, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({distributed_transactions, DistributedTxnConfigList}, {State, PostponedConfigOptions}) ->

    %% Check that all required parameters are present
    lists:foreach(fun(RequiredKey) -> case proplists:lookup(RequiredKey, DistributedTxnConfigList) of 
                                          {RequiredKey, _} -> ok;
                                          none             -> throw({missing_required_distributed_transactions_config, RequiredKey, DistributedTxnConfigList})
                                      end
                  end,
                  ?DISTRIBUTED_TRANSACTIONS_REQUIRED_KEYS),

    %% Distributed transaction support has a bunch of sub-options -- process them.
    %%
    %% Also, we need a support process to resolve any prepared txns that we discover when cleaning up after a client 
    %% (group_valmap_cleanup sends us an untagged event for these).
    lists:foldl(fun(ConfigArg, AccStateAndPostponedConfigOptions) -> 
                        handle_distributed_transactions_option_before_env_open(ConfigArg, AccStateAndPostponedConfigOptions) 
                end, 
                {State#state{support_procs = [#support_proc{name      = incrementally_resolve_prepared_txns, 
                                                            pid       = proc_lib:spawn_link(fun() -> incrementally_resolve_prepared_txns_process(
                                                                                                       State#state.bdb_port_handle,
                                                                                                       DistributedTxnConfigList) 
                                                                                            end),
                                                            start_fun = fun(SupportProc) -> SupportProc#support_proc.pid ! go end,
                                                            stop_fun  = fun(SupportProc) -> SupportProc#support_proc.pid ! stop end}
                                              | State#state.support_procs]},
                 PostponedConfigOptions},
                DistributedTxnConfigList);

%% ... these are handled by special-case code in init/1, but we need to accept them here
handle_init_option_before_env_open({port_start_function, _}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({env_dir, _}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;

%% ... and anything else is a typo
handle_init_option_before_env_open(Other, _StateAndPostponedConfigOptions) ->
    throw( {unrecognized_config_option, Other} ).


%% Handlers for replication sub-options
handle_replication_option_before_env_open({this_hostname_port, {Hostname, TCPIPPort}}, {State, _} = StateAndPostponedConfigOptions) ->

%%     %% Check that the host-name provided for use by BDB (repmgr) does actually match this host name, or we will have chaos.
%%     %% TODO: is this misguided?   (answer; probably -- hence it is commented out for now)
%%     {ok, ThisHostNameString} = inet:gethostname(),
%%     {ok, HostEnt} = inet:gethostbyname(ThisHostNameString, inet),
%%     true = lists:member(Hostname, HostEnt#hostent.h_aliases),

    ok = ?BDB:repmgr_set_local_site(State#state.bdb_port_handle, Hostname, TCPIPPort, []),
    StateAndPostponedConfigOptions;
handle_replication_option_before_env_open({this_site_priority, Priority}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Priority), Priority >= 0 ->
    ok = ?BDB:rep_set_priority(State#state.bdb_port_handle, Priority),
    StateAndPostponedConfigOptions;
handle_replication_option_before_env_open({peer_hostnames_ports, HostnameTCPIPPortsList}, {State, _} = StateAndPostponedConfigOptions) ->
    %% TODO: Eventually we want to allow DB_REPMGR_PEER to be set.  (Only one, the last one, is used).
    %% But repmgr BDB v4.6 is going to improve that feature, so we leave it for now.
    lists:foreach(fun({Hostname, TCPIPPort}) when is_integer(TCPIPPort) -> 
                          ok = ?BDB:repmgr_add_remote_site(State#state.bdb_port_handle, Hostname, TCPIPPort, [])
                  end,
                  HostnameTCPIPPortsList),
    StateAndPostponedConfigOptions;

handle_replication_option_before_env_open({initial_role_and_threadcount, {_Role, _NumReplicationThreads}} = Option, {State, PostponedConfigOptions}) when is_atom(_Role), is_integer(_NumReplicationThreads), _NumReplicationThreads >= 0 ->
    %% These are the arguments for repmgr_start(), which we cannot call until after env_open has been called.
    %% So we move it to the AfterEnvOpenConfigOptions section.
    {State, [Option | PostponedConfigOptions]};

handle_replication_option_before_env_open({ack_policy, Policy}, {State, _} = StateAndPostponedConfigOptions) when is_atom(Policy) ->
    ok = ?BDB:repmgr_set_ack_policy(State#state.bdb_port_handle, Policy),
    StateAndPostponedConfigOptions;

handle_replication_option_before_env_open({limit_transfer_per_call_bytes, Bytes}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Bytes), Bytes > 0, Bytes < 1024 * 1024 * 1024 ->
    ok = ?BDB:rep_set_limit(State#state.bdb_port_handle, 0, Bytes),  % always less than 1GB :)
    StateAndPostponedConfigOptions;

handle_replication_option_before_env_open({{timeout, acks_microseconds}, Microseconds}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Microseconds), Microseconds >= 0 ->
    ok = ?BDB:rep_set_timeout(State#state.bdb_port_handle, bdb_DB_REP_ACK_TIMEOUT, Microseconds),    
    StateAndPostponedConfigOptions;
handle_replication_option_before_env_open({{timeout, election_milliseconds}, Milliseconds}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Milliseconds), Milliseconds >= 0 ->
    ok = ?BDB:rep_set_timeout(State#state.bdb_port_handle, bdb_DB_REP_ELECTION_TIMEOUT, Milliseconds * 1000),
    StateAndPostponedConfigOptions;
handle_replication_option_before_env_open({{timeout, election_retry_milliseconds}, Milliseconds}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Milliseconds), Milliseconds >= 0 ->
    ok = ?BDB:rep_set_timeout(State#state.bdb_port_handle, bdb_DB_REP_ELECTION_RETRY, Milliseconds * 1000),
    StateAndPostponedConfigOptions;
handle_replication_option_before_env_open({{timeout, connection_retry_milliseconds}, Milliseconds}, {State, _} = StateAndPostponedConfigOptions) when is_integer(Milliseconds), Milliseconds >= 0 ->
    ok = ?BDB:rep_set_timeout(State#state.bdb_port_handle, bdb_DB_REP_CONNECTION_RETRY, Milliseconds * 1000),
    StateAndPostponedConfigOptions;

handle_replication_option_before_env_open({event_handlers, ModuleNamesArgs}, StateAndPostponedConfigOptions) when is_list(ModuleNamesArgs) ->
    GenEventName = dynamic_replication_gen_event_proc_name(),
    lists:foreach(fun({ModuleName, Args}) ->
                          %%?INFO_LOGGER("init: calling gen_event:add_handler{~w, ~w, ~w)~n", [GenEventName, ModuleName, Args]),
                          %% TODO: we don't call add_sup_handler here as we are installing the *caller's* handler but the 
                          %% gen_event_EXIT message woud come to the port coordinator, not the caller.
                          %% Options:
                          %% 1. call add_sup_handler in the start() before returning to the caller (might miss events)
                          %% 2. add a handle_info clause for gen_event_EXIT and forward it to the process that spawned us (or require that the event_handlers config argument specify a pid to send the message too)                          
                          ok = gen_event:add_handler(GenEventName, ModuleName, Args)
                  end,
                  ModuleNamesArgs),
    StateAndPostponedConfigOptions.


%% Handlers for distributed_transactions sub-options
handle_distributed_transactions_option_before_env_open({global_transaction_manager_node, _}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions.


%% handlers for options delayed until after env_open has been called.
handle_init_option_after_env_open({initial_role_and_threadcount, {Role, NumReplicationThreads}}, State) ->
    ?INFO_LOGGER("init: calling repmgr_start~n"),
    %% This may generate an untagged event. If so, we will handle it later in handle_info.
    ok = ?BDB:repmgr_start(State#state.bdb_port_handle, NumReplicationThreads, Role),    
    State.


%%--------------------------------------------------------------------
%%
%% Helper processes started in init()
%%
%%--------------------------------------------------------------------

checkpointer_process(PortCoordBdbPortHandle, PeriodSeconds, MinKBytes, CheckpointReceiveReplyTimeout) ->
    receive go -> 
            ?INFO_LOGGER("checkpointer_process ~p starting (period: ~p seconds, min kbytes: ~p, receive_reply_timeout: ~p)~n", [self(), PeriodSeconds, MinKBytes, CheckpointReceiveReplyTimeout]),
            checkpointer_loop(PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid         = self(),
                                                                     receive_reply_timeout = CheckpointReceiveReplyTimeout}, 
                              PeriodSeconds, 
                              MinKBytes)
    end.

checkpointer_loop(BdbPortHandle, PeriodSeconds, MinKBytes) ->
    receive
        stop  -> ?INFO_LOGGER("checkpointer exiting~n");
        Other -> exit({checkpointer, {unexpected_msg, Other}})
    after
        PeriodSeconds * 1000 ->
            ok = ?BDB:txn_checkpoint(BdbPortHandle, MinKBytes, 0, []),

            %%?INFO_LOGGER("checkpointer: Successful. Sleeping ~p seconds~n", [PeriodSeconds]),
            checkpointer_loop(BdbPortHandle, PeriodSeconds, MinKBytes)
    end.
        

cache_trickler_process(PortCoordBdbPortHandle, PeriodSeconds, Percent, CheckpointReceiveReplyTimeout) ->
    receive go -> 
            ?INFO_LOGGER("cache_trickler_process ~p starting (period: ~p seconds, percent: ~p, receive_reply_timeout: ~p)~n", [self(), PeriodSeconds, Percent, CheckpointReceiveReplyTimeout]),
            cache_trickler_loop(PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid         = self(),
                                                                       receive_reply_timeout = CheckpointReceiveReplyTimeout}, 
                                PeriodSeconds, 
                                Percent)
    end.

cache_trickler_loop(BdbPortHandle, PeriodSeconds, Percent) ->
    receive
        stop  -> ?INFO_LOGGER("cache_trickler exiting~n");
        Other -> exit({cache_trickler, {unexpected_msg, Other}})
    after
        PeriodSeconds * 1000 ->
            _NumFlushedPages = ?BDB:memp_trickle(BdbPortHandle, Percent),

            %%?INFO_LOGGER("cache_trickler: Successful (flushed ~p pages). Sleeping ~p seconds~n", [NumFlushedPages, PeriodSeconds]),
            cache_trickler_loop(BdbPortHandle, PeriodSeconds, Percent)
    end.


deadlock_detector_process(PortCoordBdbPortHandle, PeriodMilliseconds, Policy) ->
    receive go -> 
            ?INFO_LOGGER("deadlock_detector_process ~p starting (period: ~p milliseconds)~n", [self(), PeriodMilliseconds]),
            deadlock_detector_loop(PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid = self()}, 
                                   PeriodMilliseconds, 
                                   Policy)
    end.

deadlock_detector_loop(BdbPortHandle, PeriodMilliseconds, Policy) ->
    receive
        stop  -> ?INFO_LOGGER("deadlock_detector exiting~n");
        Other -> exit({deadlock_detector, {unexpected_msg, Other}})
    after
        PeriodMilliseconds ->
            _NumBrokenLocks = ?BDB:lock_detect(BdbPortHandle, [], Policy),

            %%?INFO_LOGGER("deadlock_detector_process: Successful (broke ~p locks). Sleeping ~p milliseconds~n", [NumBrokenLocks, PeriodMilliseconds]),
            deadlock_detector_loop(BdbPortHandle, PeriodMilliseconds, Policy)
    end.


%% This calls txn_commit and txn_abort on prepared transactions that
%% were once owned by now-dead clients (group_valmap_cleanup sends an
%% untagged event to the port coordinator for these, and the port
%% coordinator sends them to us.)
%%
%% Note: if commit or abort fail due to being selected to resolve a
%% deadlock then we die, and kill the entire port_coordinator.  There
%% is nothing else we can do -- if we lose this txn handle, we have to
%% run recovery so we can run txn_recover.
incrementally_resolve_prepared_txns_process(PortCoordBdbPortHandle, DistributedTxnConfigList) ->
    receive go -> 
            ?INFO_LOGGER("incrementally_resolve_prepared_txns ~p starting~n", [self()]),
            incrementally_resolve_prepared_txns_loop(PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid = self()}, 
                                                     proplists:get_value(global_transaction_manager_node, DistributedTxnConfigList))
    end.

incrementally_resolve_prepared_txns_loop(BdbPortHandle, GTMNode) ->
    receive
        stop -> 
            ?INFO_LOGGER("incrementally_resolve_prepared_txns exiting~n");

        %% TODO (minor optimization): As this involves a round-trip to the GTM, don't send each one individually
        %% -- wait for a few milliseconds and send a batch request.
        {need_to_resolve, <<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID, Txn} ->

            ?INFO_LOGGER("incrementally_resolve_prepared_txns: asked to resolve txn ~w ~p~n", [TxnCounter, Txn]),

            try berkeley_db_global_transaction_manager:resolve(GTMNode, DistributedTxnGUID) of 

                {DistributedTxnGUID, committed} -> 
                    ?INFO_LOGGER("incrementally_resolve_prepared_txns: GTM says to commit dist. txn ~w.~n", [TxnCounter]),
                    ok = ?BDB:txn_commit(BdbPortHandle, Txn, []),
                    incrementally_resolve_prepared_txns_loop(BdbPortHandle, GTMNode);

                {DistributedTxnGUID, aborted} -> 
                    ?INFO_LOGGER("incrementally_resolve_prepared_txns: GTM says to abort dist. txn ~w.~n", [TxnCounter]),
                    ok = ?BDB:txn_abort(BdbPortHandle, Txn),
                    incrementally_resolve_prepared_txns_loop(BdbPortHandle, GTMNode)

            catch
                {bdb_gtm_error, {DistributedTxnGUID, _}} = Error -> 
                    ?ERROR_LOGGER("incrementally_resolve_prepared_txns: GTM:resolve returned error ~p for dist. txn ~w. Exiting~n", [Error, TxnCounter]),
                    exit({gtm_returned_error, Error})
            end;

        Other -> exit({incrementally_resolve_prepared_txns, {unexpected_msg, Other}})
    end.


%%--------------------------------------------------------------------
%% Function: handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call(acquire_bdb_port_handle, {FromPid, _RequestTag}, State) ->

    %% We choose to make it  an error for a client process to call acquire_bdb_port_handle twice.
    %% Otherwise we would have to track multiple group valmap ids, to cleanup if that process exits,
    %% and we would have to ref-count calls to link before we can unlink in release_port.
    case dict:find(FromPid, State#state.client_group_valmap_ids) of

        {ok, _GroupValmapId} -> 
            ?WARNING_LOGGER("handle_info: Registered process ~p called acquire_bdb_port_handle again~n", [FromPid]),
            {reply, {error, caller_already_has_a_bdb_port_handle}, State};

        error ->

            %% Limit the number of concurrent users
            %% note: dict:size() is very fast, or we would track it ourselves.
            CurrNumClients = dict:size(State#state.client_group_valmap_ids),
            MaxNumClients  = get(max_num_concurrent_client_procs),

            if CurrNumClients >= MaxNumClients ->

                    ?WARNING_LOGGER("handle_info: acquire_bdb_port_handle failed with too_many_client_procs~n"),
                    %% We include both numbers in the error as in future I want to mak max_num_concurrent_client_procs 
                    %% adjustable at runtime, and if someone lowers it then immediately afterwards CurrNumClients may 
                    %% actually be bigger than MaxNumClients for a while.
                    {reply, {error, {too_many_client_procs, CurrNumClients, MaxNumClients}}, State};

               true ->
                    %% We link to all active clients as we need to call group_valmap_cleanup etc if they crash.
                    %% If the caller has already exited then link/1 will fail. It was written before exceptions 
                    %% and has the following behaviour.
                    %%
                    %% From http://www.erlang.org/ml-archive/erlang-questions/200510/msg00204.html
                    %%
                    %%  link(Pid) -> error(noproc) if Pid is local and does not exist and
                    %%                             caller has process_flag(trap_exit, false)
                    %%               true          otherwise
                    %%   
                    %%  And if the Pid dies, or even was remote and did not exist
                    %%  (those cases can not be discriminated), the caller will get
                    %%  an exit signal that can be converted to a regular message
                    %%  with process_flag(trap_exit, true).

                    case (catch link(FromPid)) of 
                        {'EXIT', {noproc, _}} ->
                            %% Client process has already exited
                            {reply, {error, failed_to_link_to_caller}, State};

                        true -> 
                            %% Each client is issued a unique group-valmap-id (a 32-bit unsigned int).
                            %% If we overflow that range then we need to wrap around.
                            %% But we must never issue the server's own group-valmap-id, as that owns the 
                            %% DBENV and all of the DB handles)    
                            %%?INFO_LOGGER("handle_call(acquire_bdb_port_handle): Called by client pid ~p. Assigned group valmap-id ~p.~n", [FromPid, State#state.next_group_valmap_id]),

                            CandidateNextGroupValmapId = (State#state.next_group_valmap_id + 1) rem (1 bsl 32),

                            NextGroupValmapId = 
                                if CandidateNextGroupValmapId == ?get_server_group_valmap_id(State) ->
                                        ?get_server_group_valmap_id(State) + 1;
                                   true ->
                                        CandidateNextGroupValmapId
                                end,

                            {reply, 
                             {ok, (State#state.bdb_port_handle)#bdb_port_handle{group_valmap_id = State#state.next_group_valmap_id,
                                                                                issued_to_pid   = FromPid}}, % owned by client
                             State#state{next_group_valmap_id    = NextGroupValmapId,
                                         client_group_valmap_ids = dict:store(FromPid, 
                                                                              State#state.next_group_valmap_id, 
                                                                              State#state.client_group_valmap_ids)}}
                    end
            end
    end;

%% Convenience function for when caller only provides Filename (no DatabaseNameWithinFile)
handle_call({get_db_handle, {RequestedDBType, Filename, SpecifiedDBConfigProplist}}, FromPidRequestTag, State) when is_list(Filename) ->
    handle_call({get_db_handle, {RequestedDBType, 
                                 {Filename, ""}, 
                                 SpecifiedDBConfigProplist}}, 
                FromPidRequestTag,
                State);

handle_call({get_db_handle, {RequestedDBType, 
                             {_Filename, _DatabaseNameWithinFile} = DBName, 
                             SpecifiedDBConfigProplist}}, 
            {FromPid, _RequestTag}, 
            State) ->

    %% Combine the SpecifiedDBConfigProplist with the port-coordinator's default config.
    %% 'safe' here means an internal try/catch, so we don't kill the server if the client passes something 
    %% that is totally malformed.

    case ?BDB_U:safe_resolve_config_proplist(SpecifiedDBConfigProplist, 
                                             get(default_db_config_proplist)) of

        {ok, NewDBConfigProplist} ->
            %% Is this database already open>
            case dict:find(DBName, State#state.db_handle_cache) of

                {ok, {DB, ExistingDBType, ExistingDBConfigProplist}} ->

                    %% We need to check that DBType and DBConfigPropList are compatible
                    %% (i.e. Another process may have opened this database with a differnt type, or a different page-size etc.)
                    case RequestedDBType of 

                        ExistingDBType ->
                            %% Database is the correct type.  
                            %% Is the config OK?
                            case find_mismatched_db_config_entries(NewDBConfigProplist, ExistingDBConfigProplist) of 
                                [] ->
                                    %%?INFO_LOGGER("handle_call(get_db_handle): Database ~p is already open, so returning existing DB handle (~p) to client pid ~p ~n", [DBName, DB, FromPid]),
                                    %%We don't reference-count DB handles, as there is no way to make it accurate.
                                    {reply, {ok, DB}, State};

                                MismatchList ->
                                    ?WARNING_LOGGER("handle_call(get_db_handle): Attempt by client pid ~p to open existing DB handle (~p) with same type (~p) but different config (mismatches: ~p)~n", [FromPid, DBName, ExistingDBType, MismatchList]),
                                    {reply, {error, {mismatched_db_config, MismatchList}}, State}
                            end;

                        _ ->
                            ?WARNING_LOGGER("handle_call(get_db_handle): Attempt by client pid ~p to open existing DB handle (~p) with wrong type (existing: ~p, new ~p)~n", [FromPid, DBName, ExistingDBType, RequestedDBType]),
                            {reply, {error, {mismatched_db_type, ExistingDBType, RequestedDBType}}, State}
                    end;

                error ->
                    %% We have not yet opened a database with this filename.
                    %% The database may or may not already exist on disk.
                    %% If it doesn't exist on disk, we create it.
                    open_unknown_database(RequestedDBType, 
                                          DBName, 
                                          NewDBConfigProplist, 
                                          FromPid, 
                                          State)
            end;

        {error, Error} ->
            ?WARNING_LOGGER("handle_call(get_db_handle): Attempt by client pid ~p to open DB (~p) with malformed config ~p (error: ~p)~n", [FromPid, DBName, SpecifiedDBConfigProplist, Error]),
            {reply, {error, {bad_db_config, Error, SpecifiedDBConfigProplist}}, State}

    end;


handle_call({close_unused_database, Filename}, From, State) when is_list(Filename) ->
    handle_call({close_unused_database, {Filename, ""}}, From, State);

handle_call({close_unused_database, DBName}, {FromPid, _RequestTag}, State) ->

    ?INFO_LOGGER("handle_call(close_unused_database): From client process ~p. DBName = ~p~n", [FromPid, DBName]),

    case dict:find(DBName, State#state.db_handle_cache) of

        {ok, {DB, _ExistingDBType, _ExistingDBConfigProplist}} ->

            try
                ok = ?BDB:db_close(State#state.bdb_port_handle, DB, []),

                %% Remove the DB handle from our cache
                {reply, 
                 {ok, ok},
                 State#state{db_handle_cache = dict:erase(DBName, State#state.db_handle_cache)}}

            catch
                Class:Error ->
                    %% The BDB documentation for db_close says: 
                    %%  "The DB handle may not be accessed again after DB->close is called, regardless of its return."
                    {reply, 
                     {ok, {db_close_failed, {Class, Error}}}, 
                     State#state{db_handle_cache = dict:erase(DBName, State#state.db_handle_cache)}}
            end;

        error ->
            %% We have not yet opened a database with this filename 
            {reply, 
             {error, {unknown_database, DBName}},
             State}
    end;


handle_call(release_bdb_port_handle, {FromPid, _RequestTag}, State) ->

    %% We use dict:find instead of dict:fetch as a badly behaving client 
    %% should not be able to kill the server.
    case dict:find(FromPid, State#state.client_group_valmap_ids) of

        {ok, _ClientGroupValmapId} -> 
            %%?INFO_LOGGER("handle_call(release_bdb_port_handle): From client process ~p. Unlinking and cleaning group_valmap_id ~p~n", [FromPid, ClientGroupValmapId]),

            unlink(FromPid),

            {reply, 
             {ok, ok},  % tuple required as client call is inside ?THROW_ERRORS
             State#state{client_group_valmap_ids = dict:erase(FromPid, State#state.client_group_valmap_ids)}};

        error ->
            %%?INFO_LOGGER("handle_call(release_bdb_port_handle): From NON-client pid ~p~n", [FromPid]),
            %% This used to be possible before R11B as unlink was asynchronous, so if a client called release_bdb_port_handle
            %% correctly (causing the above unlink) but then exited, we might still get an exit signal.
            %% With R11B unlink is still asynchronous but all subsequent exit events from that process to this one are discarded.
            %% So this error case should really only happen if the caller is misbehaving.  
            %% So we return an error to throw in the caller.
            {reply, 
             {error, caller_does_not_have_a_bdb_port_handle},
             State#state{client_group_valmap_ids = dict:erase(FromPid, State#state.client_group_valmap_ids)}}
    end.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State}.


%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%
%% - EXIT signals from any registered clients
%% - All port replies when we are in pipe-mode
%% - All port replies to non_block calles from this process 
%%   (e.g. group_valmap_cleanup_non_block)
%%--------------------------------------------------------------------
handle_info({ErlangPort, {data, Bytes}}, State) when ErlangPort =:= ?get_erlang_port(State) -> 

    %% This is a raw message from the pipe-driver. 
    %% We convert it to a term and forward it to the client process that sent the corresponding command.
    %%
    %% [Related sidebar: This server process itself occasionally calls blocking functions in the 
    %% ?BDB module (e.g. group_valmap_cleanup).  
    %% All such functions implicitly call ?BDB:receive_reply().
    %% While waiting for it's specific tagged reply, receive_reply() is 
    %% forced to accept all pipe-driver byte messages (to examine the tags).
    %% Its behaviour is to convert the message to term form and
    %%  - if ther tag matches, return the message
    %%  - if the tag does NOT match then extracts the ClientPid from the tag 
    %%    and forwards the message to that process.  Then recurse to wait for 
    %%    the requested message.
    %% i.e. ?BDB:receive_reply() implicitly participates in the 
    %% pipe-driver forwarding mechanism on behalf of this (or any other) server 
    %% that handles EDTK port messages in pipe-driver mode.
    %% As pipe replies are arriving all the time, this is certain to happen 
    %% eventually when we are blocking for the reply to a group_valmap_cleanup.
    %% Alternatively, servers would have to completely avoid the blocking ?BDB functions 
    %% (so avoiding ever calling receive_reply even indirectly).
    %% i.e. Servers would be restricted to using the non-blocking commands,
    %% and would have to keep the tags of all outstanding messages in 
    %% State, and handle them as they arrive asynchronously. (Much more complex.)

    {noreply,
     case ?BDB:pipe_driver_bytes_to_term(ErlangPort, Bytes) of
         {ErlangPort, TagBin, ok}         = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, ok, {_}}    = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, ok, {external_term, BinTerm}} -> 
                                                     forward_pipedriver_reply_to_client(TagBin, {ErlangPort, TagBin, ok, binary_to_term(BinTerm)}, State);
         {ErlangPort, TagBin, ok, _}      = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, error, {_}} = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, error, _}   = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);

         %% Untagged events may be generated by the driver (e.g. BDB event notifications).
         %% They should be sent to the port's connected process -- i.e. self().
         %% The handler either exits or returns a new State.
         {ErlangPort, bdb_untagged_event, Event} ->  handle_untagged_event(Event, State);

         %% 'untagged' errors can happen in pipe-mode too (see below)
         {ErlangPort, bdb_untagged_error, _} = Msg -> exit({bdb_port_error, {State#state.bdb_port_handle, Msg}});

         %% Anything else is a serious error (e.g. malformed messages).
         Other -> exit({bdb_bad_pipe_driver_reply, {State#state.bdb_port_handle, Other}})
     end};

%% Note: the is_pid() test is in the following clause is important, as in pipe-mode we might also receive exit messages from the port (that is handled later).
handle_info({'EXIT', FromPid, Reason}, State) when is_pid(FromPid) ->

    %% We use dict:find instead of dict:fetch as a badly behaving client 
    %% should not be able to kill the server.
    case dict:find(FromPid, State#state.client_group_valmap_ids) of

        {ok, GroupValmapId} -> 

            %% A registered client crashed, so we need to call group_valmap_cleanup for it.
            %%
            %% Note: ccurrently we call it in *blocking* mode.  That may need to 
            %% abort transactions, which involves disk IO and possibly replication.
            %% So this may block for siginificant time (seconds).
            %% So if a LOT of registered clients start crashing then port-coordinator 
            %% could be quite unresponsive while it processes all of the group-valmap-cleanup 
            %% operations.  
            %% Instead we could either spawn a process to call group-valmap-cleanup,
            %% or call the non_block variang and remember the tag.  The former 
            %% would require tracking all of the spawned pids in State, so that we can
            %% wait for them to finish during clean shutdown (that's a safety issue 
            %% -- not optional).  The latter (using the non_block variant) would 
            %% require storing the tags in State and doing handle_info for the 
            %% replies. And in pipe-mode, when we receive all replies, that is more 
            %% work (check every message's tag against the list of outstanding 
            %% group-valmap cleanups).
            %%
            %% Anyway, it remains to be proven that we need this.  Indeed if 
            %% large numbers of client processes are crashing then throttling the 
            %% rate at which port-coordinator issues new port handles is probably 
            %% a good thing.  My main concern is lack of responsiveness to shutdown.
            %% But remember:
            %%  
            %% Important #2: normal successful client shutdown (i.e. calls to 
            %% release_bdb_port_handle) does NOT cause this problem as we 
            %% arrange for the client process to call group_valamap_cleanup,
            %% not the server (the server merely un-registers the client).
            %%  
            %% Important #2: Also, most applications will be using ?BDB_H:do_txn
            %% or ?BDB_RG:do_txn_on_master/any_site in order to get automatic 
            %% retry on deadlock.  Those functions catch any crashes and 
            %% do 'clean' shutdown (release_bdb_port_handle) on the client side.
            %%  
            %% So in most applications the port-coordiantor should rarely see 
            %% an EXIT signal from  registered client, which is why I have not 
            %% optimized for that case.

            NumResources = ?BDB:group_valmap_cleanup(State#state.bdb_port_handle, GroupValmapId),

            ?INFO_LOGGER("handle_info: Received EXIT signal ~p from client process ~p. Cleaning group_valmap_id ~p (~p resources)~n", [Reason, FromPid, GroupValmapId, NumResources]),
            {noreply, State#state{client_group_valmap_ids = dict:erase(FromPid, State#state.client_group_valmap_ids)}};

        error ->
            %% Is it one of our support processes?
            case lists:keysearch(FromPid, #support_proc.pid, State#state.support_procs) of 

                {value, #support_proc{name = Name}} ->
                    %% If a support process dies we assume that the error must be serious, so we die too.
                    ?ERROR_LOGGER("handle_info: Received EXIT signal from support process ~p ~p (reason ~p)~n", [Name, FromPid, Reason]),
                    exit({support_process_exited, {Name, Reason}});

                false ->
                    %% We don't recognize the source of this exit signal.
                    %% It is possible that a client calls release_bdb_port_handle() and 
                    %% then exits before the response is returned.
                    %% Meanwhile the server removes the client from State#state.client_group_valmap_ids
                    %% so when the EXIT signal arrives, we reach this code-path.
                    %% We don't need to do anything.
                    ?INFO_LOGGER("handle_info: Received unexpected EXIT signal from non-client process ~p (reason ~p)~n", [FromPid, Reason]),
                    {noreply, State}
            end
    end;

handle_info({ErlangPort, bdb_untagged_event, Event}, State) when ErlangPort =:= ?get_erlang_port(State) -> 
    {noreply, handle_untagged_event(Event, State)};

handle_info({'EXIT', _ErlangPort, _} = Msg, State) when _ErlangPort =:= ?get_erlang_port(State) -> 
    %% From docs for erlang:open_port : 
    %% - During use of a port opened using {spawn, Name}, errors arising when sending messages to it are reported to the owning process using signals of the form {'EXIT', Port, PosixCode}
    ?ERROR_LOGGER("handle_info: Exiting due to ~p~n", [Msg]),
    exit({bdb_port_error, {State#state.bdb_port_handle, Msg}});

handle_info({_ErlangPort, _} = Msg, State) when _ErlangPort =:= ?get_erlang_port(State) -> 
    %% From docs for erlang:open_port : 
    %% - if the 'exit_status' option is used then:  when the external process connected to the port exits, a message of the form {Port,{exit_status,Status}} is sent to the connected process, where Status is the exit status of the external process. If the program aborts, on Unix the same convention is used as the shells do (i.e., 128+signal).
    %% - if the 'eof' option is used then:          the port will not be closed at the end of the file and produce an exit signal. Instead, it will remain open and a {Port, eof} message will be sent to the process holding the port.
    ?ERROR_LOGGER("handle_info: Exiting due to ~p~n", [Msg]),
    exit({bdb_port_error, {State#state.bdb_port_handle, Msg}});

handle_info(Unrecognized, _State) ->
    ?ERROR_LOGGER("handle_info: Unrecognized message ~p~n", [Unrecognized]),
    exit({unrecognized_message, Unrecognized}).


%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(Reason, State) ->

    ?INFO_LOGGER("terminate: (reason: ~p)~n", [Reason]),

    kill_all_support_procs_and_clients_and_close_environment(State),

    ?INFO_LOGGER("terminate: closing port~n"),
    normal = ?BDB:shutdown(State#state.bdb_port_handle),

    %% IMPORTANT: in pipe-mode shutdown is apparently asynchronous with pipe-main.
    %% That's why we use DB_REGISTER in pipe-mode (the port_coordiantor may restart
    %% while the previous pipe-main is still closing the environment).

    ?INFO_LOGGER("terminate: success~n"),
    ok.


%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%% Returns a list of incompaible config entries 
%%     [{Key, MismatchedExistingValue, NewValue}, ...]
%%  or [] if the lists are compatible.
%%
%% Policy: the new list must be a subset of the existing list
%% i.e. all config entries in the new list must be present in the existing list 
%% and must be identical, but the existing list may have additional 
%% config entries that are not present in the new list
find_mismatched_db_config_entries(NewList, ExistingList) ->

    lists:foldl(fun({Key, NewValue}, AccMismatchList)  ->
                        case proplists:lookup(Key, ExistingList) of
                            {Key, NewValue}                ->  AccMismatchList;     % ok
                            {Key, MismatchedExistingValue} ->  [{Key, MismatchedExistingValue, NewValue} 
                                                                | AccMismatchList]; % mismatch
                            none                           ->  AccMismatchList      % ok
                        end
                end,
                [],       % initial AccMismatchList
                NewList).


%% This is called if we haven't yet opened a database with name Filename.
%% The database may or may not already exist on disk.
%% If it doesn't exist, we create it.
%% Returns a valid gen_server:handle_call return value : {reply, Reply, State}
open_unknown_database(RequestedDBType, 
                      {Filename, DatabaseNameWithinFile} = DBName, 
                      NewDBConfigProplist, 
                      FromPid, 
                      State) ->

    %% As DB handles are shared across all clients, we call db_create with the server's GroupValmapId
    %% (which is a hidden argument in State#state.bdb_port_handle)
    try ?BDB:db_create(State#state.bdb_port_handle, []) of

        DB -> 
            %% Do any optional configuring of the database BEFORE we open it.

            %% Calling the BDB configure functions must be done in a try-catch, to avoid 
            %% killing the server if the client passed bad config values.
            try
                %%?INFO_LOGGER("handle_call(get_db_handle): Configuring DB (~p) for client pid ~p (config: ~p)~n", [DBName, FromPid, NewDBConfigProplist]),
                lists:foreach(fun(ConfigArg) -> handle_db_config_option(State#state.bdb_port_handle, DB, ConfigArg) end,
                              NewDBConfigProplist),

                %% If OptionTxn is non-void it belongs to the **client** process. That's OK.
                %%
                %% If the caller provided a bad filename, or a filename that exists but is the wrong DBType
                %% then this will fail and we don't want to kill the server.
                %% (Note: DB_READ_UNCOMMITTED simply *allows* dirty-reads later; it does not mean that all reads are dirty).

                try ?BDB:db_open(State#state.bdb_port_handle, 
                                 DB, 
                                 void, 
                                 Filename, 
                                 DatabaseNameWithinFile,
                                 RequestedDBType, 
                                 [bdb_DB_CREATE, bdb_DB_READ_UNCOMMITTED, bdb_DB_AUTO_COMMIT],
                                 get(file_permissions)) of
                    ok ->
                        %%?INFO_LOGGER("handle_call(get_db_handle): Successfully opened database ~p (new DB handle ~p, type=~p, config=~p) for client pid ~p ~n", [DBName, DB, RequestedDBType, NewDBConfigProplist, FromPid]),
                        {reply, 
                         {ok, DB}, 
                         State#state{db_handle_cache = dict:store(DBName, 
                                                                  {DB, RequestedDBType, NewDBConfigProplist},
                                                                  State#state.db_handle_cache)}}
                catch
                    {bdb_error, Reason} = ErrorReply ->
                        ?ERROR_LOGGER("handle_call(get_db_handle): Failed (~p) to open DB (~p) for client pid ~p ~n", [Reason, DBName, FromPid]),
                        ok = ?BDB:db_close(State#state.bdb_port_handle, DB, []),
                        {reply, {error, ErrorReply}, State}
                end

            catch
                throw:Exception -> 
                    ?ERROR_LOGGER("handle_call(get_db_handle): Exception ~p configuring or opening database (~p) for client pid ~p (db config: ~p)~n", [Exception, DBName, FromPid, NewDBConfigProplist]),
                    ok = ?BDB:db_close(State#state.bdb_port_handle, DB, []),
                    {reply, {error, {bad_db_config, Exception, NewDBConfigProplist}}, State};

                error:Exception -> 
                    ?ERROR_LOGGER("handle_call(get_db_handle): Error ~p configuring or opening database (~p) for client pid ~p (db config: ~p)~n", [Exception, DBName, FromPid, NewDBConfigProplist]),
                    ok = ?BDB:db_close(State#state.bdb_port_handle, DB, []),
                    {reply, {error, {bad_db_config, Exception, NewDBConfigProplist}}, State}
            end

    catch
        {bdb_error, _} = Error -> 
            ?ERROR_LOGGER("handle_call(get_db_handle): db_create failed ~p~n", [Error]),
            {reply, {error, {db_create_failed, Error}}, State}
    end.


handle_db_config_option(Port, DB, {page_size, NumBytes}) when is_integer(NumBytes), NumBytes > 0 ->
    ok = ?BDB:db_set_pagesize(Port, DB, NumBytes);

handle_db_config_option(Port, DB, {set_flags, Flags}) when is_list(Flags) ->
    ok = ?BDB:db_set_flags(Port, DB, Flags);

handle_db_config_option(Port, DB, {set_btree_min_keys_per_page, MinKey}) when is_integer(MinKey), MinKey > 0 ->
    ok = ?BDB:db_set_bt_minkey(Port, DB, MinKey);

handle_db_config_option(Port, DB, {set_fixed_record_size_bytes, NumBytes}) when is_integer(NumBytes), NumBytes > 0 ->
    ok = ?BDB:db_set_re_len(Port, DB, NumBytes);

handle_db_config_option(Port, DB, {set_fixed_record_pad_byte, ByteValue}) when is_integer(ByteValue), ByteValue >= 0, ByteValue =< 16#FF ->
    ok = ?BDB:db_set_re_pad(Port, DB, ByteValue);

handle_db_config_option(Port, DB, {set_queue_extent_size_db_pages, NumDBPages}) when is_integer(NumDBPages), NumDBPages > 0 ->
    ok = ?BDB:db_set_q_extentsize(Port, DB, NumDBPages);

handle_db_config_option(_, _, DbConfigOption) ->
    throw({unrecognized_db_config_option, DbConfigOption}).


%%
%% Helper to cleanly close environment.  Will throw or exit if it has problems.
%%

kill_all_support_procs_and_clients_and_close_environment(State) ->

    %%1 = ?BDB:debug(State#state.bdb_port_handle, 1),

    %% Enable automatic deadlock detection as we are about to shutdown 
    %% support processes (inc. the periodic deadlock detector if it 
    %% is running) and client processes are still running at this point.
    %% We don't want BDB to deadlock.
    %% We don't test the result, as if the auto-deadlock detector 
    %% is running, this may fail.
    ?BDB:set_lk_detect(State#state.bdb_port_handle, bdb_DB_LOCK_DEFAULT),

    %% First, tell all of the support processes to stop.  
    %% We don't use exit(,kill) as the process may be executing a BDB command 
    %% and we want the BDB command to finish before it exits (rather than
    %% killing the process while is waiting for the port reply) so
    %% that the exit message really does mean that no BDB commands are
    %% being executed that were started by that process.  
    %% This happens to not be totally essential at the time of writing, as the support
    %% processes only use the DBENV handle (so it will have an incremented 
    %% refcount if a process is still using it).  It will become important 
    %% if any support process in the future opens a transaction or cursor 
    %% (e.g. a de-fragmenter using DB->compact), so we implement it correctly now.
    %%
    %% As with all terminate() functions, we cannot rely on EXIT signals for 
    %% support procs as we may have already eaten such a message (it might 
    %% be why we are terminating). So we need to monitor them.

    SupportProcMonitorRefs
        = lists:map(fun(SupportProc) ->
                            %%?INFO_LOGGER("close_environment: stopping ~p (pid ~p)~n", [SupportProc#support_proc.name, SupportProc#support_proc.pid]),
                            MonitorRef = erlang:monitor(process, SupportProc#support_proc.pid),

                            StopFun = SupportProc#support_proc.stop_fun,
                            StopFun(SupportProc),

                            {SupportProc#support_proc.name, MonitorRef}
                    end,
                    State#state.support_procs),

    %% Theoretically, we don't actually need to do anything, as when the connected 
    %% process dies, the port will close and will do valmap cleanup.
    %% (If the gen_server crashes then we rely on all that auto-cleanup).
    %%
    %% ... the simplest possible explicit cleanup (not actually necessary)...
    %%{ok, normal} = ?BDB:shutdown(State#state.bdb_port_handle).

    %% However, as port shutdown (in linked-driver mode) blocks the 
    %% entire Erlang VM until it is finished, it is actually beneficial 
    %% to to do the cleanup steps here, as it has less impact on the VM's scheduler.
    %% TODO: there is a race condition here in pipe-mode -- replies could 
    %% still arrive after this point for commands that were sent earlier
    %% (pipe-main does not know that we are terminating).
    
    %% First kill all of our registered client processes, and 
    %% run group_valmap_cleanup on each of their valmap group ids.
    %% We want to do this in parallel to minimize shutdown down, 
    %% so we use the non_block variant of group_valmap_cleanup.
    %% (That is less painful than spawning a cleanup process for 
    %% each one, as in pipe mode the replies all still come to this 
    %% server process).

    CleanGroupValmapTags = 
        dict:fold(fun(ClientPid, ClientGroupValmapId, CleanGroupValmapTags) ->

                          %% We don't need the {'EXIT',ClientPid,killed} messages from the clients we are about to kill.
                          unlink(ClientPid),

                          %%?INFO_LOGGER("close_environment: killing client pid ~p and cleaning group_valmap_id ~p~n", [ClientPid, ClientGroupValmapId]),
                          exit(ClientPid, kill),

                          [{ClientPid, ClientGroupValmapId, ?BDB:group_valmap_cleanup_non_block(State#state.bdb_port_handle, ClientGroupValmapId)}
                           | CleanGroupValmapTags]
                  end,
                  [], 
                  State#state.client_group_valmap_ids),

    %% Wait for all of the support processes to exit
    lists:foreach(fun({ProcName, MonitorRef}) ->
                          receive
                              {'DOWN', MonitorRef, process, _, normal} ->  % this is the expected value
                                  ok;
                              {'DOWN', MonitorRef, process, _, Reason} -> 
                                  ?WARNING_LOGGER("close_environment: support proc ~p exited with an error: ~p~n", [ProcName, Reason])
                          end
                  end,
                  SupportProcMonitorRefs),

    %% We cannot close Db handles until all cursors and transactions are closed.
    %% So we *must* block until all of the above parallel operations finish, 
    %%
    %% If any of the group_valmap_cleanup operations fail then we 
    %% *cannot* safely close Db handles here, so we just close the port, and let it's 
    %% stop() function impose the correct ordering on all valmaps that are left).
    %% (Note that if we do close the port under those circumstances, other group_valmap_cleanup 
    %% operations may still be in-progress, and cleanup processes may even attempt to start
    %% more of them. That's OK; the port's _stop() function can handle the former, and Erlang 
    %% will just discard any commands sent to a port that is closing.).

    PortHandleWithGroupValmapCleanupTimeout = (State#state.bdb_port_handle)#bdb_port_handle{receive_reply_timeout = ?TERMIMATE_GROUP_VALMAP_CLEANUP_TIMEOUT_MILLISECONDS},
    lists:foreach(fun({ClientPid, ClientGroupValmapId, Tag}) ->
                          ?INFO_LOGGER("close_environment: waiting for response to group_valmap_cleanup_non_block for client ~p, group-id ~p.~n", [ClientPid, ClientGroupValmapId]),
                          NumResources = ?BDB:receive_group_valmap_cleanup_reply(PortHandleWithGroupValmapCleanupTimeout, Tag),
                          ?INFO_LOGGER("close_environment: got reply for group_valmap_cleanup_non_block for client ~p, group-id ~p (NumResources ~p).~n", [ClientPid, ClientGroupValmapId, NumResources])
                  end,
                  CleanGroupValmapTags),

    %% All valmap groups belonging to client processs have been cleaned up.

    %% Note that, if a client had a prepared (unresolved) transaction, then it is *still open*.
    %% (group_valmap_cleanup will have sent us an untagged event about it, but NOT aborted it.)
    %%
    %% As of BDB v4.5, we CANNOT (normally) close DB handles if there are any prepared (unresolved) transactions 
    %% still open.  This is because DB::close logs a record that makes the log unrecoverable if there is an open txn before it.
    %%
    %% Sleepycat have given us a patch to allow this (it prevents DB::close from logging), but currently it must be explicitly 
    %% activated (SR [#14785])
    %% Sleepycat promise to support this officially at some point (perhaps via a different mechanism).

    case proplists:get_value(distributed_transactions, get(config_proplist)) of
        disabled ->
            0;
        _ ->
            ?INFO_LOGGER("close_environment: Checking for prepared transactions~n"),

            %% We have a special BDB driver command for this.
            %% It counts the number of prepared-but-unresolved txns, and if there are any,
            %% it activates the hack provided by Sleepycat (SR [#14785]) that makes it safe 
            %% to close DB handles with open transactions (it disables logging of DB::close).
            %% Sleepycat do plan to support this functionality officially, although the mechanism 
            %% may change (ideally would be automatic -- then this command could be removed).
            %%
            %% The command was written speicifically for the current circumstances.  Therefore 
            %% if it finds any transactions still in use, it returnx a badarg error (as there shouldn't 
            %% be any).
            %%
            %% Note: originally this was implemented by checking to see if we have any 'need to resolve' 
            %% untagged event messages in our message queue -- group_valmap_cleanup on client group ids 
            %% will generate these if the client has an unresolved transaction when it was killed.
            %% However, that's a bit of a pain to do for 2 reasons
            %%   1. In pipe mode, the messages need to be converted first (easy to dobut more code).
            %%   2. It fails to count any unresolved txns that were being processed by the 'incremental resolve'
            %%      support process.
            %% Rather than track all of the unresolved txns in State, it's easier just to call some C code to 
            %% do it, especially as if there _are_ any, we need to call some C code to activate the hack.

            NumUnresolvedTxns = ?BDB:make_db_close_safe_if_there_are_unresolved_transactions(State#state.bdb_port_handle),
            ?INFO_LOGGER("close_environment: There are ~w prepared-but-unresolved transactions ~n", [NumUnresolvedTxns])
    end,

    ?INFO_LOGGER("close_environment: Closing DB handles~n"),

    %% Close all Db handles. 
    %%
    %% Important: we loop over the db_handle_cache and close each handle explicitly rather than doing group_valmap_cleanup 
    %% on the server's group-id, for 2 reasons
    %%
    %% 1.  If group_valmap_cleanup attempts to close a DB handle and finds that there is a prepared-but-unresolved txn open, 
    %%     it HARD-ABORTS the OS process.  group_valmap_cleanup must assume that the process will continue to run -- that 
    %%     clients may attempt to abort or commit the prepared txns, so aborting the process is the only safe course.
    %%
    %% 2.  Closing DB handles is potentially expensive, and while group_valmap_cleanup would not block the Erlang VM 
    %%     (as it uses an async thread), it closes each handle sequentially.  Here we can do them in parallel.
    
    DbCloseTags = dict:fold(fun(DBName, {DB, _ExistingDBType, _ExistingDBConfigProplist}, DbCloseTags) ->
                                    %%?INFO_LOGGER("close_environment: closing DB ~p~n", [DBName]),
                                    [{DBName, ?BDB:db_close_non_block(State#state.bdb_port_handle, DB, [])} | DbCloseTags]
                            end,
                            [], 
                            State#state.db_handle_cache),

    PortHandleWithDbCloseTimeout = (State#state.bdb_port_handle)#bdb_port_handle{receive_reply_timeout = ?TERMIMATE_DB_CLOSE_TIMEOUT_MILLISECONDS},
    lists:foreach(fun({_DBName, Tag}) ->
                          %%?INFO_LOGGER("close_environment: waiting for response to db_close_non_block for database ~p~n", [Filename]),
                          ok = ?BDB:receive_db_close_reply(PortHandleWithDbCloseTimeout, Tag)
                          %%?INFO_LOGGER("close_environment: got reply for db_close_non_block for ~p~n", [DBName]),
                  end,
                  DbCloseTags),

    %% Close the DbEnv handle.  It should be the only thing still open.
    %% It is safe to do this even if NumUnresolvedTxns /= 0.

    NumServerResources = ?BDB:group_valmap_cleanup(State#state.bdb_port_handle, ?get_server_group_valmap_id(State)),
    ?INFO_LOGGER("close_environment: cleaned server group valmap id ~p (NumResources ~p)~n", [?get_server_group_valmap_id(State), NumServerResources]),

    State#state{db_handle_cache         = dict:new(), 
                next_group_valmap_id    = ?get_server_group_valmap_id(State) + 1,   % will be 1
                client_group_valmap_ids = dict:new(), 
                support_procs           = []}. 


%%
%% Helpers for handle_info
%%

forward_pipedriver_reply_to_client(TagBin, Reply, State) ->
    %% (Remember, these messages may ALSO be forwarded by ?BDB:receive_reply)
    case binary_to_term(TagBin) of 
        {_Ref, ClientPid} when is_reference(_Ref), is_pid(ClientPid) -> 
            ClientPid ! Reply,
            State
    end.


handle_untagged_event({need_to_resolve_prepared_txn, DistributedTxnGUID, Txn}, State) -> 
    %% Pass it to the incremenatl-resolver process.
    {value, #support_proc{pid = ResolverPid}} = lists:keysearch(incrementally_resolve_prepared_txns, #support_proc.name, State#state.support_procs),
    ResolverPid ! {need_to_resolve, DistributedTxnGUID, Txn},
    State;

handle_untagged_event({sleepycat_event, bdb_DB_EVENT_PANIC}, State) -> 
    %% We hard-kill all of our clients (incase they might be trapping exits), and then exit ourselves.
    %% We don't publish this event via replication_gen_event, as exiting will kill the gen_event process anyway.
    ?ERROR_LOGGER("handle_untagged_event: Got DB_EVENT_PANIC : hard-killing all clients and then exiting~n"),
        dict:fold(fun(ClientPid, _, _) ->
                          ?ERROR_LOGGER("PANIC: killing client pid ~p~n", [ClientPid]),
                          exit(ClientPid, kill)
                  end,
                  [], 
                  State#state.client_group_valmap_ids),
    exit(bdb_event_panic);

handle_untagged_event({sleepycat_event, bdb_DB_EVENT_REP_CLIENT}, State) -> 
    %% This site just transitiond to become a replica.  Either
    %%   - Started replication (as a replica)
    %%   - Downgraded from Master
    ?INFO_LOGGER("handle_untagged_event: Got DB_EVENT_REP_CLIENT~n"),

    %% Publish event to all interested parties
    gen_event:notify(get_replication_gen_event_pid(State), 
                     {?get_proc_name(State), bdb_DB_EVENT_REP_CLIENT}),
    
    %% If we downgraded from Master, we might want to kill all of our clients. 
    %% But the ones just doing reads would have been OK, and the ones doing writes 
    %% will get errors from db_put etc. anyway. So we don't do anything
    State;

handle_untagged_event({sleepycat_event, bdb_DB_EVENT_REP_MASTER}, State) -> 
    %% "The replication group of which this site is a member has just elected the local site as its master."
    ?INFO_LOGGER("handle_untagged_event: Got DB_EVENT_REP_MASTER (we have been elected master)~n"),

    case proplists:get_value(distributed_transactions, get(config_proplist)) of
        disabled ->
            ok;
        _ ->
            %% We may need to run txn_recover here.  However, clients might be trying to start new transactions.
            %% BDB does catch that and returns errors for txn_begin.   (I've requested that Sleepcat change this 
            %% and allow txn_begin in this circumstance -- to improve write-availablity of the master.
            %% txn_begin is safe here as the new master will still have the locks owned by the unresolved txn, 
            %% -- unlike after normal (startup-time) recovery (when there are unresolved transactions discovered 
            %% by txn_recover), as normal recovery does not acquire locks (it is single-threaded).
            ?INFO_LOGGER("handle_untagged_event: Calling do_distributed_transaction_recovery after DB_EVENT_REP_MASTER~n"),
            do_distributed_transaction_recovery(State#state.bdb_port_handle, bdb_DB_FIRST)
    end,

    %% Send notification to other parties AFTER txn_recover is finished.
    gen_event:notify(get_replication_gen_event_pid(State), 
                     {?get_proc_name(State), bdb_DB_EVENT_REP_MASTER}),

    State;

handle_untagged_event({sleepycat_event, bdb_DB_EVENT_REP_NEWMASTER}, State) -> 
    %% "The replication group of which this site is a member has just elected a new master; the local site is not the new master."
    ?INFO_LOGGER("handle_untagged_event: Got DB_EVENT_REP_NEWMASTER (we are NOT the master)~n"),

    gen_event:notify(get_replication_gen_event_pid(State), 
                     {?get_proc_name(State), bdb_DB_EVENT_REP_NEWMASTER}),
    State;

handle_untagged_event({sleepycat_event, bdb_DB_EVENT_REP_STARTUPDONE}, State) -> 
    %% This site is a replica and has finished sync-up (is now receiving and acking live updates from the master)
    ?INFO_LOGGER("handle_untagged_event: Got DB_EVENT_REP_STARTUPDONE~n"),

    %% Publish event to all interested parties
    gen_event:notify(get_replication_gen_event_pid(State), 
                     {?get_proc_name(State), bdb_DB_EVENT_REP_STARTUPDONE}),
    State;

handle_untagged_event({sleepycat_event, bdb_DB_EVENT_WRITE_FAILED}, State) -> 
    ?INFO_LOGGER("handle_untagged_event: Got DB_EVENT_WRITE_FAILED (failed to write to disk)~n"),

    gen_event:notify(get_replication_gen_event_pid(State), 
                     {?get_proc_name(State), bdb_DB_EVENT_WRITE_FAILED}),
    State.



%% More general helper functions

dynamic_replication_gen_event_proc_name() ->
    {registered_name, PortCoordinatorProcessName} = process_info(self(), registered_name),
    list_to_atom(atom_to_list(PortCoordinatorProcessName) ++ ?REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX).

get_port_coordinator_proc_name() ->
    {registered_name, Pid} = process_info(self(), registered_name),
    Pid.

get_replication_gen_event_pid(State) ->
    {value, #support_proc{pid = Pid}} = lists:keysearch(replication_gen_event, #support_proc.name, State#state.support_procs),
    Pid.


%% Helper function for distributed transaction support.
%% TxnRecoverFlags should be bdb_DB_FIRST when called from outside.
%% Returns nothing useful. Exits or throws if there is a problem.
%%
%% If replication is disabled this is called after env_open.
%% If replication is enabled this is called when DB_EVENT_REP_MASTER is received.
do_distributed_transaction_recovery(BdbPortHandle, TxnRecoverFlags) ->
    
    case ?BDB:txn_recover(BdbPortHandle, TxnRecoverFlags) of
        [] ->
            ?INFO_LOGGER("do_distributed_transaction_recovery: Finished.~n"),
            ok;

        ListDistributedTxnGUIDsTxns ->
            ?INFO_LOGGER("do_distributed_transaction_recovery: Need to resolve ~p~n", [ListDistributedTxnGUIDsTxns]),

            %% Ask GTM for decision for all DistTxnGUDS
            ListTxnResolutions
                = berkeley_db_global_transaction_manager:resolve(
                    lists:map(fun({DistributedTxnGUID, Txn}) when is_binary(DistributedTxnGUID), size(DistributedTxnGUID) == ?DB_XIDDATASIZE, is_record(Txn, bdb_db_txn) -> 
                                      DistributedTxnGUID;
                                 (_) ->
                                      %% If txn_recover could not allocate memory it puts an atom in the list
                                      exit({txn_recover_returned_malformed_results, ListDistributedTxnGUIDsTxns})
                              end, 
                              ListDistributedTxnGUIDsTxns)),

            ?INFO_LOGGER("do_distributed_transaction_recovery: Got resolutions from GTM ~p~n", [ListTxnResolutions]),

            %% Paranoia: check that the GTM returned results for ALL of the transactions that we asked it for.
            %% It is sufficient to check that length(ListTxnResolutions) == length(ListDistributedTxnGUIDsTxns).
            %% If it is shorter, some txns are missing.  If it is longer then either there are multiple entries for 
            %% some transactions (which would have been caught when they are aborted or committed twice) or 
            %% there are spurious entries that we did not ask for, which would have been caught in find_txn_handle().
            if length(ListTxnResolutions) /= length(ListDistributedTxnGUIDsTxns) ->
                    exit({global_transaction_manager_resolve_results_are_wrong_length, ListTxnResolutions, ListDistributedTxnGUIDsTxns});
               true -> ok
            end,

            FindTxnHandle = fun(DistributedTxnGUID) ->
                                    case lists:keysearch(DistributedTxnGUID, 1, ListDistributedTxnGUIDsTxns) of
                                        {value, {DistributedTxnGUID, TxnHandle}} ->
                                            TxnHandle;
                                        false ->
                                            exit({global_transaction_manager_resolve_results_are_missing_an_entry, ListTxnResolutions, ListDistributedTxnGUIDsTxns})
                                    end
                            end,

            %% Loop over GTM results and call txn_commit or txn_abort on each txn
            %% (Paranoia: if the GTM has a bug that causes it to return multiple entries for the same 
            %% transaction then commit or abort will catch it, as the generation-id of the bdb_db_txn won't 
            %% be valid the second time).
            lists:foreach(
              fun({ok, {<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID, committed}}) -> 
                      ?INFO_LOGGER("do_distributed_transaction_recovery: GTM says to commit dist. txn ~w.~n", [TxnCounter]),
                      ok = ?BDB:txn_commit(BdbPortHandle, FindTxnHandle(DistributedTxnGUID), []);

                 ({ok, {<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID, aborted}}) -> 
                      ?INFO_LOGGER("do_distributed_transaction_recovery: GTM says to abort dist. txn ~w.~n", [TxnCounter]),
                      ok = ?BDB:txn_abort(BdbPortHandle, FindTxnHandle(DistributedTxnGUID));

                 ({error, {<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>>, _}} = Error) -> 
                      ?ERROR_LOGGER("do_distributed_transaction_recovery: GTM:resolve returned error ~p for dist. txn ~w~n", [Error, TxnCounter]),
                      exit({global_transaction_manager_resolve_returned_error, Error})
                 end,
              ListTxnResolutions),

            %% Repeat until txn_recover returns an empty list
            do_distributed_transaction_recovery(BdbPortHandle, bdb_DB_NEXT)
    end.


logger(String) ->
    logger(String, []).
logger(String, Args) ->
    case process_info(self(), registered_name) of
        {registered_name, ProcessName} ->
            io:format("~p: " ++ String, [ProcessName | Args]);
        [] ->
            %% This requires that we only create support processes with proc_lib:spawn_link
            io:format("~p: support process ~p: " ++ String, [hd(get('$ancestors')) | [self() | Args]])
    end.


