%%%-------------------------------------------------------------------
%%% File    : berkeley_db_replication_group.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%%   See also berkeley_db_replication_group_helpers.erl (for functions
%%%   that support redirects and retries when attempting to talking to 
%%%   a replication group when one or more sites might be down).
%%%
%%%
%%%   This module implements a gen_server that both manages one site
%%%   in a BDB-HA replication group and indirectly constructs and
%%%   maintains an entire replication group.  One of its primary
%%%   duties is to track the location of the master and forward
%%%   transactions (funs) to that site.
%%%
%%%   All replication_group servers (that are part of the same named group) form a 
%%%   'peer supervision' ring.  Each site attempts to find all of the others, and will 
%%%   attempt to remote-spawn any that appear to be down -- with infinite retry attempts
%%%   unless the target site is explicitly removed from the group via an API.
%%%   So starting one site (with appropriate config) is sufficient to start them all.
%%%   Other sites may be added incrementally via an API.  So it is possible to start 
%%%   a single site that does not know about any others, and add others over time.
%%%
%%%   There is a convenience API called start_sites that will start (in parallel) an entire 
%%%   group of replication_group servers at once, but it does NOT link them all to the caller,
%%%   This is because gen_server:start_link is hardwired to termiante if an EXIT message 
%%%   is received from the process that called start_link.
%%%   Typically we do NOT want to kill the entire replication group if the single process 
%%%   that called caller crashes.  Instead the caller should use erlang:monitor on the 
%%%   pids returned by start_sites.  The caller MUST also remember that if a replication_group 
%%%   server dies then the other sites will attempt to re-spawn it.  So if the caller gets 
%%%   a 'DOWN' message it will need to start looking for a new registered process being 
%%%   created, and then monitor that new process.)
%%%   If some application does want the death of a single admin process to cause a 
%%%   replication_group to be torn down entirely, then a second admin process should monitor 
%%%   the first and issue 'group_shutdown' commands to all sites in the replication group.
%%%   That will be the ONLY way to shutdown all of the sites, other than all of the Erlang 
%%%   nodes failing simultaneously. 
%%%
%%%   Note that start_sites is only a convenience.  It is perfectly fine for an application's init
%%%   function to to start an individual replication_group server, with appropriate config to spawn 
%%%   the others. N.B. Again, if the local application instance dies then the replication_group server 
%%%   will continue to run (and be restarted) unless the entire group is shutdown or that 
%%%   particular site is explicitly removed from the group via an API call.   
%%%   So the model is significantly different from a pure OTP hierarchy.
%%%
%%%
%%%   Each replication_group server spawns and monitors a local port_coordinator server,
%%%   and restarts it if it dies (i.e. acts as a supervisor).
%%%
%%%   Currently, if a replication_group server dies then its port_coordinator is killed 
%%%   (i.e. they are linked).
%%%   It isn't completely obvious that this is the right thing to do.  Here is the rationale for the choice:
%%%
%%%   1. If there is a bug in replication_group server (e.g. inconsistent tracking of master) that kills the 
%%%      replication_group server then we DON'T want the port_coordinator to die if the replication_group server dies, 
%%%      because we want BDB-HA replication to continue while replication_group server restarts.
%%%
%%%   2. If the replication_group server detects a problem with the port_coordinator it can 
%%%      explicitly terminate it to restart it.
%%%       
%%%   3. On the other hand, if there is a bug in the port coordinator because if the the replication_group server 
%%%      cannot handle (and therefore dies) then the replication_server will thrash -- die/spawn/die/spawn
%%%      unless something also kills the port_coordinator.
%%%
%%%   On balance the danger of #3 outweighs the minor availability issue of #1, because the port_coordinator will be 
%%%   restarted as soon as the replication_group server is restarted by a peer, which should happen almost immediately 
%%%   as every other site in the group will attempt to restart it.
%%%       
%%%
%%%   The replication_group server subscribes to the the local port_coordinator replication_gen_event, to track the 
%%%   BDB-HA status of this site (master/{replica,not_in_sync}/{replica,in-sync} etc).
%%%
%%%   When the status of its site changes, the replication_group publishes the new status to every other 
%%%   replication_group that it knows about.   
%%%
%%%   When a replication_group server discovers another site, it subscribes to that other site's stream of status updates 
%%%   and asks for that other site's current status.
%%%
%%%
%%%   This whole thing is therefore a simple form of a classical group communication system.
%%%
%%%   There are many intentional departures from the classical design.  These will be addressed over time, possibly
%%%   by a separate group communication subsystem that will also take the place of Distributed Erlang.
%%%
%%%   - No formal concept of epochs (periods of frozen group membership in which we pretend there are no failures).
%%%     Site's come and go at will.  
%%%
%%%   - When a site learns of another site it does performs unidirectional (polling) 'state transfer' (asks for 
%%%     the other site's full state) as well simultaneously subscribing to all future state changes.
%%%     Sites are expected to discover each other by polling.  Sites do not currently tell all other existing sites about 
%%%     the new site that they just located.  (This will probably need to change to allow sites to be added 
%%%     dynamically).
%%%
%%%   - No causal order tracking of messages (vector clocks) or causal-order delivery of messages (i.e. no buffering of 
%%%     received messages until all those that causally preceed it have been received and acted upon)
%%%
%%%     Currently all status-update messages are timestamped with erlang:now() from their originating host, to allow 
%%%     ordering messages from one site (but not across all sites).
%%%
%%%     Also, all status-updates updates are *currently* absolute (idempotent).  
%%%     This simplifies things. e.g. There is a race condition in the subscribe_and_get_status operation
%%%     we have to subscribe before returning the status or we might miss events.  But then some events might 
%%%     be sent that preceed the status value that is returned by the function. 
%%%     Because we have idempotent events we can simply discard those that preceed the one obtained by the function.
%%%     (note: those timestamps cannot be meaningfully compared with our own local time, but we don't 
%%%     need to do that, we just need to order the messages from one site).
%%%     If the messages were not idempotent (deltas, not absolute) then we would need to discard 0..N events
%%%     (those that are timestamped before the OtherSiteStatus received from get_site_status) before returning 
%%%     the absolute value we got from subscribe_and_get_status, and then apply any remaining events.
%%%
%%%   - No message retransmit, and no blocking for acks from each destination.
%%%     Instead, we rely on Erlang's guarantee for sending to monitored/linked processes: that either the message is delivered 
%%%     or eventually the sender will receive a DOWN/EXIT message for the destination process.
%%%     See 'medium safe' : http://www.erlang.org/ml-archive/erlang-questions/199910/msg00043.html
%%%
%%%
%%%   The replication_group server accepts requests to run read or write transactions (funs) and
%%%   forwards them to the correct port_coordinator (i.e. write transactions go to the master). 
%%%
%%%   The reply includes an opaque MasterLocationHint which is an opaque object that can be passed to the 
%%%   next call to do_txn_on_master to reduce the likelyhood of an extra hop to find the master again.
%%%
%%%   A function is provided to obtain the Node from a MasterLocationHint, to allow applications to move 
%%%   work to the current master node and so avoid multiple inter-node hops if they intend to call 
%%%   do_txn_on_master multiple times.
%%%
%%%
%%% API
%%%
%%% Function: do_txn_on_master()
%%%
%%% Description: 
%%%
%%%   Forwards a transaction fun to be executed on the master site.
%%%
%%%   This is similar in intent to berkeley_db_helpers:do_txn/3 but with 
%%%   some important differences:
%%%
%%%      - if the current host happens to be the master then the TxnFun will be executed by a spawned process 
%%%        (TODO: we could just execute it in the client process).
%%
%%%      - if the current host is NOT the master then the unless the current host is the master, the TxnFun will be sent to another 
%%
%%%      - the return value is different (contains more information) -- see below.
%%%
%%%   TxnFun must be  
%%%
%%%      {Module, Function, ClientDataTerm}
%%%
%%%   where the following exported function must exist on the master node:
%%%
%%%      Module:Function(BdbPortHandle, ClientDataTerm) -> ResultTerm
%%%
%%%   i.e. TxnFun is (unfortunately) NOT a normal fun (closure) or even a new-style 'fun M:F/A'.
%%%   This is to allow the code on the server to be upgraded independently of the code on the caller.
%%%   Normal Erlang hot-code loading fails here (and normal funs break) because TxnFun will almost 
%%%   certainly contain BDB API constants (converted from #defines in db.h to -defines in berkeley_db.xml
%%%   and berkeley_db_drv.hrl), and those constants can (and usually) do change between BDB releases 
%%%   (even patch releases).  So using the wrong constants could do anything -- including corrupting data.
%%%   Therefore the version of the fun's code used to actually call BDB (the version on the server) must
%%%   be compiled with the new berkeley_db_drv.hrl.  And the fun invocation sent by the client must match 
%%%   that new version.  fun invocations *can* invoke the *old* code (because funs contain an internal 
%%%   module-id, they don't find the module by name) -- but in this case we absolutely don't want that.
%%%   Fortunately upgrading BDB cannot be done via hot-code loading -- the VM must be stopped and restarted 
%%%   which clears out the old code completely.
%%%
%%%   Using the Module:Function form to specify a TxnFun means that the client code can be a different version
%%%   than the version on the server.
%%%
%%%   IMPORTANT: TxnFun must **NOT** call berkeley_db_port_coordinator:release_bdb_port_handle() on BdbPortHandle
%%%   as the server may choose to re-use that handle.   This should be obvious as the TxnFun is passed the handle 
%%%   and does not obtain it itself.
%%%
%%% Returns    
%%%
%%%   {ResultOfTxnFun, MasterLocationHint}
%%%
%%%   Note: MasterLocationHint is an opaque term which may be passed as the first argument of do_txn_on_master/2
%%%   to hopefully avoid an extra hop.  Currently it is a Distributed Erlang pid, for speed.
%%%
%%% or throws one of...
%%%
%%%   {txn_fun_error, Reason}                            (e.g. if TxnFun called throw)
%%%   {txn_fun_error, {bdb_error, bdb_DB_LOCK_DEADLOCK}} (if TxnFun hit BDB deadlocks and exhausted NumDeadlockRetries)
%%%   {txn_fun_error, {return_with_open_resources, 
%%%                    NumResources, 
%%%                    TxnFunResult}}                    (if TxnFun returned with an open cursor/txn)
%%%
%%% or a system error...
%%%
%%%    {bdb_replication_group_error, Reason}             (*** e.g. Reason might be 'no_master' ***)
%%%
%%% or an internal error (these are examples, *not* an exhaustive list)...
%%%
%%%   {bdb_helper_error, {acquire_bdb_port_handle, Reason}}
%%%   {bdb_helper_error, {release_bdb_port_handle, Reason, TxnFunResult}}
%%%   {bdb_helper_error, {group_valmap_cleanup, Reason}}
%%%   {bdb_helper_error, Reason}
%%% 
%%%
%%% NOTE: a TxnFun will be executed in it's own process and can live as long as it wishes.
%%% It can send messages to other processes (via pids bound to closure variables), and wait to receive 
%%% messages from them.  In this way a TxnFun may open a cursor and then send batches of key/values to 
%%% other processes on demand.  Note: a TxnFun should guarantee to terminate.  If it uses receive then 
%%% there should be a non-infinite timeout, or the TxnFun should monitor/link to the process it is expecting 
%%% to receive messages from.  If a TxnFun crashes then any cursors that it opened will be closed and any 
%%% txns it opened will be aborted.
%%%
%%% Example of a simple TxnFun
%%%
%%%         my_txn(BdbPortHandle, {MyDbName, MyKey, MyData}) -> 
%%%
%%%                 %% we almost always need at least one DB handle (which we must NOT close)
%%%                 {ok, DB} = berkeley_db_port_coordinator:get_db_handle(BdbPort, bdb_DB_BTREE, MyDbName),
%%%          
%%%                 %% now do some work and auto-commit it...
%%%                 {ok, 0}  = ?BDB:db_put(BdbPort, DB, void, MyKey, MyData, []),
%%%          
%%%                 %% and/or do a multi-part transaction and explicitly commit it...
%%%                 {ok, Txn} = ?BDB:txn_begin(BdbPort, void, []),
%%%                 {ok, 0}   = ?BDB:db_put(BdbPort, DB, Txn, MyKey, MyData, []),
%%%                 {ok, 0}   = ?BDB:db_put(BdbPort, DB, Txn, <<"some other key">>, MyData, []),
%%%                 {ok, 0}   = ?BDB:txn_commit(BdbPort, Txn, []),
%%%         end
%%% 
%%% 
%%% Function: do_txn_on_any_site()
%%%
%%% Description: 
%%%
%%%   Executes a READ-ONLY transaction
%%%
%%%   Again, this is similar in intent to berkeley_db_helpers:do_txn/3.
%%%
%%%   It executes the TxnFun on a site chosen from the following preference order:
%%%
%%%       - the local site if it is in-sync (even if it happens to be the master)
%%%       - any other in-sync replica (i.e. excluding the master)
%%%       - the master (as it is in-sync)
%%%       - the local site (which is not in-sync)
%%%       - [TODO] if the local site is down, we could pick any other site
%%%
%%%  TODO: the last entries above means that this function does NOT gairnatee to exectute 
%%%  on an in-sync site.  I intend to add a parameter to control this.
%%%
%%% Returns    
%%%
%%%   ResultOfTxnFun
%%%
%%% or throws one of the errors described earlier.
%%%
%%% 
%%% TODO
%%%
%%% - Since this module was written, Sleepycat have changed DbEnv::repmgr_add_remote_site() 
%%%   to return the internal opaque site id of each site.
%%%   Such a site id is passed with the NEWMASTER event to the DbEnv::set_event_notify() handler 
%%%   which generates our {sleepycat_event, bdb_DB_EVENT_REP_NEWMASTER} messages.
%%%   We don't currently use that feature, and that means there is a a serious race condition 
%%%   in the current code -- if NEWMASTER occurs locally **after** we have received 
%%%   the status update from the new master then this site will forget who the master is.
%%%   This is unlikely to occur (as NEWMASTER is local and receiving an update from the new master 
%%%   is remote). But using the opaque site ids fixes this (the propogation of site status from the 
%%%   new master should simply agree with the site_id from the NEWMASTER callback).
%%%   Also, using this feature allows replicas to discover the location of the new master
%%%   a bit faster than waiting for the master site to receive DB_REP_MASTER event and then tell everyone 
%%%   about it.  i.e. We will be able to forwared write transactions (in do_txn_on_master) a bit sooner than we 
%%%   could before.
%%%   To support this the port_coordinator needs to remember the site ids, and convert the opaque site id in 
%%%   the {sleepycat_event, bdb_DB_EVENT_REP_NEWMASTER} message to some kind of meaningful-to-us site id 
%%%   structure (e.g. like BDBHAHostPort) before it forwards the event via gen_event:notify().  
%%%   Then replication_group would convert that to a SiteLocator
%%%   Note that adding this means that we have 2 sources that can claim a different site is master 
%%%   -- messages from that site, and the {sleepycat_event, bdb_DB_EVENT_REP_NEWMASTER} messages.
%%%   We need to decide what to do if these disagree (there is an implicit race condition -- either might 
%%%   arrive first, although it is very likely that the sleepycat_event will get here first).
%%%
%%% - Fix shutdown_group API to make it use persistent consensus (so if a site is restarted during the shutdown, it doesn't ressurect the 
%%%   entire group).
%%%
%%% - Add add_site/remove_site APIs (we need corresponding repmgr functionality from Sleepycat)
%%%
%%% - Why can't 2 sites elect a master (with full or normal elections)?  Alan says that there is special-case code for 2 sites
%%%   - ask about that.
%%%
%%% - If do_txn_on_master crashes in a way that suggests that the site might not be master (i.e. db_put returns 'cannot write to database')
%%%   then reset entire replication_group's state (don't kill all port_coords, but poll local bdb rep_state to get status).
%%%
%%% - add a do_distributed_txn API (main part of which will be in a similar API in port_coordinator 
%%%   -- this laye just needs to ensure that the Master in each replication group is the the site 
%%%   that participates).
%%%
%%% - consistency 
%%%
%%% We are simply trying to track the result of BDB's own consensus (election) algorithm.  
%%%
%%% We don't currently use any proven group-comm (e.g. spread, horus, isis, ensemble) or consensus algorithm 
%%% (paxos), but in theory we should.
%%%
%%% We do rely on the fact thag BDB replication events to a single site are well-spaced in time.
%%%
%%% There are a number of race conditions in the system
%%%
%%%  1. The underlying BDB call to the set_event_notify callback is implicitly a race condition because 
%%%     it is subject to thread scheduling.
%%%
%%%  2. That callback enqueues an operation to be executed by the Erlang VM in ready_async.
%%%     The queue then defines an order on the events, but the order of enqueing is again subject to 
%%%     thread scheduling.  Using a static counter as a logical clock (to send with the message) 
%%%     would fix this but not affect #1, so would be useless.
%%%
%%%  3. Because it is not a proper causal-order group comm system, messages may be delivered from two 
%%%     different processes in an order that is different from the ordering of event in those two processes.
%%% 
%%%-------------------------------------------------------------------
-module(berkeley_db_replication_group).

-behaviour(gen_server).

%% API
-export([start_sites/3,              % returns [GoodPidsSites, BadSites], does NOT link
         start/4,                    % returns Pid or throws, does NOT link. Used by start_sites.
         start_link/4,               % returns {ok, Pid} or {error, Error} --- for use by conventional supervisors;
         get_proc_name/2,
         get_port_coordinator_proc_name/2,
         get_master_location_hint/2,
         get_master_location_hint/3, % also takes a ClientServerTimeout
         do_txn_on_master/2, 
         do_txn_on_master/3,         % also takes TxnFunTimeout   (that timeout does NOT imply waiting for a master to be discovered)
         do_txn_on_master/4,         % also takes NumDeadlockRetries
         do_txn_on_any_site/2,       % Will use an in-sync site if there is one. Prefers local site, then any other non-master site, then master.
         do_txn_on_any_site/3,
         do_txn_on_any_site/4,
         remove_site/1, 
         get_group_status/1,
         stop_entire_group/1, 
         get_node_from_master_location_hint/1
        ]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% Private API
-export([remote_find_or_spawn_site/3   % exported for use by rpc:call
        ]).


-define(DEFAULT_CONFIG_PROPLIST, 
        [{common_port_coordinator_config, []},
         {find_other_sites_period_ms,     5000}, % TODO: add random fuzz to de-phase sites
         {spawn_other_sites,              true}  % (set to 'disabled' if each site is in a tranditional supervisor)
         ]).


%% A 'site id' is:
%%
%%    {PortCoordName, RemoteNodeName}
%%
%% Note this does NOT include doesn't include the ReplicationGroupName 
%% (but that is always known by context as it is the same for all sites)
%% And it does NOT include replication_group server's process name 
%% (but that can be calculated from the site's PortCoordName name
%% and ReplicationGroupName).

-include("berkeley_db_replication_group_internal.hrl").

-record(state, {current_master,     % unknown | this_site | OtherSiteId
                this_site_status,   % #site_status (when status cannot be 'unknown' and both replication_group_pid and monitor_ref are 'unknown')
                other_sites_status, % [{OtherSiteId, #site_status}, ...] 
                other_sites,        % stored in State because we can add and remove sites (eventually this will be in a BDB on disk)'
                notify_clients      % list of {From, Now, Timeout} of clients to notify when we discover a master
               }).
%% TODO: Eventually 'OtherSites' will be stored in a group-membership record in a replicated BDB.

-record(site_status, {status,                 % {unknown | master | {replica, not_in_sync|in_sync}, 
                      last_updated_time,      % local erlang:now() of last change to status | unknown
                      replication_group_pid,  % ReplicationGroupPid | unknown
                      monitor_ref             % Ref | unknown
                     }).
%% Note that in term comparisons, atom (e.g. 'unknown') < any erlang:now(), as any atom < any tuple

-define(get_this_site_status(State),      (State#state.this_site_status)#site_status.status).
-define(get_this_site_status_time(State), (State#state.this_site_status)#site_status.last_updated_time).


-define(BDB_H, berkeley_db_helpers).
-define(BDB_U, berkeley_db_utilities).

-define(STRINGIZE(Arg), ??Arg).

-define(seed_rng(), 
        {RNGSeed1, RNGSeed2, RNGSeed3} = now(),
        random:seed(RNGSeed1, RNGSeed2, RNGSeed3)).


%%====================================================================
%% API
%%====================================================================

-define(THROW_SIMPLE_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_replication_group_error, Error})
           end).

%% This must be used for any API that takes a SiteLocator or MasterLocationHint.
%% Note that it throws any term from {error, Term} as is -- it does NOT wrap it in {bdb_replication_group_error, _}
%% That's because some errors (specifically txn_fun_error) are not replication_group errors - they are caller errors 
%% that we are passing through.
-define(THROW_REMOTE_SITE_ERRORS(Expr, SiteLocator),
           try Expr of
               {ok, Result}    -> Result;
               {error, Reason} -> throw(Reason)
           catch
               exit:Reason ->
                   %% OTP will cause the client to exit if the destination process is down.
                   %% We convert that to an exception so the client can retry with another site.
                   throw({bdb_replication_group_error, {site_is_down, SiteLocator, Reason}})
           end).

%%--------------------------------------------------------------------
%% Function: start_sites() -> {ok,[ListOfPids]} | {error,Error}
%%
%% Description: Starts MULTIPLE sites on any combination of local and remote-nodes.
%%
%% AllSites must be 
%%
%%           [ {{SiteName, NodeName}, BdbEnvDir, {HostName, Port}, SiteSpecificPortCoordConfigProplist}, 
%%             ... ]
%%
%% See also definition of DEFAULT_CONFIG_PROPLIST.
%%
%% Returns   {GoodSitesPids, ErrorSiteaReasons}
%%
%%              where a 'good site/pid' is     {Site, Pid}
%%              and an 'error' site/reason' is {Site, Reason}
%%
%% The order of sites in the result is not guaranteed to match the order of the parameter list.
%%
%% This does NOT link or monitor the resulting pids - the caller should do that if desired.
%% The caller can easily call erlang:monitor on the returned GoodPids.
%%
%% Rationale:
%%
%% This is a convenience function that is somewhat redundant as 
%% starting a single site somewhere will attempt to spawn all of the others, so this function  
%% is somewhat redundant.
%%
%% It exists because: 
%%
%%  - the caller might not want a site on it's local node
%%  - the caller can learn which sites were successfully spawned and which ones failed
%%
%% This calls start() on the required nodes (inc local node if necessary).
%%--------------------------------------------------------------------
start_sites(ReplicationGroupName, AllSites, ConfigProplist) ->    

    %% gen_server does not have a remote-spawn-link facility, so we must use an rpc to 
    %% spawn on remote nodes.
    %%
    %% There may be many sites (almost certainly more than 1), so we start them in parallel.
    %%
    %% For now we use this (which spawns a local process per list element, each of which calls rpc:call and blocks until it returns):
    %%   ?BDB_U:pmap(Fun/1, List) -> ResultsList
    %%   Evaluates Fun for every element Elem in List, in parallel. Returns the list of return values, in the same order as the input list.
    %%
    %% An alternative would be to use
    %%   rpc:async_call(Node, Module, Function, Args) -> Key
    %%   Implements call streams with promises, a type of RPC which does not suspend the caller until the result is finished. Instead, a key is returned which can be used at a later stage to collect the value. The key can be viewed as a promise to deliver the answer. 
    %%   In this case, the key Key is returned, which can be used in a subsequent call to yield/1 or nb_yield/1,2 to retrieve the value of evaluating apply(Module, Function, Args) on the node Node. 
    %%
    %% Important: pmap spawns short-lived processes to do call our Fun, so obviously the Fun cannot use start_link as the spawned 
    %% server would be linked to the shor-lived pmap helper process.  But were not using links anyway.
   
    lists:foldl(fun({Site, {ok, Pid}},       {Good, Bad})  -> {[{Site, Pid} | Good], Bad};
                   ({Site, {error, Reason}}, {Good, Bad})  -> {Good, [{Site, Reason} | Bad]};
                   ({Site, Other},           {Good, Bad})  -> {Good, [{Site, Other} | Bad]}
                end,
                {[], []},
                ?BDB_U:pmap(
                   fun({{_, NodeName}, _, _, _} = Site) ->
                           {Site, 
                            try rpc:call(NodeName, ?MODULE, start, [ReplicationGroupName, 
                                                                    Site,
                                                                    AllSites -- [Site], 
                                                                    ConfigProplist]) 
                                of
                                Pid when is_pid(Pid)  -> {ok, Pid};
                                Other                 -> {error, Other}
                            catch
                                Class:Reason          -> {error, {Class, Reason}}
                            end}
                   end,
                   AllSites)).


%%--------------------------------------------------------------------
%% Function: start() -> Pid or throws
%%
%% Description: Starts ONE site on the LOCAL node.
%%
%% This does NOT link the caller to the spawned process.
%% If that is desired then the caller should use erlang:monitor() on the returned pid.
%%
%% ThisSite      must be   {{SiteName, node()}, BdbEnvDir, {HostName, Port}, SiteSpecificPortCoordConfig}
%% OtherSites    must be [ {{SiteName, NodeName}, BdbEnvDir, {HostName, Port}, SiteSpecificPortCoordConfig}, ... ]
%%
%%--------------------------------------------------------------------
start(ReplicationGroupName, 
      {{SiteName, NodeName}, _, {_, _}, _} = ThisSite,
      OtherSites, 
      ConfigProplist) ->

    NodeName = node(), % assert

    ?THROW_SIMPLE_ERRORS( gen_server:start({local, get_proc_name(ReplicationGroupName, SiteName)}, 
                                           ?MODULE, 
                                           {ReplicationGroupName, ThisSite, OtherSites, ConfigProplist},
                                           []) ).


%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%%
%% Description: Starts ONE site on the LOCAL node.
%%
%% This is for use with conventional supervisors.
%% Returns 
%%         {ok, Pid}
%%         {error, Error}
%%
%%--------------------------------------------------------------------
start_link(ReplicationGroupName, 
           {{SiteName, NodeName}, _, {_, _}, _} = ThisSite,
           OtherSites, 
           ConfigProplist) ->

    NodeName = node(), % assert

    gen_server:start_link({local, get_proc_name(ReplicationGroupName, SiteName)}, 
                          ?MODULE, 
                          {ReplicationGroupName, ThisSite, OtherSites, ConfigProplist},
                          []).


%%--------------------------------------------------------------------
%% Function: get_proc_name(ReplicationGroupName, SiteName)
%%
%% Description: Finds the locally-registered name of the replication_group process 
%%              for a given group/site
%%
%%--------------------------------------------------------------------
get_proc_name(ReplicationGroupName, SiteName) ->
    list_to_atom("bdb_" ++ atom_to_list(ReplicationGroupName) ++ "_" ++ atom_to_list(SiteName) ++ "_replication_group").


%%--------------------------------------------------------------------
%% Function: get_port_coordinator_proc_name(ReplicationGroupName, SiteName)
%%
%% Description: Finds the locally-registered name of the port_coordinator process 
%%              for a given group/site
%%
%%--------------------------------------------------------------------
get_port_coordinator_proc_name(ReplicationGroupName, SiteName) ->
    list_to_atom("bdb_" ++ atom_to_list(ReplicationGroupName) ++ "_" ++ atom_to_list(SiteName) ++ "_port_coordinator").


%%--------------------------------------------------------------------
%% Function: get_master_location_hint(SiteLocator, FindMasterTimeout)
%%
%% Description: See API documentation at top of file.
%%
%% This supports FindMasterTimeout 0 .. infinity, but ClientServerTimeout should be 1..N (i.e. not infinity).
%% If FindMasterTimeout is infinity then ClientServerTimeout is ignored (i.e. set to infinity too).
%%
%% SiteLocator is {ReplicationGroupName, SiteName, SiteNode}
%%
%% Returns an opaque MasterLocationHint object.
%%
%%--------------------------------------------------------------------

%% This variant uses the default gen_server 5000ms client-server timeout
get_master_location_hint({ReplicationGroupName, SiteName, SiteNode} = SiteLocator, FindMasterTimeout) ->
    get_master_location_hint({ReplicationGroupName, SiteName, SiteNode} = SiteLocator, FindMasterTimeout, 5000).

get_master_location_hint({ReplicationGroupName, SiteName, SiteNode} = SiteLocator, FindMasterTimeout, ClientServerTimeout) ->

    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               {get_master_location_hint, FindMasterTimeout},
                                               case FindMasterTimeout of 
                                                   infinity -> infinity;
                                                   _        -> FindMasterTimeout + ClientServerTimeout
                                               end),
                               SiteLocator).

%%--------------------------------------------------------------------
%% Function: do_txn_on_master
%%
%% Description: See API documentation at top of file.
%%
%% There are convenience functions provide a default timeout and NumDeadlockRetries.
%% For each function there are 2 clauses:
%%
%%  - first clause takes a MasterLocationHint which is currently a Distributed Erlang Pid
%%       (this may be changed in future but it's the fastest option now)
%%
%%  - second clause takes {ReplicationGroupName, SiteName, Node} and is used by 
%%    clients when they don't have a MasterLocationHint.
%%    The MasterLocationHint might eventually change to this format (or some other).
%%
%%--------------------------------------------------------------------

%% Convenience functions tha provide defaults
do_txn_on_master(MasterLocationHint, TxnFun) when is_record(MasterLocationHint, master_location_hint) ->
    do_txn_on_master(MasterLocationHint, TxnFun, 0, 5000); % default NumDeadlockRetries & gen_server timeout
do_txn_on_master(SiteLocator, TxnFun) ->
    do_txn_on_master(SiteLocator, TxnFun, 0, 5000).

do_txn_on_master(MasterLocationHint, TxnFun, NumDeadlockRetries) when is_record(MasterLocationHint, master_location_hint) ->
    do_txn_on_master(MasterLocationHint, TxnFun, NumDeadlockRetries, 5000); % default gen_server timeout
do_txn_on_master(SiteLocator, TxnFun, NumDeadlockRetries) ->
    do_txn_on_master(SiteLocator, TxnFun, NumDeadlockRetries, 5000).

%% The main function
do_txn_on_master(MasterLocationHint, TxnFun, NumDeadlockRetries, TxnFunTimeout) when is_record(MasterLocationHint, master_location_hint) ->

    ?THROW_REMOTE_SITE_ERRORS( gen_server:call(MasterLocationHint#master_location_hint.pid,
                                               {do_txn_on_master, TxnFun, NumDeadlockRetries, TxnFunTimeout},
                                               TxnFunTimeout),
                               MasterLocationHint#master_location_hint.site_locator );

do_txn_on_master({ReplicationGroupName, SiteName, SiteNode} = SiteLocator, % a tuple (single parameter) so it can also be a MasterLocationHint
                 TxnFun, 
                 NumDeadlockRetries,
                 TxnFunTimeout) -> 

    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               {do_txn_on_master, TxnFun, NumDeadlockRetries, TxnFunTimeout},
                                               TxnFunTimeout),
                               SiteLocator ).


%%--------------------------------------------------------------------
%% Function: do_txn_on_any_site
%%
%% Description: See API documentation at top of file.
%%
%% There are convenience functions provide a default timeout and NumDeadlockRetries.
%%--------------------------------------------------------------------

%% Convenience functions tha provide defaults
do_txn_on_any_site(SiteLocator, TxnFun) ->
    do_txn_on_any_site(SiteLocator, TxnFun, 0, 5000).  % default NumDeadlockRetries & gen_server timeout

do_txn_on_any_site(SiteLocator, TxnFun, NumDeadlockRetries) ->
    do_txn_on_any_site(SiteLocator, TxnFun, NumDeadlockRetries, 5000). % default gen_server timeout

do_txn_on_any_site({ReplicationGroupName, SiteName, SiteNode} = SiteLocator,
                   TxnFun, 
                   NumDeadlockRetries,
                   TxnFunTimeout) -> 

    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               {do_txn_on_any_site, TxnFun, NumDeadlockRetries, TxnFunTimeout},
                                               TxnFunTimeout),
                               SiteLocator ).


%%--------------------------------------------------------------------
%% Function: remove_site(ReplicationGroupName, {SiteName, SiteNode})
%%
%% Description: 
%%
%%  Removes a given site from the group and shuts down that site.
%%  The site can be added back into the group with a call to add_site().
%%
%% TODO:
%%
%%  Currently this just makes all sites forget about the target site, and 
%%  then shuts it down.
%%  We we have persistent group membership, this will change that membership.
%%
%%  If the target site happens to be the master then we must do a clean 
%%  safe fail-over
%%    - suspend writes on that site (not all sites, so not same as stop_entire_group)
%%    - shutdown that site gracefully (i.e. so that an election is called immediately 
%%      and the ex-master does not participate)
%%      e.g. set ex-master's site-priority to   and then call rep_start(CLIENT) to downgrade it,
%%      (check with Sleepycat that repmgr  calls an election immediately if the master does it). 
%%
%%--------------------------------------------------------------------
remove_site({ReplicationGroupName, SiteName, SiteNode} = SiteLocator) ->
    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               remove_self),
                               SiteLocator ).


%%--------------------------------------------------------------------
%% Function: get_group_status(SiteLocator)
%%
%% Description: 
%%
%%  Returns current status of all the sites.
%%
%%           [{SiteId, Status}, ... ]
%%
%%  where SiteId = {SiteName, SiteNode}
%%        Status = unknown | master | {replica, not_in_sync} | {replica, in_sync}
%%
%%  Of course, this information may be out of date even before the  caller receives it.
%%
%%--------------------------------------------------------------------
get_group_status({ReplicationGroupName, SiteName, SiteNode} = SiteLocator) ->
    ?THROW_REMOTE_SITE_ERRORS( gen_server:call({get_proc_name(ReplicationGroupName, SiteName), SiteNode},
                                               get_group_status),
                               SiteLocator ).


%%--------------------------------------------------------------------
%% Function: stop_entire_group(SiteLocator)
%%
%% Description: 
%%
%%   Performs a *temporary* shutdown of ALL sites in the group.
%%   'Temporary' means that the group membership is NOT changed 
%%   (when we eventually store group membership in a replicated database).
%%
%% This take the same kind of 'site locator' as do_txn_on_master
%% -- i.e. a MasterLocationHint or any {ReplicationGroupName, SiteName, Node}
%%
%% TODO: 
%%
%% Currently this command just shuts down all sites and RETURNS WITHOUT WAITING.
%% (it does not wait as it is hard to avoid deadlocks).
%%
%% It needs to be completely rewritten to do the following:
%%
%%    - suspend writes (on all sites not just current master, so even if master fails we won't start taking writes again)
%%    - wait for all sites to sync-up (with a timeout)
%%    - shutdown all sites.
%%    - return when complete (or an error occurs)
%%
%% This will use the persistent group-membership database as a group consensus persistent state machine.
%%
%%--------------------------------------------------------------------
stop_entire_group(MasterLocationHint) when is_pid(MasterLocationHint) ->
    ok = gen_server:cast(MasterLocationHint, {stop_entire_group, self()});

stop_entire_group({ReplicationGroupName, SiteName, SiteNode}) ->
    ok = gen_server:cast({get_proc_name(ReplicationGroupName, SiteName), SiteNode}, {stop_entire_group, self()}).


%%--------------------------------------------------------------------
%% Function: get_node_from_master_location_hint(MasterLocationHint)
%%
%% Description: Allows application to move work to the master node,
%%              to reduce number of hops when invoking do_txn_on_master 
%%              multiple times.
%%
%%--------------------------------------------------------------------
get_node_from_master_location_hint(#master_location_hint{site_locator = {_, _, Node}}) ->
    Node.


%%-define(INFO_LOGGER,    error_logger:info_msg).
%%-define(WARNING_LOGGER, error_logger:warning_msg).
%%-define(ERROR_LOGGER,   error_logger:error_msg).
-define(INFO_LOGGER,    logger).
-define(WARNING_LOGGER, logger).
-define(ERROR_LOGGER,   logger).


%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%
%%--------------------------------------------------------------------
init({ReplicationGroupName, 
      {{ThisSiteName, ThisSiteNode} = ThisSiteId, ThisSiteEnvDir, {ThisSiteHostName, ThisSitePort}, ThisSitePortCoordConfig} = ThisSite, 
      OtherSites, 
      SpecifiedConfigProplist}) ->

    ThisSiteNode = node(),  % assert

    ?INFO_LOGGER("starting ~w~n", [self()]),
    process_flag(trap_exit, true),

    ThisSitePortCoordName = get_port_coordinator_proc_name(ReplicationGroupName, ThisSiteName),

    %% Validate config
    lists:foreach(fun({common_port_coordinator_config, _}) -> ok;
                     ({find_other_sites_period_ms, _})     -> ok;
                     ({spawn_other_sites, _})              -> ok;
                     (Other)                               -> throw({unrecognized_config_option, Other})
                  end,
                  SpecifiedConfigProplist),

    %% Reconcile with default config
    ConfigProplist = ?BDB_U:get_dominant_proplist_values([SpecifiedConfigProplist,
                                                          ?DEFAULT_CONFIG_PROPLIST]),

    %% Prepare port_coordinator configuration for this site.
    %%  - tell BDB-HA which host IP/names and ports to use.
    %%  - tell the local port_coordinator that we we want to receive all replication_gen_event messages 

    PortCoordConfig = 
        [{replication, [{this_hostname_port, {ThisSiteHostName, ThisSitePort}},
                        {peer_hostnames_ports, lists:map(fun({_, _, HostNamePort, _}) -> 
                                                                 HostNamePort
                                                         end,
                                                         OtherSites)},
                        {event_handlers, [{berkeley_db_replication_group_gen_event_handler,
                                           %% It is OK to use 'self()' here, rather than a process name.  
                                           %% If we die and restart and find that port_coordinators are 
                                           %% already running then we will need to get an update of absolute state anyway.
                                           {forward_events_as_messages_to, local_sleepycat_event, self()}}]}
                       ]}]
        ++ ThisSitePortCoordConfig 
        ++ proplists:get_value(common_port_coordinator_config, ConfigProplist),

    %% Spawn our local port_coordinator first, as this will validate the config.
    {ok, PortCoordPid} = berkeley_db_port_coordinator:start_link(ThisSitePortCoordName, ThisSiteEnvDir, PortCoordConfig),

    ?INFO_LOGGER("local port_coordinator: ~w (registered as ~w)~n", [PortCoordPid, ThisSitePortCoordName]),
    
    %% TODO: is it worth using the port_coordinator's 'support_procs' technique (an element in the State record to hold spawned pids)?
    %% So far, storing the pids in the process dictionary (as they are immutable -- we die if they die) seems reasonable.
    %% The port_coordinator uses support_procs to register 'clean shutdown' functions, to be called in terminate().
    %% But we don't need clean shutdown functions yet.
    %% And it doesn't seem likely that we will have a lot of support procs, whereas the port_coordinator may accumulate more (de-frag etc).

    %% Spawn a helper process to start trying to connect to peer sites that are 'unknown' (and spawn them if they are not running).
    ThisSiteReplicationGroupPid = self(),
    put(find_or_spawn_peers_pid, 
        proc_lib:spawn_link(
          fun() -> find_or_spawn_peers_process(
                     ReplicationGroupName, 
                     ThisSiteId,
                     ThisSiteReplicationGroupPid, 
                     [SiteId || {SiteId, _, _, _} <- OtherSites],  % SiteIds to find/spawn
                     proplists:get_value(find_other_sites_period_ms, ConfigProplist),
                     case proplists:get_value(spawn_other_sites, ConfigProplist) of
                         true ->                       
                             %% info required when spawning a site
                             {[ThisSite | OtherSites],             % AllSites
                              ConfigProplist};         
                         disabled ->
                             disabled
                     end)
          end)),

    State = #state{current_master     = unknown,
                   this_site_status   = #site_status{status            = {replica, not_in_sync},
                                                     last_updated_time = now()},                     % other members are intentionally left as 'undefined'                   
                   other_sites_status = [{OtherSiteId, #site_status{status                = unknown,
                                                                    last_updated_time     = unknown, % we don't use the default of 'undefined' as we want to detect bugs of omission (failing to set a field)
                                                                    replication_group_pid = unknown,
                                                                    monitor_ref           = unknown}}
                                         || {OtherSiteId, _, _, _} <- OtherSites],
                   other_sites        = OtherSites,
                   notify_clients      = []},

    %% Save immutable state
    put(config_proplist,            ConfigProplist),
    put(replication_group_name,     ReplicationGroupName),
    put(this_site_port_coord_name,  ThisSitePortCoordName),
    put(this_site_id,               ThisSiteId),
    put(this_site_name,             ThisSiteName),
    put(this_site_port_coord_pid,   PortCoordPid),

    assert_state_invariants(State),

    ?seed_rng(),
    {ok, State}.


%%--------------------------------------------------------------------
%% Function: handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------

handle_call({get_master_location_hint, FindMasterTimeout}, From, State) ->

    case State#state.current_master of 

        this_site ->
            ?INFO_LOGGER("get_master_location_hint: returning this_site~n"),
            {reply, 
             {ok, #master_location_hint{site_locator = {get(replication_group_name), get(this_site_name), node()},
                                        pid          = self()}},
             State};

        {MasterSiteName, MasterSiteNode} = MasterSiteId ->

            ?INFO_LOGGER("get_master_location_hint: returning ~w~n", [MasterSiteId]),
            {value, {_, MasterSiteStatus}} = lists:keysearch(MasterSiteId, 1, State#state.other_sites_status),

            {reply, 
             {ok, #master_location_hint{site_locator = {get(replication_group_name), MasterSiteName, MasterSiteNode},
                                        pid          = MasterSiteStatus#site_status.replication_group_pid}},
            State};

        unknown ->
            ?INFO_LOGGER("get_master_location_hint: waiting to discover master~n"),

            case FindMasterTimeout of
                0 ->
                    %% caller was just polling -- reply immediately
                    {reply, {error, {bdb_replication_group_error, no_master}}, State};

                infinity ->
                    %% Add this client to the list of clients to notify if we do discover a master.
                    ?INFO_LOGGER("get_master_location_hint: will notify client ~w when master is discovered (infinite timeout)~n", [From]),
                    {noreply, State#state{notify_clients = [{From, infinity} | State#state.notify_clients]}};

                _ ->
                    %% Add this client to the list of clients to notify if we do discover a master,
                    %% *and* also send a message to ourself on a timer to tell the client that there is no_master 
                    %% (if one has not been discovered by then).
                    %% (Note: we use erlang:start_timer not timer:send_after as apparently the latter is very inefficient).
                    %% start_timer sends a message: {timeout, TimerRef, Msg}.
                    %% We store the TimerRef in notify_clients so we can cancel the timer if we do discover a master.

                    TimerRef = erlang:start_timer(FindMasterTimeout, self(), {notify_client_timer_expired, From}), 
                    ?INFO_LOGGER("get_master_location_hint: will notify client ~w (~w ms timeout, TimerRef ~w)~n", [From, FindMasterTimeout, TimerRef]),

                    {noreply, State#state{notify_clients = [{From, TimerRef} | State#state.notify_clients]}}
            end
    end;

    
handle_call({do_txn_on_master, _TxnFun, _NumDeadlockRetries, _Timeout} = Command, From, State) ->

    case State#state.current_master of 

        this_site ->
            ?INFO_LOGGER("do_txn_on_master on local site~n"),
            ok = async_do_txn_on_master_local_site(Command, From),
            {noreply, State};

        {_, _} = MasterSiteId ->
            ?INFO_LOGGER("do_txn_on_master: we are not master - forwarding to master site ~w~n", [MasterSiteId]),

            %% Forward command to remote site that is master
            %% If the destination server dies or hangs then the gen_server:call will timeout.

            %% Note: MasterSiteId is {SiteName, Node}, so we have to translate it to find 
            %% the replication_group process.
            %% There are two translations:
            %%   1.   {get_proc_name(get(replication_group_name), SiteName), Node}
            %%   2.   Look for MasterSiteId in State#other_sites_status and use the replication_group_pid element.

            {value, {_, #site_status{replication_group_pid = MasterPortCoordPid}}} 
                = lists:keysearch(MasterSiteId, 1, State#state.other_sites_status),

            ok = forward_command_to_remote_site(MasterSiteId, MasterPortCoordPid, Command, From),
            {noreply, State};

        unknown ->
            {reply, {error, {bdb_replication_group_error, no_master}}, State}
    end;


handle_call({do_txn_on_any_site, _TxnFun, _NumDeadlockRetries, _Timeout} = Command, From, State) ->

    %% We prefer sites in the following order
    %%   - local site if it is in-sync (even if it happens to be the master)
    %%   - any other in-sync replica (i.e. excluding the master)
    %%   - the master
    %%   - any site that we can find that is running (prefering the local site of course)

    case ?get_this_site_status(State) of 

        {replica, in_sync} ->
            ?INFO_LOGGER("do_txn_on_any_site: using local site as it is an in-sync replica~n"),
            ok = async_do_txn_on_any_site_local_site(Command, From),
            {noreply, State};

        master ->
            ?INFO_LOGGER("do_txn_on_any_site: using local site as it is master~n"),
            ok = async_do_txn_on_any_site_local_site(Command, From),
            {noreply, State};

        _ -> 
            %% Local site is not in sync
            %% Are there any in-sync replicas at all?
            case [SiteId || {SiteId, SiteStatus} <- State#state.other_sites_status, 
                            SiteStatus#site_status.status =:= {replica, in_sync}] of
                [] ->
                    %% There are not any in-sync replicas.
                    %% Use the master if there is one 
                    %% (we know that the master cannot be the local site, as we checked earlier).
                    case State#state.current_master of 
                        {_, _} = MasterSiteId ->
                            %% We do have a master, so use that
                            {value, {_, #site_status{replication_group_pid = MasterPortCoordPid}}} 
                                = lists:keysearch(MasterSiteId, 1, State#state.other_sites_status),
                            ok = forward_command_to_remote_site(MasterSiteId, MasterPortCoordPid, Command, From);

                        unknown ->
                            %% There is not even a master.
                            %% If the client really wanted an in-sync site, we should return an error here.
                            %% TODO: that should be a command parameter.
                            %% For the moment we choose to say that any site is better than failure,
                            %% so we execute on the local site.
                            ?INFO_LOGGER("do_txn_on_any_site: using local site as the least-worst option (**there are no in-sync sites and no master**)~n"),
                            ok = async_do_txn_on_any_site_local_site(Command, From)
                    end;

                InSyncReplicas ->
                    %% We have a choice of in-sync sites.
                    %% Pick one at random (the RNG was seeded in init()).
                    SiteId = lists:nth( random:uniform(length(InSyncReplicas)), InSyncReplicas),
                    {value, {_, #site_status{replication_group_pid = PortCoordPid}}} 
                        = lists:keysearch(SiteId, 1, State#state.other_sites_status),
                    ok = forward_command_to_remote_site(SiteId, PortCoordPid, Command, From)
            end,
            {noreply, State}
    end;


handle_call(remove_self, _, State) ->
    assert_state_invariants(State),

    ?INFO_LOGGER("remove_self starting~n"),

    %% TODO: 
    %% 1. run a txn on the master to change group membership
    %% 2. if this site happens to be master then perform graceful failover
    %%    - suspend writes on this site only
    %%    - call an immediate election in which this site will not participate (e.g. set this site's priority to 0 and downgrade to replica)

    %% We are shutting down this site so we CAN block the server while we do the following:

    %% Stop trying to connect to/spawn ALL other sites
    %% Even if find_or_spawn_peers_pid _just_ died, we _know_ that we cannot possibly have eaten the EXIT message
    %% yet, so blocking for it is safe.   (That is not the case in terminate() of course.)
    FindOrSpawnPid = get(find_or_spawn_peers_pid),
    exit(FindOrSpawnPid, kill),
    receive {'EXIT', FindOrSpawnPid, _} -> ok end,

    %% Tell all other sites not to reconnect to / respawn this site
    %%
    %% Note: even when BDB (repmgr) allows us to break connections to sites (and prevent retries) 
    %% we don't need to call that API from this site as we are about to call terminate() and 
    %% shutdown the port_coordinator.

    %% Do this in parallel as the  group might be large
    ?BDB_U:pmap(fun({OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid}}) 
                   when is_pid(OtherSiteReplicationGroupPid) -> 

                        case (catch gen_server:call(OtherSiteReplicationGroupPid, {forget_other_site, get(this_site_id)})) of
                            ok ->
                                ?INFO_LOGGER("remove_self: successfully told other site ~w to forget this site~n", [OtherSiteId]);
                            Other ->
                                ?WARNING_LOGGER("remove_self: failed to tell other site ~w to forget this site (~w)~n", [OtherSiteId, Other])
                        end;

                   ({OtherSiteId, _}) ->
                        ?INFO_LOGGER("remove_self: not connected to ~w so don't need to tell it to forget about this site~n", [OtherSiteId])
                end,
                State#state.other_sites_status),

    %% Terminate
    %% We don't change State as there is nothing really meaningfull to change it too.
    {stop, normal, ok, State};


%% This is an internal command used by remove_self
handle_call({forget_other_site, OtherSiteId} = Msg, {FromPid, _}, State) ->
    assert_state_invariants(State),

    ?INFO_LOGGER("forget_other_site ~w~n", [OtherSiteId]),

    %% Stop trying to connect to/spawn the other site
    %% Note race condition (we might be connecting right now). So subscribe needs to handle case where OtherSiteId has been forgotten.
    get(find_or_spawn_peers_pid) ! Msg,

    %% If we have a connection to OtherSite then we need to de-monitor it.
    case lists:keysearch(OtherSiteId, 1, State#state.other_sites_status) of

        {value, {OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                           monitor_ref           = MonitorRef} = OtherSiteStatus}} when is_reference(MonitorRef) ->

            ?INFO_LOGGER("handle_cast(forget_other_site): de-monitoring OtherSiteId ~w (Status = ~w)~n", [OtherSiteId, OtherSiteStatus]),
            OtherSiteReplicationGroupPid = FromPid,    % assert: should only be told to forget a site by that site
            ok = erlang:demonitor(process, MonitorRef);

        {value, {OtherSiteId, OtherSiteStatus}} ->
            ?INFO_LOGGER("handle_cast(forget_other_site): no need to de-monitor OtherSiteId ~w (Status = ~w)~n", [OtherSiteId, OtherSiteStatus])

        %% We should never get 'false' (never be asked to forget the same site twice)
    end,

    %% TODO: when BDB (repmgr) allows us to break connections to sites (and prevent retries) we need to call that API here

    NewState = State#state{other_sites_status = lists:keydelete(OtherSiteId, 1, State#state.other_sites_status),
                           other_sites        = [OtherSite || {SiteId, _, _, _} = OtherSite <- State#state.other_sites, SiteId /= OtherSiteId]},

    assert_state_invariants(NewState),
    {reply, ok, NewState};


handle_call(get_group_status, _, State) ->
    assert_state_invariants(State),

    {reply, 
     {ok, [{get(this_site_id), ?get_this_site_status(State)}
           | [{SiteId, SiteStatus#site_status.status} || {SiteId, SiteStatus} <- State#state.other_sites_status]]},
     State}.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------

%% Another site wants to subscribe to our status updates.
%% This may be the first time we have learned of the other site (if so, we need to start monitoring it).
handle_cast({subscribe, OtherSiteId, OtherSiteReplicationGroupPid}, State) ->
    assert_state_invariants(State),

    %% We may or may not already know about the caller site.
    %% And independently of that, we may or may not have 'connected' to it (asked it for it's status).
    %%
    %% If we first learn of OtherSite via this request to subscribe, then:
    %%
    %% - We record its pid/monitor ref (i.e. before find_or_spawn_peers_pid finds them, 
    %%   so handle_cast(found_other_site) must be prepared for that).
    %%
    %% - We leave its status alone (as 'unknown')
    %%   TODO: the caller could include it's own status with the subscribe message, to allow us to fill 
    %%   it in here.  However, I'd like to have a single point that updates the status -- or we'd have 
    %%   to run the whole state machine here as well.
    %%
    %% -  We don't bother sending a 'found_other_site' message to ourselves, as find_or_spawn_peers_pid 
    %%    should do that very shortly and we don't want to do it multiple times.
    %%
    %% So the main thing we want to do here is to start monitoring the caller if we are not already 
    %% doing so.
    
    NewState = 
        case lists:keysearch(OtherSiteId, 1, State#state.other_sites_status) of

            {value, {OtherSiteId, #site_status{replication_group_pid = unknown} = OtherSiteStatus}} ->

                ?INFO_LOGGER("subscribe: Valid subscribe from entirely unknown site ~w (pid ~w)~n", [OtherSiteId, OtherSiteReplicationGroupPid]),

                %% If this doesn't get there then the other site will eventually ask us again.
                send_site_status(get(this_site_id), State, OtherSiteReplicationGroupPid),

                State#state{other_sites_status 
                            = lists:keyreplace(OtherSiteId, 
                                               1, 
                                               State#state.other_sites_status,
                                               {OtherSiteId, 
                                                OtherSiteStatus#site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                                                            monitor_ref           = erlang:monitor(process,
                                                                                                                   OtherSiteReplicationGroupPid)}})};

            {value, {OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid} = OtherSiteStatus}} ->
                ?INFO_LOGGER("subscribe: Valid subscribe from known site ~w (OtherSiteStatus ~w)~n", [OtherSiteId, OtherSiteStatus]),
                %% We may have learned about the other site via find_or_spawn_peers.
                %% So we don't need to monitor it (we are already), but we do need to allow it to subscribe.
                %% If this doesn't get there then the other site will eventually ask us again.
                send_site_status(get(this_site_id), State, OtherSiteReplicationGroupPid),
                State;

            {value, {OtherSiteId, OtherSiteStatus}} ->
                ?INFO_LOGGER("subscribe: Discarding INCONSISTEN attempt to subscribe from site ~w for a different pid (~w) than is already registered (OtherSiteStatus ~w)~n", [OtherSiteId, OtherSiteReplicationGroupPid, OtherSiteStatus]),
                %% A different Pid is claiming to be OtherSiteId.
                %% This can legitimately happen; another site could die, be respawned and attempt to subscribe to us 
                %% before the DOWN notice for the previous one reaches us (to eradicate the old pid).
                %% So we just ignore this -- the caller will retry indefinitely and the DOWN messate is guaranteed to reach 
                %% us eventually.  See handle_info('DOWN') for more discussion.
                State
        end,

    assert_state_invariants(NewState),
    {noreply, NewState};


%% This handles status updates from other sites.
%%
%% As find_or_spawn_peers_process subscribes to the other site on our behalf, 
%% this may be the first time we learn of the other site (if so, we need to start monitoring it).
%%
%% In certain cases we might get events out of order.  We use their last_updated timestamps (generated on 
%% the sending site -- *cannot* be compared to our own time) to order the messages and discard ones from the past.
%%
handle_cast({other_site_status, OtherSiteId, OtherSiteReplicationGroupPid, NewOtherSiteState, NewOtherSiteUpdatedAtTime}, State) ->
    assert_state_invariants(State),

    NewState = 
        case lists:keysearch(OtherSiteId, 1, State#state.other_sites_status) of

            {value, {OtherSiteId, #site_status{replication_group_pid = unknown} = OldOtherSiteStatus}} ->

                ?INFO_LOGGER("other_site_status: First status ~w for ~w (was ~w)~n", [NewOtherSiteState, OtherSiteId, OldOtherSiteStatus]),

                %% We need to know if the other site dies or becomes partitioned
                OtherSiteMonitorRef = erlang:monitor(process, OtherSiteReplicationGroupPid),

                %% Tell find_or_spawn_peers_pid that it can stop trying OtherSiteId.
                %% (If we ever receive a DOWN message for OtherSiteMonitorRef then 
                %% we tell find_or_spawn_peers_pid to start trying to find/spawn OtherSiteId again)
                get(find_or_spawn_peers_pid) ! {stop_trying_site, OtherSiteId},

                %% Now push the state change through the official state machine.
                %% We need to pre-insert the pid and monitor ref, but we let handle_other_site_status_update update the status field.

                ConnectedOtherSiteStatus = OldOtherSiteStatus#site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                                                          monitor_ref           = OtherSiteMonitorRef},
        
                handle_other_site_status_update(OtherSiteId, 
                                                ConnectedOtherSiteStatus, 
                                                NewOtherSiteState, 
                                                NewOtherSiteUpdatedAtTime, 
                                                State#state{other_sites_status = lists:keyreplace(OtherSiteId, 
                                                                                                  1, 
                                                                                                  State#state.other_sites_status, 
                                                                                                  {OtherSiteId, ConnectedOtherSiteStatus})});

            {value, {OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid,
                                               last_updated_time     = OldOtherSiteUpdatedAtTime} = OldOtherSiteStatus}} 
            when OldOtherSiteUpdatedAtTime < NewOtherSiteUpdatedAtTime ->

                ?INFO_LOGGER("other_site_status: Status update ~w for ~w (was ~w)~n", [NewOtherSiteState, OtherSiteId, OldOtherSiteStatus]),

                %% Tell find_or_spawn_peers_pid that it can stop trying OtherSiteId.
                %% (This is necessary here as well as the previous clause as we might have set replication_group_pid 
                %% when the site subscribed to us).
                get(find_or_spawn_peers_pid) ! {stop_trying_site, OtherSiteId},

                handle_other_site_status_update(OtherSiteId, OldOtherSiteStatus, NewOtherSiteState, NewOtherSiteUpdatedAtTime, State);

            {value, {OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid} = OldOtherSiteStatus}} ->
                ?INFO_LOGGER("other_site_status: Ignoring status update ~w for ~w as it is same age or older (~w) than the one we already have (~w)~n", [NewOtherSiteState, OtherSiteId, NewOtherSiteUpdatedAtTime, OldOtherSiteStatus]),
                State;

            {value, {OtherSiteId, OldOtherSiteStatus}} ->
                ?INFO_LOGGER("other_site_status: Discarding INCONSISTENT status update ~w for ~w as new pid (~w) does not match the pid we have recorded (~w)~n", [NewOtherSiteState, OtherSiteId, OtherSiteReplicationGroupPid, OldOtherSiteStatus]),
                %% A different Pid is already claiming to be OtherSiteId.
                %% This could legitimately happen; another site could die, be respawned and find_or_spawn_peers_pid could find it 
                %% before the DOWN notice for the previous one reaches us (to eradicate the old pid).
                %% So we just ignore this -- the caller will retry indefinitely and the DOWN message is guaranteed to reach 
                %% us eventually.  See handle_info('DOWN') for more discussion.
                State

        end,

    assert_state_invariants(NewState),
    {noreply, NewState};


handle_cast({stop_entire_group, FromPid}, State) ->
    assert_state_invariants(State),

    ?INFO_LOGGER("stop_entire_group~n"),

    %% Coordinate a clean **temporary** shutdown of all running sites.
    %% Important: in future when we have group membership recorded in a replicated BDB, 
    %% this API will *NOT* alter that membership database.

    %% TODO:
    %% The main purpose of this API is to:
    %% - suspend writes on all sites (not just the current master).  do_txn_on_master should throw an appropriate error.
    %% - wait for all sites to sync up  (with a timeout)
    %% - then shutdown all sites (as below)

    %% Stop trying to connect to/spawn ALL other sites
    %% Even if find_or_spawn_peers_pid _just_ died, we _know_ that we cannot possibly have eaten the EXIT message
    %% yet, so blocking for it is safe.   (That is not the case in terminate() of course.)
    FindOrSpawnPid = get(find_or_spawn_peers_pid),
    exit(FindOrSpawnPid, kill),
    receive {'EXIT', FindOrSpawnPid, _} -> ok end,
    

    %% Tell all other sites to shutdown
    %%
    %% We DON'T simply call remove_site() on all of the sites that we know about as
    %%   - at some point remove_site will remove the site from the group membership database.
    %%   - the site that we happen to be running on might not have a connection to some site.
    %%     i.e. We need stop_entire_group to be transitive.
    %%
    %% Currently this is implemented by sending stop_entire_group to all sites that we know about 
    %% (and they in turn will send it to sites that they know about).
    %% To minimise the messages, if we received 'stop_entire_group' from a process that we recognise as a site 
    %% then we don't send the command to the site 
    %%
    %% Note: even when BDB (repmgr) allows us to break connections to sites (and prevent retries) 
    %% we don't need to call that API from this site as we are about to call terminate() and 
    %% shutdown the port_coordinator.

    %% Do this in parallel as the group might be large and we want to shutdown quickly.
    lists:foreach(fun({OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid}}) 
                     when is_pid(OtherSiteReplicationGroupPid), OtherSiteReplicationGroupPid /= FromPid -> 

                          ?INFO_LOGGER("stop_entire_group: forwarding to ~w~n", [OtherSiteId]),
                          ok = gen_server:cast(OtherSiteReplicationGroupPid, {stop_entire_group, self()});

                     ({OtherSiteId, _}) ->
                          ?INFO_LOGGER("stop_entire_group: not connected to ~w (or command was received from that site) so don't need to tell it to stop_entire_group~n", [OtherSiteId])
                  end,
                  State#state.other_sites_status),

    %% Terminate
    %% We don't change State as there is nothing really meaningfull to change it too.
    {stop, normal, State}.


%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------

handle_info({local_sleepycat_event, Event}, State) ->
    %% This is a message from berkeley_db_replication_group_gen_event_handler 
    %% from the port_coordinator on THIS site.
    %% handle_this_site_sleepycat_event() does pre- and post- calls to assert_state_invariants(),
    %% so we don't have to do it here.
    
    {noreply, handle_this_site_sleepycat_event(Event, State)};


handle_info({timeout, TimerRef, {notify_client_timer_expired, From}}, State) ->

    %% A client was only willing to wait until now to discover a master

    %% Of course, we might actually have discovered a master (and told the client about it) before this timeout happened.
    %% So we only send 'no_master' to the client if 'From' is still in the notify_clients list.
    %% (We don't have to worry about the same client process asking for notification again -- after we've previously 
    %% replied to it but before this timeout arrives -- as 'From' contains a Ref and so is unique, not just a pid.)

    {noreply,
     case ?BDB_U:lists_find_and_remove_first({From, TimerRef}, State#state.notify_clients) of

         {found, ListAfterRemoval} ->

             ?INFO_LOGGER("notify_client_timer_expired: sending no_master to ~w~n", [From]),
             gen_server:reply(From, {error, {bdb_replication_group_error, no_master}}),    
             State#state{notify_clients = ListAfterRemoval};

         not_found ->
             ?INFO_LOGGER("notify_client_timer_expired: ~w is not in notify_list~n", [From]),
             State
     end};


handle_info({'DOWN', MonitorRef, process, DownPid, DownReason} = Msg, State) ->
    assert_state_invariants(State),

    %% A site we were monitoring has disappeared.  
    ?INFO_LOGGER("Process ~w is DOWN : ~w~n", [DownPid, DownReason]),

    NewState = 
        %% Find out which site
        case [{OtherSiteId, OtherSiteStatus}
              || {OtherSiteId, #site_status{replication_group_pid = Pid,
                                            monitor_ref           = Ref} = OtherSiteStatus} 
                     <- State#state.other_sites_status, Pid == DownPid, Ref == MonitorRef] of

            [{OtherSiteId, OtherSiteStatus}] ->

                ?INFO_LOGGER("Site ~w is DOWN (had status ~w) : ~w~n", [OtherSiteId, OtherSiteStatus, DownReason]),

                %% Tell find_or_spawn_peers_pid to start attempting to find/spawn it again
                get(find_or_spawn_peers_pid) ! {find_or_spawn_site, OtherSiteId},

                case State#state.current_master of

                    OtherSiteId ->
                        ?WARNING_LOGGER("We lost the master site ~w~n", [OtherSiteId]),
                        
                        %% Just because we lost connection with the master we cannot assume that everyone else has.
                        %%
                        %% If Info is 'noconnection' (note no '_') then we have a network partition which might only be affecting this 
                        %% one node-to-node connection.
                        %%
                        %% If Info is 'noproc' then the process died between find_or_spawn finding or spawning it, and 
                        %% subscribe calling erlang:monitor() on it.
                        %%
                        %% If 'Info' is anything else then it is the ExitReason of the process.
                        %%
                        %% In the latter two cases we DO know that all sites have lost the master (at some point).
                        %% However, we DON'T act on it here as there is a race condition.  Those sites might already
                        %% have found another master, synced with it, and told us.  So we cannot legitimately 
                        %% change their status to not_in_sync.  We simply have to track their state changes as they happen -- 
                        %% we cannot make any guesses.
                        %%
                        %% However, as we _know_ that this_site_status is changing (to not_in_sync) here, we do send 
                        %% out that status update.  And all other affected sites will do the same of course.  
                        %% There is a race condition of course; the present 'DOWN' message may have arrived *after* we found and synced to 
                        %% a new master.  But we prevent that by:
                        %%
                        %% - handle_cast(found_new_site) explicitly refuses to subscribe to a new site if there is already a pid 
                        %%   registered for that site. The pid might be the same (i.e. find_or_spawn_peers just retried unecessarily 
                        %%   because of the race condition with telling it to stop; that's fine), or the pid might be a different pid
                        %%   which means that find_or_spawn_peers found or spawned the site before we got the DOWN message for the old
                        %%   site.  In that case we just ignore the found_new_site message; find_or_spawn_peers will retry indefinitely 
                        %%   and the DOWN message will eventually arrive.
                        %%
                        %% - subscribe explicitly refuses a subscribe attempt if a different pid is still registered 
                        %%   for that site.  (The caller will retry indefinitely and we will eventually get a DOWN message for the old pid).
                        %%
                        %% This is another place where we would benefit from a proper group view/epoch management system 
                        %% - to determine which messages were sent before the membership changed (master died) and which 
                        %% were sent after.
                        
                        State2 = State#state{current_master     = unknown,
                                             this_site_status   = #site_status{status            = {replica, not_in_sync},
                                                                               last_updated_time = now()},
                                             other_sites_status = lists:keyreplace(OtherSiteId, 
                                                                                   1, 
                                                                                   State#state.other_sites_status, 
                                                                                   {OtherSiteId, #site_status{status                = unknown,
                                                                                                              last_updated_time     = unknown,
                                                                                                              replication_group_pid = unknown,
                                                                                                              monitor_ref           = unknown}})},

                        assert_state_invariants(State2), % before we use it

                        %% Tell all other sites our new status
                        publish_state_change(State, State2),
                        State2;

                    StillMaster ->
                        ?INFO_LOGGER("We didn't lose the master site ~w~n", [StillMaster]),
                        %% Our state does not change, so there is no event to publish.
                        State#state{other_sites_status = lists:keyreplace(OtherSiteId, 
                                                                          1, 
                                                                          State#state.other_sites_status, 
                                                                          {OtherSiteId, #site_status{status                = unknown,
                                                                                                     last_updated_time     = unknown,
                                                                                                     replication_group_pid = unknown,
                                                                                                     monitor_ref           = unknown}})}                        
                end;

            _ ->
                ?ERROR_LOGGER("Got unexpected DOWN message ~w (OtherSitesStatus = ~w)~n", [Msg, State#state.other_sites_status]),
                exit({unexpected_DOWN_msg, Msg, State#state.other_sites_status})
        end,

    assert_state_invariants(NewState),
    {noreply, NewState}.


%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(Reason, _State) ->
    ?INFO_LOGGER("terminate: (reason: ~w)~n", [Reason]),

    %% Shut down support processes.
    %% As always in terminate(), the support processes might already have died
    %% and we might already have eaten the EXIT message.
    %% Fortunately erlang:monitor does what we want -- we will get a DOWN message
    %% even if the process is not running when we try to monitor it.

    FindOrSpawnPid = get(find_or_spawn_peers_pid),
    FindOrSpawnMonitorRef = erlang:monitor(process, FindOrSpawnPid),
    exit(FindOrSpawnPid, kill),

    PortCoordPid = get(this_site_port_coord_pid),
    PortCoordMonitorRef = erlang:monitor(process, PortCoordPid),
    ok = berkeley_db_port_coordinator:stop(PortCoordPid),

    %% Wait for the support processes to exit, otherwise they are killed when terminate returns (and their parent exits).
    receive {'DOWN', FindOrSpawnMonitorRef, process, _, _} -> ok end,
    receive {'DOWN', PortCoordMonitorRef, process, _, _} -> ok end.


%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

notify_clients_of_discovered_master(State) ->

    case State#state.current_master of 

        this_site ->

            notify_clients_of_discovered_master(
              State#state.notify_clients,
              {#master_location_hint{site_locator = {get(replication_group_name), get(this_site_name), node()},
                                     pid          = self()}, 
               State});

        {MasterSiteName, MasterSiteNode} = MasterSiteId ->

            {value, {_, MasterSiteStatus}} = lists:keysearch(MasterSiteId, 1, State#state.other_sites_status),

            notify_clients_of_discovered_master(
              State#state.notify_clients,
              {#master_location_hint{site_locator = {get(replication_group_name), MasterSiteName, MasterSiteNode},
                                     pid          = MasterSiteStatus#site_status.replication_group_pid}, 
               State})
    end.

notify_clients_of_discovered_master([], {_, State}) ->
    ?INFO_LOGGER("notify_clients_of_discovered_master: done all clients~n"),
    State#state{notify_clients = []};

notify_clients_of_discovered_master([{From, TimerRef} | T], {MasterLocationHint, _} = Constants) ->

    ?INFO_LOGGER("notify_clients_of_discovered_master: sending to ~w and cancelling timer ~w~n", [From, TimerRef]),
    gen_server:reply(From, {ok, MasterLocationHint}),

    if TimerRef /= infinity -> erlang:cancel_timer(TimerRef);
       true                 -> ok
    end,
                               
    notify_clients_of_discovered_master(T, Constants).


%%
%% find_or_spawn_peers_process
%%

find_or_spawn_peers_process(ReplicationGroupName, 
                            ThisSiteId, 
                            ThisSiteReplicationGroupPid, 
                            OtherSitesIds,
                            FindOtherSitesPeriodMs,
                            SpawnOtherSitesInfo) ->

    ?INFO_LOGGER("find_or_spawn_peers_process ~w starting~n", [self()]),

    put(find_other_sites_period_ms, FindOtherSitesPeriodMs),

    find_or_spawn_peers_loop(ReplicationGroupName, 
                             ThisSiteId,
                             ThisSiteReplicationGroupPid, 
                             SpawnOtherSitesInfo,
                             [{SiteId, 0} || SiteId <- OtherSitesIds],  % ThisPassSiteIds : [{SiteId, NumAttempts}, ...]
                             [],                                        % NextPassSiteIds
                             0).                                        % ReceiveTimeout -- start first pass immediately

%% The desired behaviour is:
%%
%% Every N milliseconds attempt to find or spawn all of the FindSiteIds
%% Don't pause between each one (or large groups will take a long time to spawn).
%% But do check (between each one) for requests (messages) to start or stop trying other sites,
%% and handle those requests immediately.
%% Eventually we will want all nodes to be out of phase (or use different retry periods),
%% so that the attempts to spawn other sites are well distributed in time (to minimize recovery latency).

find_or_spawn_peers_loop(ReplicationGroupName, 
                         ThisSiteId,
                         ThisSiteReplicationGroupPid, 
                         SpawnOtherSitesInfo,
                         [], 
                         [],
                         ReceiveTimeout) when ReceiveTimeout /= infinity ->

    %% no work to do, so wait forever (for a message)
    ?INFO_LOGGER("find_or_spawn_peers: transitioned to having no work to do: setting timeout to infinity~n"),
    find_or_spawn_peers_loop(ReplicationGroupName,
                             ThisSiteId,
                             ThisSiteReplicationGroupPid, 
                             SpawnOtherSitesInfo,
                             [],
                             [],
                             infinity);


find_or_spawn_peers_loop(ReplicationGroupName, 
                         ThisSiteId,
                         ThisSiteReplicationGroupPid, 
                         SpawnOtherSitesInfo,
                         [], 
                         NextPassSiteIds,
                         _) when NextPassSiteIds /= [] ->

    %% All ThisPassSiteIds are exhausted -- start a new pass
    ?INFO_LOGGER("find_or_spawn_peers: starting new pass after ~w ms with ~w~n", [get(find_other_sites_period_ms), NextPassSiteIds]),
    find_or_spawn_peers_loop(ReplicationGroupName,
                             ThisSiteId,
                             ThisSiteReplicationGroupPid, 
                             SpawnOtherSitesInfo,
                             NextPassSiteIds,
                             [],
                             get(find_other_sites_period_ms));  %% wait for a while before starting


find_or_spawn_peers_loop(ReplicationGroupName, 
                         ThisSiteId,
                         ThisSiteReplicationGroupPid, 
                         SpawnOtherSitesInfo,
                         ThisPassSiteIds,
                         NextPassSiteIds,
                         ReceiveTimeout) ->


    %% Invariant: ReceiveTimeout == infinity iff there is some work to do
    if (ReceiveTimeout == infinity) and (ThisPassSiteIds == []) and (NextPassSiteIds == [])  -> ok;
       (ReceiveTimeout /= infinity) and ((ThisPassSiteIds /= []) or (NextPassSiteIds /= [])) -> ok;
       true -> erlang:error({receive_timeout_invalid, ReceiveTimeout, ThisPassSiteIds, NextPassSiteIds})
    end,

    %% As our replication_group process sends us a 'stop_trying_site' message every time 
    %% it receives a site update, a constant stream of those messages would starve the 
    %% retry mechanism if we re-used the same receive timeout each time.
    %% so we adjust the the timeout each time.
    StartReceiveTime = now(),

    ?INFO_LOGGER("find_or_spawn_peers: ThisPassSiteIds = ~w, NextPassSiteIds = ~w, ReceiveTimeout = ~w~n", [ThisPassSiteIds, NextPassSiteIds, ReceiveTimeout]),
    receive

        {find_or_spawn_site, OtherSiteId} -> 
            ?INFO_LOGGER("find_or_spawn_peers: received find_or_spawn_site ~w~n", [OtherSiteId]),
            %% This generally means that we were monitoring a site and we got a DOWN message.
            %% Note that we start monitoring sites when they subscribe to us, not just 
            %% when we subscribe to them.  Therefore it is possible that we get this 
            %% message for a site that we are already looking for, so we need to prevent
            %% duplicate entries.
            %% We immediately try the site (if it crashed we want to respawn it 
            %% as quickly as possible), and then continue with this pass
            %% (placing the new site at the end of NextPassSiteIds).
            %% We also reset NumAttempts for that site.

            find_or_spawn_site(ReplicationGroupName, 
                               ThisSiteId,
                               ThisSiteReplicationGroupPid, 
                               SpawnOtherSitesInfo,
                               OtherSiteId),

            find_or_spawn_peers_loop(ReplicationGroupName,
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     SpawnOtherSitesInfo, 
                                     lists:keydelete(OtherSiteId, 1, ThisPassSiteIds),
                                     lists:keydelete(OtherSiteId, 1, NextPassSiteIds) ++ [{OtherSiteId, 0}],
                                     case ReceiveTimeout of 
                                         infinity -> get(find_other_sites_period_ms);                  % start retries
                                         0        -> 0;                                                % in the middle of a pass (continue)
                                         _        -> reduced_timeout(ReceiveTimeout, StartReceiveTime) % waiting for next pass (adjust time)
                                     end);

        {stop_trying_site, OtherSiteId} -> 
            ?INFO_LOGGER("find_or_spawn_peers: received stop_trying_site ~w~n", [OtherSiteId]),
            find_or_spawn_peers_loop(ReplicationGroupName,
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     SpawnOtherSitesInfo,
                                     lists:keydelete(OtherSiteId, 1, ThisPassSiteIds),
                                     lists:keydelete(OtherSiteId, 1, NextPassSiteIds),
                                     case ReceiveTimeout of 
                                         infinity -> infinity;
                                         0        -> 0;                                                  % in the middle of a pass (continue)
                                         _        -> reduced_timeout(ReceiveTimeout, StartReceiveTime)   % waiting for next pass (adjust time)
                                     end);

        {forget_other_site, OtherSiteId} -> 
            ?INFO_LOGGER("find_or_spawn_peers: received forget_other_site ~w~n", [OtherSiteId]),
            find_or_spawn_peers_loop(ReplicationGroupName,
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     case SpawnOtherSitesInfo of
                                         {AllSites, ConfigProplist} ->
                                             {[OtherSite || {SiteId, _, _, _} = OtherSite <- AllSites, SiteId /= OtherSiteId], 
                                              ConfigProplist};
                                         _ -> 
                                             SpawnOtherSitesInfo
                                     end,
                                     lists:keydelete(OtherSiteId, 1, ThisPassSiteIds),
                                     lists:keydelete(OtherSiteId, 1, NextPassSiteIds),
                                     case ReceiveTimeout of 
                                         infinity -> infinity;
                                         0        -> 0;                                                 % in the middle of a pass (continue)
                                         _        -> reduced_timeout(ReceiveTimeout, StartReceiveTime)  % waiting for next pass (adjust time)
                                     end);

        stop -> 
            ?INFO_LOGGER("find_or_spawn_peers exiting gracefully~n");

        Other -> 
            exit({find_or_spawn_peers, {unexpected_msg, Other}})

    after
        ReceiveTimeout ->
            
            %% As timeout was not infinity there must be some work to do.
            [ {SiteToFindId,  NumAttempts} | TailThisPassSiteIds ] = ThisPassSiteIds,

            %% Try the next site.
            %%
            %% Even if successful, this we do NOT drop SiteId from the work queues as 
            %% 'success' isn't achieved until we have received the first status update 
            %% from the found/spawned site (at which point our replication_group will 
            %% send us a 'stop_trying_site' message)
            %% Until then anything might go wrong, so we keep retrying.
            %% Note: it would seem that we could stop retrying as soon as replication_group is 
            %% monitoring the found site's pid.  If replication_group gets a DOWN 
            %% message it will send us a 'find_or_spawn_site' message.
            %% However, we also begin monitoring a site if it subscribes to us
            %% -- but that doesn't mean that we are subscribed to that site yet.

            find_or_spawn_site(ReplicationGroupName, 
                               ThisSiteId,
                               ThisSiteReplicationGroupPid, 
                               SpawnOtherSitesInfo,
                               SiteToFindId),

            %% Move this SiteToFind to the end of the NextPass queue.
            %%
            %% We always set ReceiveTimeout to 0 in order to process the next site in TailThisPassSiteIds 
            %% immediately.  
            %% If TailThisPassSiteIds is empty then an earlier function clause will 
            %% move NextPassSiteIds to ThisPassSiteIds and 
            %% set ReceiveTimeout to the full between-passses timeout.

            find_or_spawn_peers_loop(ReplicationGroupName, 
                                     ThisSiteId,
                                     ThisSiteReplicationGroupPid, 
                                     SpawnOtherSitesInfo,
                                     TailThisPassSiteIds,
                                     NextPassSiteIds ++ [{SiteToFindId, NumAttempts + 1}],
                                     0)

    end.


reduced_timeout(ReceiveTimeout, StartReceiveTime) ->
    case ReceiveTimeout - ?BDB_U:now_diff_milliseconds(now(), StartReceiveTime) of
        T when T >= 0 -> T;
        _             -> 0
    end.
             

find_or_spawn_site(ReplicationGroupName, 
                   ThisSiteId,
                   ThisSiteReplicationGroupPid, 
                   SpawnOtherSitesInfo,
                   {_, SiteToFindNode} = SiteToFindId) ->

    ?INFO_LOGGER("find_or_spawn_site trying ~w~n", [SiteToFindId]),

    case rpc:call(SiteToFindNode, 
                  ?MODULE, 
                  remote_find_or_spawn_site, 
                  [ReplicationGroupName, 
                   SiteToFindId, 
                   SpawnOtherSitesInfo]) of

        {ok, SiteToFindReplicationGroupPid} ->
            ?INFO_LOGGER("find_or_spawn_peers found ~w for ~w~n", [SiteToFindReplicationGroupPid, SiteToFindId]),

            %% Point the two sites at each other (mutually subscribe)
            ok = gen_server:cast(SiteToFindReplicationGroupPid, {subscribe, ThisSiteId,   ThisSiteReplicationGroupPid}),
            ok = gen_server:cast(ThisSiteReplicationGroupPid,   {subscribe, SiteToFindId, SiteToFindReplicationGroupPid});

        Other ->
            ?INFO_LOGGER("find_or_spawn_site failed on ~w with ~w~n", [SiteToFindId, Other])

    end.

%% This helper must be exported for rpc to find it
remote_find_or_spawn_site(ReplicationGroupName, 
                          {SiteToFindName, SiteToFindNode} = SiteToFindId, 
                          SpawnOtherSitesInfo) ->

    SiteToFindNode = node(), % assert that we are spawning on the correct node

    case whereis( get_proc_name(ReplicationGroupName, SiteToFindName) ) of

        Pid when is_pid(Pid) -> 
            ?INFO_LOGGER("find_or_spawn_site found ~w for ~w~n", [Pid, SiteToFindId]),
            {ok, Pid};

        undefined -> 
            case SpawnOtherSitesInfo of

                {AllSites, ConfigProplist} ->

                    %% We should attempt to respawn the other site
                    [SiteToFind] = [Site || {SiteId, _, _, _} = Site <- AllSites, SiteId == SiteToFindId],

                    try ?MODULE:start(ReplicationGroupName, 
                                      SiteToFind, 
                                      AllSites -- [SiteToFind],  % AllSites from the perspective of SiteToFind
                                      ConfigProplist) of 

                        Pid -> 
                            ?INFO_LOGGER("find_or_spawn_site successfully spawned ~w for ~w~n", [Pid, SiteToFindId]),
                            {ok, Pid}

                    catch
                        {error, {already_started, Pid}} -> 
                            %% More than one site will try to spawn a missing site at the same time, so some of them may 
                            %% find that it does exists, even though the 'whereis' check above failed.
                            ?INFO_LOGGER("find_or_spawn_site spawned found ~w while attempting to spawh ~w~n", [Pid, SiteToFindId]),
                            {ok, Pid};

                        Error -> Error
                    end;

                disabled ->
                    %% We are not allowed to spawn the other site
                    undefined
            end
    end.


%%
%% Helpers for do_txn_on_...()
%%

forward_command_to_remote_site({SiteName, SiteNode} = SiteId, 
                               PortCoordPid, 
                               {_CommandName, _TxnFun, _NumDeadlockRetries, Timeout} = Command, 
                               From) ->

    %% Note: if we passed 'From' to the Master then it could reply directly (we wouldn't 
    %% need the call to gen_server:reply() here, and we would only need to spawn a process here 
    %% to avoid blocking this replication_group process while we send the command -- and of course 
    %% we do send messages to othersites elswhere anyway.  Of course, gen_server:call would need 
    %% to become gen_server:cast() -- or the current process will block for the reply.
    %%
    %% The above assumes that gen_server replies can be sent from a different process to that to which 
    %% the request was sent.  I have not tested that.
    %% Note that the *real* client will be monitoring the original destination 
    %% process, so if the process to which we are forwarding crashes, the client won't get a reply
    %% at all (will incur the full timeout).
    %% So that is one behavioural difference here.
    %%
    %% Note that it is possible (albeit unlikely) that we will forward more than once -- and 
    %% even in infinite cycles (if the group status is totally inconsistent across sites, in just 
    %% the right way).  This is because we forward the original command, and depend on that command 
    %% to actually agree to to execute TxnFun on that site, and not simply forward it again.  
    %% If this is a problem/risk, we could simply forward TxnFun to a terminal "do_txn_on_local_site" command
    %% (but that's just a bit more code.)
    %%
    %% I'm undecided whether to call plain spawn() or proc_lib:spawn() here -- proc_lib is certainly 
    %% slower, and do we really need the error reports for this worker (which is catch-all anyway,
    %% unless it is given malformed arguments).
    %%
    %% It's a _little_ bit clearer that we should call spawn() rather than spawn_link() here.
    %% If we did the latter then we would have to keep the pid (associated with From presumably)
    %% and handle any EXIT messages etc.  That's a _lot_ of churning state management 
    %% for very little real benefit.
    %% If the spawned process dies then the client will eventually get a timeout, and the original 
    %% replication_group server (to which the client sent the request) won't care anyway.
    %% But that should never happen, as we carefully catch all possible errors from attempting to 
    %% forward the command.

    spawn(
      fun() -> 
              ?INFO_LOGGER("forward_command_to_remote_site: starting worker ~p (forwarding to ~w, ~w)~n", [self(), SiteId, PortCoordPid]),
              gen_server:reply(
                From, 
                %% Use the same timeout (assume we have used zero time so far)
                try gen_server:call(PortCoordPid, Command, Timeout)
                catch
                    %% TODO: do we need to catch the SPECIFIC reason here (noproc, or nodedown, or noconnection, or whatever it is).
                    exit:Reason ->
                        %% OTP will cause the client to exit if the destination process is down.
                        %% That is safe here (it won't kill this site) because we are in a spawned process.
                        %% But if we don't catch it we won't reply to the client, so the client will 
                        %% block for the full timeout.
                        %% We convert it to an error return that is thrown as an exception on the 
                        %% client, so the client can retry with another site.
                        ?INFO_LOGGER("forward_command_to_remote_site: (failed to forward to ~w, ~w; returning site_is_down (reason is ~w)~n", [SiteId, PortCoordPid, Reason]),
                        {error, {bdb_replication_group_error, 
                                 {site_is_down, 
                                  {get(replication_group_name), SiteName, SiteNode},
                                  Reason}}}
                end),
              ?INFO_LOGGER("forward_command_to_remote_site: worker ~p finished~n", [self()])
      end),
    ok.


async_do_txn_on_master_local_site({_CommandName, TxnFun, NumDeadlockRetries, _Timeout}, 
                                  From) ->

    %% We don't need to send TxnFun to the port_coordinator, we can execute it ourselves and avoid the extra hop.
    %% The earlier comments about proc_lib vs plain, and spawn vs spawn_link apply here too.
    PortCoordPid = get(this_site_port_coord_pid),

    %% TODO: to avoid generating (small) garbage on every txn, only create a new MasterLocationHint 
    %% if the client didn't specify one, or if the new one would be different 
    %% (most of the time neither of those will be true).
    MasterLocationHint = #master_location_hint{site_locator = {get(replication_group_name), get(this_site_name), node()},
                                               pid          = self()},
    spawn(
      fun() -> 
              ?INFO_LOGGER("do_txn_on_master: starting worker ~p~n", [self()]),
              gen_server:reply(
                From, 
                try ?BDB_H:do_txn_no_throw(PortCoordPid, TxnFun, NumDeadlockRetries) of 
                    {ok, Result}            -> {ok, {Result, MasterLocationHint}};
                    {error, _} = ErrorReply -> ErrorReply
                catch
                    %% do_txn_no_throw is NOT completely 'no throw'
                    %% - it doesn't catch crashes that might be due to bugs.
                    %% We do so here otherwise the client will be left waiting for a full gen_server timeout.
                    Class:Reason -> 
                        {error, {bdb_helper_error, {internal_error, Class, Reason, erlang:get_stacktrace()}}}                        
                end),
              ?INFO_LOGGER("do_txn_on_master: worker ~p finished~n", [self()])
      end),
    ok.


async_do_txn_on_any_site_local_site({_CommandName, TxnFun, NumDeadlockRetries, _Timeout}, 
                                    From) ->
    %% Execute TxnFun on local site 
    %% See comments in do_txn_on_master -- the only difference here is that we don't return a new MasterLocationHint.
    PortCoord = get(this_site_port_coord_pid),
    proc_lib:spawn(
      fun() -> 
              ?INFO_LOGGER("do_txn_on_any_site_local_site: starting worker ~p~n", [self()]),
              gen_server:reply(
                From, 
                try ?BDB_H:do_txn_no_throw(PortCoord, TxnFun, NumDeadlockRetries) of 
                    {ok, Result}            -> {ok, Result};
                    {error, _} = ErrorReply -> ErrorReply
                catch
                    Class:Reason -> 
                        {error, {bdb_helper_error, {internal_error, Class, Reason, erlang:get_stacktrace()}}}                        
                end),
              ?INFO_LOGGER("do_txn_on_any_site_local_site: worker ~p finished~n", [self()])
      end),
    ok.


%%
%% Handle local Sleepycat events
%%

handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_CLIENT}, State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    assert_state_invariants(State),

    ?INFO_LOGGER("sleepycat event: Local site is a replica, and might not be in-sync yet~n"),

    NewState = case State#state.current_master of 

                   this_site ->
                       ?INFO_LOGGER("sleepycat event: This site was master and got DB_EVENT_REP_CLIENT (**downgrade**)~n"),
                       State#state{current_master   = unknown,
                                   this_site_status = #site_status{status            = {replica, not_in_sync},
                                                                   last_updated_time = now()}};

                   _ ->
                       %% We have no reason to believe that the master has changed.
                       State#state{this_site_status = #site_status{status            = {replica, not_in_sync},
                                                                   last_updated_time = now()}}
               end,

    assert_state_invariants(NewState),

    publish_state_change(State, NewState),
    NewState;


handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_NEWMASTER}, State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    assert_state_invariants(State),

    ?INFO_LOGGER("sleepycat event: This site has discovered a new master and is beginning sync-up (OLD master was ~w)~n", [State#state.current_master]),

    %% TODO: At the C level, the site_id of the NEW master is supplied.  That's not yet exposed here.
    %%       The main reason being that repmgr allocates the site_id's internally, so we wouldn't know which site it is anyway.
    %%
    %% We cannot therefore make any change to current_master, as we may have learned about the new master from that site
    %% before we ourselves got NEWMASTER

    NewState = case State#state.current_master of 

                   this_site ->
                       ?ERROR_LOGGER("sleepycat event: This site got DB_EVENT_REP_NEWMASTER while it is master.  Exiting as there may be multiple masters~n"),
                       exit(this_site_got_NEWMASTER_while_it_is_master);

                   _ ->
                       State#state{this_site_status = #site_status{status            = {replica, not_in_sync},
                                                                   last_updated_time = now()}}
               end,

    assert_state_invariants(NewState),

    publish_state_change(State, NewState),
    NewState;


handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_MASTER}, State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    assert_state_invariants(State),

    ?INFO_LOGGER("sleepycat event: This site is now the master (OLD master was ~w)~n", [State#state.current_master]),

    NewState = case State#state.current_master of 

                   this_site ->
                       ?WARNING_LOGGER("sleepycat event: This site got DB_EVENT_REP_MASTER but is already master~n"),
                       State;

                   {_, _} = ExMasterSiteId ->
                       %% We have been elected as new master but we still think we are connected to an existing master.
                       %% This doesn't necessarily mean more than one master -- e.g. we might be elected as the new master 
                       %% before we get the DOWN message for the ex-master (unlikely but possible).  
                       %% So we change the ex-masters status to '{replica, not_in_sync}', as the DOWN message must arrive 
                       %% some time (and if the master simply downgraded rather than disconnecting that is also the correct new state).
                       {value, {_, ExMasterSiteStatus}} = lists:keysearch(ExMasterSiteId, 1, State#state.other_sites_status),

                       State2 
                           = State#state{this_site_status   = #site_status{status            = master,
                                                                           last_updated_time = now()},
                                         current_master     = this_site,
                                         other_sites_status = lists:keyreplace(ExMasterSiteId, 
                                                                               1, 
                                                                               State#state.other_sites_status, 
                                                                               {ExMasterSiteId, 
                                                                                ExMasterSiteStatus#site_status{status = {replica, not_in_sync}}})},
                       assert_state_invariants(State2),
                       State2;

                   unknown ->
                       State2 = notify_clients_of_discovered_master(           % returns State2 with empty notify_clients list
                                  State#state{this_site_status = #site_status{status            = master,
                                                                              last_updated_time = now()},
                                              current_master   = this_site}),

                       assert_state_invariants(State2),
                       State2

               end,

    publish_state_change(State, NewState),
    NewState;


handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_REP_STARTUPDONE}, State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    assert_state_invariants(State),    

    ?INFO_LOGGER("sleepycat event: This site (state was ~w) but is is now in-sync with the master (master = ~w)~n", [State#state.this_site_status, State#state.current_master]),

    NewState = case ?get_this_site_status(State) of 

                   {replica, not_in_sync} ->
                       State#state{this_site_status = #site_status{status            = {replica, in_sync},
                                                                   last_updated_time = now()}};

                   {replica, in_sync} ->
                       ?WARNING_LOGGER("sleepycat event: This site got DB_EVENT_REP_STARTUPDONE but was already in_sync~n"),
                       State;

                   unknown ->
                       ?WARNING_LOGGER("sleepycat event: This site got DB_EVENT_REP_STARTUPDONE but was in 'unknown'~n"),
                       State#state{this_site_status = #site_status{status            = {replica, in_sync},
                                                                   last_updated_time = now()}};

                   master ->
                       ?ERROR_LOGGER("sleepycat event: This site got DB_EVENT_REP_STARTUPDONE but was in 'master'~n"),
                       exit(got_STARTUPDONE_while_master)
               end,

    assert_state_invariants(NewState),

    publish_state_change(State, NewState),
    NewState;


handle_this_site_sleepycat_event({PortCoordName, bdb_DB_EVENT_WRITE_FAILED}, State) ->

    PortCoordName = get(this_site_port_coord_name),  % assert
    assert_state_invariants(State),

    ?ERROR_LOGGER("sleepycat event: This site's BDB failed to write to disk~n"),
    %% TODO: we might be a replica or a master.;
    %% If we are master then we should crash to kill the port_coordinator and force an election.
    exit(this_site_bdb_failed_to_write_to_disk).


publish_state_change(OldState, NewState) ->

    if  NewState#state.this_site_status /= OldState#state.this_site_status ->

            %% Status has changed so send it to all other replication_group sites that we know of

            ThisSiteId = get(this_site_id),
            lists:foreach(fun({OtherSiteId, #site_status{replication_group_pid = OtherSiteReplicationGroupPid}}) 
                             when is_pid(OtherSiteReplicationGroupPid) ->

                                  ?INFO_LOGGER("publish_state_change: ~w sending to ~w (~w)~n", [?get_this_site_status(NewState), OtherSiteId, OtherSiteReplicationGroupPid]),
                                  send_site_status(ThisSiteId, NewState, OtherSiteReplicationGroupPid);

                             ({OtherSiteId, OtherSiteStatus}) ->
                                  ?INFO_LOGGER("publish_state_change: ~w NOT sending to ~w as replication_group_pid is unknown ~w~n", [?get_this_site_status(NewState), OtherSiteId, OtherSiteStatus])
                          end,
                          NewState#state.other_sites_status);

        true ->
            ?INFO_LOGGER("publish_state_change: no change so not sending~n"), 
            ok
    end.


send_site_status(ThisSiteId, State, DestinationReplicationGroupPid) ->

    ok = gen_server:cast(DestinationReplicationGroupPid, 
                         {other_site_status, ThisSiteId, self(), ?get_this_site_status(State), ?get_this_site_status_time(State)}).


%%
%% Handle state-changes from other replication_group servers
%%

handle_other_site_status_update(OtherSiteId, OldOtherSiteStatus, {replica, _} = NewStatus, NewUpdatedAtTime, State) ->

    State#state{other_sites_status = lists:keyreplace(OtherSiteId, 
                                                      1, 
                                                      State#state.other_sites_status, 
                                                      {OtherSiteId, OldOtherSiteStatus#site_status{status            = NewStatus,
                                                                                                   last_updated_time = NewUpdatedAtTime}})};

handle_other_site_status_update(OtherSiteId, OldOtherSiteStatus, master = NewStatus, NewUpdatedAtTime, State) ->

    case State#state.current_master of 

        this_site ->
            ?ERROR_LOGGER("handle_other_site_status_update: Site ~w claims to be master when this site is master~n", [OtherSiteId]),
            exit({other_site_claims_to_be_master_when_this_site_is_master, OtherSiteId});

        OtherSiteId ->
            ?WARNING_LOGGER("handle_other_site_status_update: Site ~w claims to be master when it is already master~n", [OtherSiteId]),
            State;

        {_, _} = ExMasterSiteId ->
            ?INFO_LOGGER("handle_other_site_status_update: Master has moved from ~w to ~w~n", [ExMasterSiteId, OtherSiteId]),

            %% We need to change OtherSiteId to master and ExMasterSiteId to {replica, not_in_sync}
            %% (when changing ExMaster's status we need to leave last_updated_time **alone** -- we 
            %% don't have the current time for that site, and future events from that site must sort later than this.).
            {value, {_, ExMasterSiteStatus}} = lists:keysearch(ExMasterSiteId, 1, State#state.other_sites_status),

            NewState =
                State#state{current_master     = OtherSiteId,
                            other_sites_status = ?BDB_U:lists_multikeyreplace(
                                                    [{OtherSiteId, 
                                                      {OtherSiteId, OldOtherSiteStatus#site_status{status            = NewStatus,
                                                                                                   last_updated_time = NewUpdatedAtTime}}}, 
                                                     {ExMasterSiteId, 
                                                      {ExMasterSiteId, ExMasterSiteStatus#site_status{status = {replica, not_in_sync}}}}],
                                                    1, 
                                                    State#state.other_sites_status)},
            assert_state_invariants(NewState),
            NewState;

        unknown ->
            ?INFO_LOGGER("handle_other_site_status_update: Master discovered ~w (was unknown)~n", [OtherSiteId]),

            NewState =
                notify_clients_of_discovered_master(         % returns the passed State but with an empty notify_clients list
                  State#state{current_master     = OtherSiteId,
                              other_sites_status = lists:keyreplace(
                                                     OtherSiteId, 
                                                     1, 
                                                     State#state.other_sites_status,
                                                     {OtherSiteId, OldOtherSiteStatus#site_status{status            = NewStatus,
                                                                                                  last_updated_time = NewUpdatedAtTime}})}),
            assert_state_invariants(NewState),
            NewState
    end.

%%
%% Assert invariants
%%

assert_state_invariants(#state{current_master = unknown} = State) ->
    assert_this_site_status_invariants(State#state.this_site_status),
    ?get_this_site_status(State) /= master,
    ?get_this_site_status(State) /= {replica, in_sync},
    assert_all_other_sites_have_status_entry(State),
    [] = [X || {_SiteId, #site_status{status = master}} = X <- State#state.other_sites_status];
%% TODO: is this a valid assert?
%% If we get DOWN from the master then we set all replicas to not_in_sync (until we get STARTUPDONE messages from them).
%% But if another site discovers the master and syncs up before we discover the master then this assert might trigger
%%     [] = [X || {_SiteId, #site_status{status = {replica, in_sync}}} = X <- State#state.other_sites_status];

assert_state_invariants(#state{current_master = this_site} = State) ->
    assert_this_site_status_invariants(State#state.this_site_status),
    ?get_this_site_status(State) == master,
    assert_all_other_sites_have_status_entry(State),
    [] = [X || {_SiteId, #site_status{status = master}} = X <- State#state.other_sites_status],
    [] = State#state.notify_clients;

assert_state_invariants(#state{} = State) ->   %% current master is another site
    assert_this_site_status_invariants(State#state.this_site_status),
    ?get_this_site_status(State) /= master,
    assert_all_other_sites_have_status_entry(State),
    %%?INFO_LOGGER("Current master must be another site ~w~n", [State#state.other_sites_status]),
    [_ExactlyOne] = [X || {_SiteId, #site_status{status = master}} = X <- State#state.other_sites_status],
    [] = State#state.notify_clients.


assert_this_site_status_invariants(#site_status{status = unknown} = ThisSiteStatus) ->
    throw({this_site_unknown, ThisSiteStatus});
assert_this_site_status_invariants(#site_status{status                = _,
                                                last_updated_time     = Time,
                                                replication_group_pid = undefined,
                                                monitor_ref           = undefined} = ThisSiteStatus) when is_tuple(Time), size(Time) == 3 ->
    assert_site_status_invariants(ThisSiteStatus).

assert_all_other_sites_have_status_entry(#state{} = State) ->
    lists:foreach(fun({FindSiteId, _, _, _}) ->
                          {value, {FindSiteId, SiteStatus}} = lists:keysearch(FindSiteId, 1, State#state.other_sites_status),
                          assert_site_status_invariants(SiteStatus)
                  end,
                  State#state.other_sites),
    %% and the inverse...
    lists:foreach(fun({FindSiteId, SiteStatus}) ->
                          assert_site_status_invariants(SiteStatus),
                          %% some restrictions for other site's status
                          case SiteStatus of 
                              #site_status{replication_group_pid = undefined} -> throw(other_site_status_cannot_have_undefined);
                              _                                               -> ok
                          end,
                          [_ExactlyOne] = [SiteId || {SiteId, _, _, _} <- State#state.other_sites, SiteId == FindSiteId]
                  end,
                  State#state.other_sites_status).
    
assert_site_status_invariants(#site_status{} = SiteStatus) ->
    assert_valid_site_status_value(SiteStatus#site_status.status),
    %% if status is unknown the last_updated_time must be unknown and vice-versa
    case SiteStatus of 
        #site_status{status            = unknown,
                     last_updated_time = unknown}                    -> ok;
        #site_status{last_updated_time = Time} when Time =/= unknown -> ok
    end,
    %% if replication_group_pid is not unknown then monitor_ref must not be unknown, and vice-versa
    case SiteStatus of 
        #site_status{replication_group_pid = undefined,
                     monitor_ref           = undefined} -> ok; % must be this site's status
        #site_status{replication_group_pid = unknown,
                     monitor_ref           = unknown} -> ok;
        #site_status{replication_group_pid = Pid,
                     monitor_ref           = Ref} when is_pid(Pid) and is_reference(Ref) -> ok
    end.

assert_valid_site_status_value(unknown)          -> ok;
assert_valid_site_status_value(master)                 -> ok;
assert_valid_site_status_value({replica, not_in_sync}) -> ok;
assert_valid_site_status_value({replica, in_sync})     -> ok.


%% Utilities 
%% TODO: move to utilities.erl and change name to something better (it's only called via a macro anyway)

logger(String) ->
    logger(String, []).
logger(String, Args) ->
    case process_info(self(), registered_name) of
        {registered_name, ProcessName} ->
            io:format("~w: " ++ String, [ProcessName | Args]);
        [] ->
            case get('$ancestors') of
                undefined ->
                    io:format("~w: " ++ String, [self() | Args]);
                Ancestors ->
                    %% This requires that we only create support processes with proc_lib:spawn_link
                    io:format("~w: support process ~w: " ++ String, [hd(Ancestors) | [self() | Args]])
            end
    end.
