%%%-------------------------------------------------------------------
%%% File    : berkeley_db_replication_group_helpers.erl
%%% Author  : Chris Newcombe <>
%%% Description : 
%%%
%%% Helper functions to make it reasonably convient for applications to 
%%% access replication groups without requiring a global directory 
%%% service to find sites/masters etc.
%%%
%%% Problem statement:
%%%
%%% berkeley_db_replication_group:do_txn_on_master() will find the 
%%% master if it moves, but the caller must first find _some_ site in 
%%% the replication group, in order to call do_txn_on_master at all.
%%% Having a MasterLocationHint is no use if it is the master that fails.
%%%
%%% Also, clients might want to run a txn on a replication_group that 
%%% does not currently have a site on the current node (either the 
%%% site has failed, or it really is a different cluster).
%%% 
%%% Also (in the latter case), clients might want to run a read-only
%%% txn on _any_ site.  They would presumeably  prefer to use a local
%%% site if one exists, but would like to select a random site 
%%% otherwise.
%%%
%%% If a site fails (cannot be contacted), the caller would like to avoid
%%% trying it again for some period.  But if all sites appear to fail,
%%% we probably _do_ want to try them again (rather than waiting for 
%%% some timeout period).
%%%
%%% Solution:
%%%
%%% The caller can create a 'caller_state' object (one per replication group)
%%% and pass it to do_txn_on_master() and do_txn_on_any_site().
%%%
%%% do_txn_on_master() will use a master-location-hint (cached in the caller-state).
%%% If the master dies it will try other random sites until it finds it again,
%%% or until all sites have been tried, or until a timeout is reached.
%%%
%%% do_txn_on_any_site{} picks any site that is believed to be 'up', 
%%% but prefers local sites first, if they exist. 
%%%
%%% Both functions will tell the caller if the caller_state has changed, so
%%% that the caller knows whether to update it's own cache of caller_state 
%%% (e.g. store it in an ETS table).
%%% Of course, if multiple callers are using a caller_state object,
%%% they risk using slightly stale copies.  But that should be fine -- 
%%% merely a little inefficient, not fatal.  If the objects are 
%%% stored in an ETS table, the 'default' last-writer-wins policy
%%% should be pretty effective (i.e. we don't need to attempt to 
%%% merge caller_state objects, or try to identify the 'best' or 
%%% 'latest' one). 
%%%
%%% Created : 26 Oct 2006 by Chris Newcombe <>
%%%
%%% TODO: 
%%%
%%% - do_txn_on_any_site wants to do it on **in_sync** sites.
%%%   So instead of using rpc:call, there should be a method
%%%   ?BDB_RG:do_txn_on_any_in_sync_site().
%%%   We pick a random site (prefering local sites), 
%%%   and call the method on that site.
%%%   If that site happens to not be in-sync, in picks a site 
%%%   that is and forwards the request.
%%%   It should return an indication of whether the _original_ 
%%%   site was in/not-in sync.  caller_state should keep that
%%%   so up_site_locators should be in_sync_site_locators.
%%%   and not_in_sync flags should be moved to down_sites and polled.
%%%
%%% - In pick_random_not_down_site, if all sites are down and we 
%%%   try to contact one (random) site, then if we _do_ contact it
%%%   we want to get the current group state (or all traffic will be 
%%%   sent to that single site until the others emerge from the timeout period).
%%%   So we need the same mechanism as the next point (?BDB_RG:get_current_caller_state()).
%%%   And that means we don't use rpc:call here either -- we call the above 
%%%   method instead.
%%%
%%% - When replication_group supports removing sites, caller_state will drift out of 
%%%   sync with the real site membership.  
%%%   So whenever we fail to connect to a site, and later successfully connect to a different site,
%%%   we should ask that site for the new membership. 
%%%   i.e. There should be a ?BDB_RG:get_current_caller_state() method, which 
%%%   would be a smarter version of make_initial_caller_state().
%%%   It could be invoked via do_txn_on_any_site() using the current (flawed) ?CS,
%%%   to get a better one.
%%%   Either than, or clients need to hook into replication_group events 
%%%   -- and even exposing those to the clients is a problem (which sites send them?)
%%%   or every call to every ?BDB_RG method all returns the current group status 
%%%   - obviously far too heavyweight.
%%%
%%% - Is a timeout the correct mechanism to limit these fumctions?
%%%   e.g. should caller_state contain a max_num_sites_to_try as well? 
%%%
%%% - In pick_random_not_down_site
%%%   As RgCallerState is presumably rarely/never transmitted across nodes, 
%%%   we could cache LocalUpSiteLocators and avoid the list comprehension here.
%%%
%%% - Ideally we would have cached pids for each UpSiteLocator
%%%   (like master_location_hint) does.
%%%   That would save a call to whereis (the first clause of do_txn_with_retry_on_deadlock_no_throw),
%%%   but that of course risks flagging a site as down if it has crashed since we got the cached pid 
%%%   -- even if it is back up now.  (Using process names fixes that.)
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_replication_group_helpers).

%% API
-export([make_initial_caller_state/3,
         update_caller_state_from_group_status/2,
         get_replication_group_site_pid/1,
         get_local_replication_group_site_pid/2,
         pick_random_not_down_site/1,       % if all sites have failed recently, this function will attempt to connect to sites to see if they are up.
         pick_random_not_down_site_fast/1,  % may give false negative results, as if all sites have failed recently, this function won't attempt to connect to any sites to see if they are up.
         do_txn_on_master/4,
         do_txn_on_any_site/4
        ]).

%% Internal API
%% Exported for use with rpc:call
-export([]).

-include("berkeley_db_replication_group_helpers.hrl").
-include("berkeley_db_replication_group_internal.hrl").

-define(CS, berkeley_db_replication_group_caller_state).

-define(BDB,    berkeley_db_drv).
-define(BDB_U,  berkeley_db_utilities).
-define(BDB_H,  berkeley_db_helpers).
-define(BDB_PC, berkeley_db_port_coordinator).
-define(BDB_RG, berkeley_db_replication_group).


-define(random_list_element(List),
    lists:nth( random:uniform(length(List)), List)).


% Debug helpers (poor-man's tracing)
-define(AT, io:format("~nAt line ~w~n", [?LINE])).   %% remove with regex    ^ *\?AT, *^J


%%====================================================================
%% API
%%====================================================================

%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%   AllSites is the same list required by ?BDB_RG:start_sites()
%%--------------------------------------------------------------------

make_initial_caller_state(ReplicationGroupName, AllSites, IgnoreDownSitesForMilliseconds) ->

    AllSiteLocators = lists:map(
                        fun({{SiteName, SiteNode}, _BdbDataDir, _BDBHAHostNamePort, _SiteSpecificConfig}) ->
                                {ReplicationGroupName, SiteName, SiteNode}
                        end,
                        AllSites),

    %% TODO: for now we just set master_location_hint to a random SiteLocator.
    %% (That's valid -- all public functions that accept a MasterLocationHint also accept a SiteLocator)
    %%
    %% We could also do this:
    %%   master_location_hint = #master_location_hint(site_locator = MasterSiteLocator,
    %%                                                pid          = get_replication_group_site_pid(MasterSiteLocator)},
    %% but then we have to handle the case when it is down.
    %% Or we could even set pid to 'unknown' or something 
    %% (but that would require more code in the functions that accept master location hints)

    MasterSiteLocator = ?random_list_element(AllSiteLocators),

    #?CS{replication_group_name   = ReplicationGroupName,
         all_site_locators        = AllSiteLocators,         
         master_location_hint     = MasterSiteLocator,
         up_site_locators         = AllSiteLocators,   % we assume that everything is running
         down_sites               = [],
         ignore_down_sites_for_ms = IgnoreDownSitesForMilliseconds}.


update_caller_state_from_group_status(RgCallerState, GroupStatus) ->

    ReplicationGroupName = RgCallerState#?CS.replication_group_name,

    UpSiteLocators = [{ReplicationGroupName, SiteName, SiteNode} || {{SiteName, SiteNode}, SiteStatus} <- GroupStatus, SiteStatus =/= unknown],
    
    %% Note: this resets the ignore_down_sites_for_ms time of any site that is already known to be down.
    %% (i.e. sucn a site might be about to recover).
    %% We could extract the times from RgCallerState#?CS.down_sites, but I don't think it is worth it really.
    Now = now(),
    DownSites = [{{ReplicationGroupName, SiteName, SiteNode}, Now} || {{SiteName, SiteNode}, SiteStatus} <- GroupStatus, SiteStatus =:= unknown],

    MasterSiteLocator
        = case [SiteId || {SiteId, master} <- GroupStatus] of
              [] ->
                  %% no site is currently master, at least according to the site that provided GroupStatus.
                  %% It is still best to specify _some_ up site.
                  %% We prefer a site on *this* node if there is one.
                  %% Otherwise we prefer any other site that is running.
                  case [SiteLocator || {_, _, SiteNode} = SiteLocator <- UpSiteLocators, SiteNode == node()] of
                      [] -> 
                          %% No local site, so consider them all
                          ?random_list_element(UpSiteLocators);

                      LocalUpSiteLocators ->
                          ?random_list_element(LocalUpSiteLocators)
                  end;

              [{SiteName, SiteNode}] ->
                  {ReplicationGroupName, SiteName, SiteNode}
          end,

    RgCallerState#?CS{master_location_hint = #master_location_hint{site_locator = MasterSiteLocator,
                                                                   pid          = get_replication_group_site_pid(MasterSiteLocator)},
                      up_site_locators     = UpSiteLocators,
                      down_sites           = DownSites}.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%% Parameters:  SiteLocator
%% Returns:     site_not_running | Pid
%%--------------------------------------------------------------------

get_replication_group_site_pid({ReplicationGroupName, SiteName, _SiteNode}) when _SiteNode == node() ->
    %% Optimization for when the site is on the current node
    get_local_replication_group_site_pid(ReplicationGroupName, SiteName);

get_replication_group_site_pid({ReplicationGroupName, SiteName, SiteNode}) ->
    
    try rpc:call(SiteNode, ?MODULE, get_local_replication_group_site_pid, [ReplicationGroupName, SiteName])
    catch
        {badrpc, nodedown} ->  site_not_running
    end.

%% This is passed to rpc:call for execution on the remote-node.
get_local_replication_group_site_pid(ReplicationGroupName, SiteName) ->

    case whereis(?BDB_RG:get_proc_name(ReplicationGroupName, SiteName)) of
        undefined -> site_not_running;
        Pid       -> case erlang:is_process_alive(Pid) of
                         true  ->  Pid;
                         false ->  site_not_running
                     end
    end.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%           This prefers sites on the current node, if there are any.
%% Returns
%%           {SiteLocator, {caller_state_has/not_changed, RgCallerState}}
%% Throws:
%%           {berkeley_db_replication_group_error, {no_sites_available, NewRgCallerState}}
%%--------------------------------------------------------------------

pick_random_not_down_site(RgCallerState) ->

    try pick_random_not_down_site_fast(RgCallerState)
    catch
        {berkeley_db_replication_group_helper_error, {wrapped_caller_state_has_changed, 
                                                      {berkeley_db_replication_group_error, {no_sites_available, _}}, 
                                                      #?CS{down_sites = DownSites} = NewRgCallerState}} = Error ->

            %% Here we believe that *all* sites are down, as they have all failed to respond 
            %% to attempts within the last ignore_down_sites_for_ms period.
            %% But that timeout is probably conservative. 
            %% To avoid waiting for the full timeout before being able to access the group again,
            %% we attempt to check a one of the sites to see if they are already running.
            %%
            %% If we do manage to contact a site that is up, we want to get the current group state 
            %% from the replication_group server (and convert it to a caller_state).
            %% Otherwise _all_ 'any site' traffic will be sent to that single site, until the others emerge from 
            %% the timeout period.
            %%
            %% Currently we only check one site (even if that site really is still down) to 
            %% minimize overhead per call.  As the site is randomly chosen, if this function is called 
            %% frequently it will find a site in reasonable time.
            %% The alternative is to check more sites -- but if they really are all down we may 
            %% incur significant overhead (connection attempts) until one of them comes back up.
            %% If the caller really does want to know the 'live' status of the whole group then 
            %% then can map is_site_running() over #CS.site_locators.
            %%
            %% We choose a randome site rather than, say, the site that has been down for longest
            %% as this call will be made frequently and repeatedly. We obviously don't want to make the 
            %% same choice each time (or add yet more state to record past choices).

            {ProbeSiteLocator, _} = ?random_list_element(DownSites),

            case get_replication_group_site_pid(ProbeSiteLocator) of

                site_not_running ->
                    throw(Error);   % rethrow same exception

                _Pid ->
                    %% We found a site that is up
                    %% Ask it for the current group state, and convert that to caller_state.
                    try ?BDB_RG:get_group_status(ProbeSiteLocator) of
                        GroupStatus ->  
                            {ProbeSiteLocator, {caller_state_has_changed,
                                                update_caller_state_from_group_status(NewRgCallerState, GroupStatus)}}
                    catch
                        %% It appears that ProbeSiteLocator site really is not up after all (e.g. may have just crashed).
                        %% Rethrow the original error.
                        _ -> throw(Error)
                    end
            end
    end.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%           This variant does not attempt to connect to any down site (to see if it is up again yet).
%% Returns
%%           {SiteLocator, {caller_state_has/not_changed, RgCallerState}}
%% Throws:
%%           {berkeley_db_replication_group_error, {no_sites_available, NewRgCallerState}}
%%--------------------------------------------------------------------

pick_random_not_down_site_fast(#?CS{up_site_locators         = UpSiteLocators,
                                    down_sites               = DownSites,
                                    ignore_down_sites_for_ms = IgnoreDownSitesForMs} = RgCallerState) ->

    %% First resurrect any dowm_sites that haven't been tried for IgnoreDownSitesForMs.
    Now = now(),

    {NewUpSiteLocators, StillDownSites, CallerStateStatus} 
        = lists:foldl(
            fun({SiteLocator, TimeLastTried} = DownSite, {NewUpSiteLocators, StillDownSites, CallerStateStatus}) ->

                    case ?BDB_U:now_diff_milliseconds(Now, TimeLastTried) of

                        T when T > IgnoreDownSitesForMs ->
                            %% resurrect this site
                            {[SiteLocator|NewUpSiteLocators], StillDownSites, caller_state_has_changed};

                        _ ->
                            %% this site is (probably) still down
                            {NewUpSiteLocators, [DownSite|StillDownSites], CallerStateStatus}
                    end
            end,
            {UpSiteLocators, [], caller_state_not_changed},
            DownSites),

    case NewUpSiteLocators of 
        [] ->
            %% We want to pass NewRgCallerState to the caller.  
            %% They are expected to catch this, cache the new caller state, and rethrow the underlying exception
            %% The exception is:  {Tag, Reason} 
            %% where           Tag is berkeley_db_replication_group_helper_error
            %%              Reason is {wrapped_caller_state_has_changed, UnderlyingException, NewRgCallerState}
            %% UnderlyingException is {berkeley_db_replication_group_error, {no_sites_available, ReplicationGroupName}
            throw({berkeley_db_replication_group_helper_error, 
                   {wrapped_caller_state_has_changed,
                    {berkeley_db_replication_group_error, {no_sites_available, RgCallerState#?CS.replication_group_name}},
                    RgCallerState#?CS{up_site_locators = NewUpSiteLocators,
                                      down_sites       = StillDownSites}}});

        _ ->
            %% Pick a random site from NewSiteLocators

            %% Optimization::prefer a site local to this node, if there are any
            %% TODO: as RgCallerState is presumably rarely/never transmitted across nodes, 
            %% we could cache LocalUpSiteLocators and avoid the list comprehension here.
           SiteLocator =
                case [SiteLocator || {_, _, SiteNode} = SiteLocator <- NewUpSiteLocators, SiteNode == node()] of
                    [] -> 
                        %% No local site, so consider them all
                        ?random_list_element(UpSiteLocators);

                    LocalUpSiteLocators ->
                        ?random_list_element(LocalUpSiteLocators)
                end,

            %% We don't change master_location_hint as it might be fine.
            {SiteLocator, {CallerStateStatus, 
                           RgCallerState#?CS{up_site_locators = NewUpSiteLocators,
                                             down_sites       = StillDownSites}}}
    end.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%% Returns
%%           NewRgCallerState
%%--------------------------------------------------------------------

adjust_for_site_is_now_down(#?CS{master_location_hint = MLH,
                                 all_site_locators    = AllSiteLocators,
                                 up_site_locators     = UpSiteLocators,
                                 down_sites           = DownSites} = RgCallerState,
                            DownSiteLocator) ->

    NewUpSiteLocators = UpSiteLocators -- [DownSiteLocator],

    NewMLH  = case MLH of
                  #master_location_hint{site_locator = DownSiteLocator} ->
                      choose_new_mlh(NewUpSiteLocators, AllSiteLocators);
                  DownSiteLocator ->
                      choose_new_mlh(NewUpSiteLocators, AllSiteLocators);
                  _ ->
                      %% Current master location hint is not affected
                      MLH
              end,

    RgCallerState#?CS{master_location_hint = NewMLH,
                      up_site_locators     = NewUpSiteLocators,
                      down_sites           = [{DownSiteLocator, now()} | DownSites]}.


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%% Returns:
%%            {TxnFunResult, NewRgCallerState}
%%
%% Exceptions include an updated RgCallerState, which may be useful for later retries.
%%--------------------------------------------------------------------

%% public version
do_txn_on_master(#?CS{} = RgCallerState, TxnFun, NumDeadlockRetries, TxnFunTimeout) ->
    do_txn_on_master(caller_state_not_changed, RgCallerState, TxnFun, NumDeadlockRetries, TxnFunTimeout).

%% private helper
do_txn_on_master(CallerStateStatus, RgCallerState, TxnFun, NumDeadlockRetries, TxnFunTimeout) 
  when TxnFunTimeout > 0 ->

    StartTime = now(),

    CurrMLH = RgCallerState#?CS.master_location_hint,
    try ?BDB_RG:do_txn_on_master(CurrMLH, TxnFun, NumDeadlockRetries, TxnFunTimeout) of

        {TxnFunResult, NewMLH} ->
            %% If the master has moved, update the cached location
            case NewMLH of
                CurrMLH -> {TxnFunResult, {CallerStateStatus, RgCallerState}};
                _       -> {TxnFunResult, {caller_state_has_changed, RgCallerState#?CS{master_location_hint = NewMLH}}}
            end
    catch
        {bdb_replication_group_error, {site_is_down, DownSiteLocator, _}} ->

            %% adjust_for_site_is_now_down() will change master_location_hint to
            %% a different (random) site, or will throw if there are no more sites.
            do_txn_on_master(caller_state_has_changed,
                             adjust_for_site_is_now_down(RgCallerState, DownSiteLocator),
                             TxnFun,
                             NumDeadlockRetries,
                             reduced_timeout(TxnFunTimeout, StartTime));

        Exception when CallerStateStatus =:= caller_state_has_changed -> 
            %% We want to pass NewRgCallerState to the caller.  
            %% They are expected to catch this, cache the new caller state, and rethrow the underlying exception
            throw({berkeley_db_replication_group_helper_error, 
                   {wrapped_caller_state_has_changed, Exception, RgCallerState}})

        %% Note that we allow {bdb_replication_group_error, no_master} to propogate unchanged, 
        %% as the site that we contacted is running, so master_location_hint is valid and 
        %% there is no new information to add that would help with a later retry.
        %% TODO: is it worth recording the time of the earliest 'no_master' -- and reset 
        %% that time on all successful requests, so we can detect if no master has been 
        %% elected for a long time?).
    end;

%% Here TxnFunTimeout must be =< 0
do_txn_on_master(caller_state_not_changed, RgCallerState, _, _, _) ->
    throw({berkeley_db_replication_group_error, {do_txn_on_master, timeout, RgCallerState#?CS.replication_group_name}});

do_txn_on_master(caller_state_has_changed, NewRgCallerState, _, _, _) ->    
    %% We want to pass NewRgCallerState to the caller.  
    %% They are expected to catch this, cache the new caller state, and rethrow the underlying exception
    %% The exception is:  {Tag, Reason} 
    %% where           Tag is berkeley_db_replication_group_helper_error
    %%              Reason is {wrapped_caller_state_has_changed, UnderlyingException, NewRgCallerState}
    %% UnderlyingException is {berkeley_db_replication_group_error, {no_sites_available, ReplicationGroupName}
    throw({berkeley_db_replication_group_helper_error, 
           {wrapped_caller_state_has_changed,
            {berkeley_db_replication_group_error, {do_txn_on_master, timeout, NewRgCallerState#?CS.replication_group_name}},
            NewRgCallerState}}).


%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%--------------------------------------------------------------------

%% IMPORTANT TODO: 
%%  - Prefer in-sync sites to not-in-sync.
%%    This requires adding ?BDB_RG:do_txn_on_any_site(), which will do it locally if the site is in-sync,
%%    or (if the current site is not in-sync) will pick a random in-sync site (ideally other than the master).
%%
%%  - Catch if the remote port_coordinator is down (and try another site).
%%    That will require making do_txn_with_retry_on_deadlock_no_throw catch gen_server exits if the target process does not exist.
%%
%%  - tests

%% public version
do_txn_on_any_site(#?CS{} = RgCallerState, TxnFun, NumDeadlockRetries, TxnFunTimeout) ->
    do_txn_on_any_site(caller_state_not_changed, RgCallerState, TxnFun, NumDeadlockRetries, TxnFunTimeout).

%% private helper
do_txn_on_any_site(CallerStateStatus, RgCallerState, TxnFun, NumDeadlockRetries, TxnFunTimeout) ->

    StartTime = now(),

    %% This prefers sites on the current node, if there are any.
    {SiteLocator, {CallerStateStatusTmp, RgCallerState2}} = pick_random_not_down_site(RgCallerState),

    CallerStateStatus2 = case {CallerStateStatus, CallerStateStatusTmp} of
                             {caller_state_not_changed, caller_state_not_changed} -> caller_state_not_changed;
                             _                                                    -> caller_state_has_changed
                         end,

    io:format("do_txn_on_any_site: Chosen site ~p~n", [SiteLocator]),

    try ?BDB_RG:do_txn_on_any_site(SiteLocator, TxnFun, NumDeadlockRetries, TxnFunTimeout) of

        TxnFunResult ->  {TxnFunResult, {CallerStateStatus2, RgCallerState2}}

    catch
        {bdb_replication_group_error, {site_is_down, SiteLocator, _Reason}} ->
            %% adjust_for_site_is_now_down() will pick a different (random) site, 
            %% or will throw if there are no more sites.
            do_txn_on_any_site(caller_state_has_changed,
                               adjust_for_site_is_now_down(RgCallerState2, SiteLocator),
                               TxnFun,
                               NumDeadlockRetries,
                               reduced_timeout(TxnFunTimeout, StartTime));

        Exception when CallerStateStatus2 =:= caller_state_has_changed -> 
            %% We want to pass NewRgCallerState to the caller.  
            %% They are expected to catch this, cache the new caller state, and rethrow the underlying exception
            throw({berkeley_db_replication_group_helper_error, 
                   {wrapped_caller_state_has_changed, Exception, RgCallerState2}});

        Exception when CallerStateStatus2 =:= caller_state_not_changed -> 
            throw(Exception)
    end.


%%====================================================================
%% Internal functions
%%====================================================================

reduced_timeout(CurrTimeout, StartTime) ->
    case CurrTimeout - ?BDB_U:now_diff_milliseconds(now(), StartTime) of
        T when T >= 0 -> T;
        _             -> 0
    end.

%% private helper
choose_new_mlh([], AllSiteLocators) ->
    %% No 'up' sites to choose from, so simply pick a random site (maybe it will be running)
    ?random_list_element(AllSiteLocators);
choose_new_mlh(UpSiteLocators, _) ->
    ?random_list_element(UpSiteLocators).


