%%%-------------------------------------------------------------------
%%% File    : berkeley_db_sequence_server.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%% This registers under a given local or global name.
%%%
%%% It maintains a 64-bit unsigned transactional sequence number,
%%% at a local specified by BDBPortCoordinatorName,DBFilename,Key.
%%%
%%% It issues values from this counter on request, currently in 
%%% blocks of 1. (A consume_N_values API could be added if necessary.)
%%
%%% It caches ranges of this value of a specified size, and issues 
%%% values 
%%% 
%%% Obviously this should participate in a supervisor tree.
%%% It must be started *after* the relevant port_coordinator 
%%% has been started.
%%%
%%% TODO: 
%%%
%%% - Handle deadlocks
%%%
%%% - The dependency on the port_coordinator is perhaps a 
%%%   case for having the port_coordinator own these
%%%   (creating them via a port_coordinator API).
%%%
%%% - When BDB-HA is done, this should optionally talk to an
%%%   HA cluster (master), not just any old port_coordinator.
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_sequence_server).

-behaviour(gen_server).

%% API
-export([start_link/5, consume_one_value/1, stop/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% When nextValue = startNextBatchValue, we need to consume another batch
-record(state, {bdbPortHandle, 
                db, 
                dbKey, 
                nextValue, 
                startNextBatchValue,
                persistEveryNConsumedValues}).


-define(BDB,    berkeley_db_drv).
-define(BDB_PC, berkeley_db_port_coordinator).

%%-define(INFO_LOGGER,    error_logger:info_msg).
%%-define(WARNING_LOGGER, error_logger:warning_msg).
%%-define(ERROR_LOGGER,   error_logger:error_msg).
-define(INFO_LOGGER,    logger).
-define(WARNING_LOGGER, logger).
-define(ERROR_LOGGER,   logger).

logger(String) ->
    logger(String, []).
logger(String, Args) ->
    case process_info(self(), registered_name) of
        {registered_name, ProcessName} ->
            io:format("~p: " ++ String, [ProcessName | Args]);
        [] ->
            %% This requires that we only create support processes with proc_lib:spawn_link
            io:format("~p: support process ~p: " ++ String, [hd(get('$ancestors')) | [self() | Args]])
    end.



%%====================================================================
%% API
%%====================================================================

-define(THROW_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_sequence_server_error, Error})
           end).

%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%
%% Notes
%%
%% PersistEveryNConsumedValues
%% - If this is small we do frequent transactions to consume message-ids.
%%   (If it is 1 then we do a transaction every time).
%%   If it is large we do less transactions but more counter values are wasted 
%%   if the server exits (even normally), because on restart we have to start 
%%   consuming the next batch.
%%--------------------------------------------------------------------
start_link(SequenceServerName, BDBPortCoordinatorName, DBName, DBKey, PersistEveryNConsumedValues) when is_integer(PersistEveryNConsumedValues), PersistEveryNConsumedValues > 0 ->
    gen_server:start_link({local, SequenceServerName}, 
                          ?MODULE, 
                          {BDBPortCoordinatorName, DBName, DBKey, PersistEveryNConsumedValues}, 
                          []).

consume_one_value(SequenceServerName) ->
    ?THROW_ERRORS( gen_server:call(SequenceServerName, consume_one_value) ).

stop(SequenceServerName) ->
    gen_server:cast(SequenceServerName, stop).


%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%--------------------------------------------------------------------
init({PortCoordName, DBName, DBKey, PersistEveryNConsumedValues}) ->

    ?INFO_LOGGER("starting (~p)~n", [self()]),
    process_flag(trap_exit, true),

    %%?INFO_LOGGER("init: PortCoordName = ~p DBFilename = ~p, DBKey = ~p, PersistEveryNConsumedValues =~p~n", [PortCoordName, DBFilename, DBKey, PersistEveryNConsumedValues]),

    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(PortCoordName),
    DB            = ?BDB_PC:get_db_handle(BdbPortHandle, bdb_DB_BTREE, DBName),

    %% Begin by initializing state.nextValue and state. startNextBatchValue from 
    %% the database (transactinally consuming one range of values in the process).
    {ok, consume_next_range(#state{bdbPortHandle               = BdbPortHandle,
                                   db                          = DB,
                                   dbKey                       = if is_binary(DBKey) -> DBKey; 
                                                                    true             -> list_to_binary(DBKey) 
                                                                 end,
                                   persistEveryNConsumedValues = PersistEveryNConsumedValues})}.


%%--------------------------------------------------------------------
%% Function: %% handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call(consume_one_value, _FromPid, State) ->

    if State#state.nextValue < State#state.startNextBatchValue ->

            %%?INFO_LOGGER("handle_call(consume_one_value): Returning ~w to ~w (~w values left in range)~n", [State#state.nextValue, FromPid, State#state.startNextBatchValue - (State#state.nextValue + 1)]),
            {reply, {ok, State#state.nextValue}, State#state{nextValue = State#state.nextValue + 1}};

       true ->
            %% Current batch is exhausted, need to start another one.
            NewState = consume_next_range(State),

            %%?INFO_LOGGER("handle_call(consume_one_value): NEW RANGE: Returning ~w to ~w (~w values left in range)~n", [NewState#state.nextValue, FromPid, NewState#state.startNextBatchValue - (NewState#state.nextValue + 1)]),
            {reply, {ok, NewState#state.nextValue}, NewState#state{nextValue = NewState#state.nextValue + 1}}
    end.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State}.

%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------
handle_info(Info, State) ->
    ?ERROR_LOGGER("Unexpected msg ~p~n", [Info]),
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(Reason, State) ->

    ?INFO_LOGGER("terminate: (reason: ~p)~n", [Reason]),
    
    %% Even though we may be shutting down cleanly, it is not safe to attempt to 
    %% avoid wasting values by writing the (unconsumed) State#state.nextValue 
    %% back to the database.  If this were guaranteed to be the only 
    %% process using this database entry, then it would be OK.  
    %% But we cannot assume that.
    %%
    %% So all we do is to nicely release the bdb port handle, just to avoid 
    %% the port_coordinator worrying about us dying.
    ok = ?BDB_PC:release_bdb_port_handle(State#state.bdbPortHandle),

    ?INFO_LOGGER("terminate: success~n"),
    ok.

%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%% Returns a new State.
%% This IGNORES (overwrites) the current State#state.nextValue and State#state.startNextBatchValue 
%% (it uses the database to generate new such values).
%% This works even if the counter did not previously exist (so long as the port and database are writeable).
consume_next_range(State) ->

    Txn    = ?BDB:txn_begin(State#state.bdbPortHandle, void, []),
    
    Cursor = ?BDB:db_cursor(State#state.bdbPortHandle, State#state.db, Txn, []),

    NewState = 
        try ?BDB:c_get(State#state.bdbPortHandle, Cursor, State#state.dbKey, <<>>, [bdb_DB_SET, bdb_DB_RMW]) of 

            {Key, <<NextValue:64>>} when Key =:= State#state.dbKey -> 
                %% There is a value in the database.  
                %% Do some sanity checks, but remember that we might still be initializing our 
                %% in-memory value (i.e. nextValue might be undefined)
                if is_integer(State#state.nextValue), State#state.nextValue =< NextValue;
                   State#state.nextValue =:= undefined ->

                        StartNextBatchValue = NextValue + State#state.persistEveryNConsumedValues,
                        %%?INFO_LOGGER("consume_next_range: START NEW RANGE: Storing StartNextBatchValue = ~w~n", [StartNextBatchValue]),
                        ok = ?BDB:c_put(State#state.bdbPortHandle, 
                                        Cursor, 
                                        State#state.dbKey, 
                                        <<StartNextBatchValue:64>>, 
                                        bdb_DB_CURRENT),

                        State#state{nextValue           = NextValue,
                                    startNextBatchValue = StartNextBatchValue}
                end;

            {Key, Value} when Key =:= State#state.dbKey -> 
                ?INFO_LOGGER("consume_next_range: Invalid sequence value in database: ~p~n", [Value]),
                throw({invalid_sequence_value_in_database, Value});

            Other ->
                ?ERROR_LOGGER("consume_next_range: c_get error ~p~n", [Other]),
                throw({database_error, Other})

        catch
            {bdb_error, bdb_DB_NOTFOUND} -> 

                %% No persistent entry at all. As a sanity check, we better not think we already have a valid nextValue...
                if is_atom(State#state.nextValue), State#state.nextValue == undefined ->

                        %%?INFO_LOGGER("consume_next_range: START NEW SEQUENCE: Key '~p'~n", [State#state.dbKey]),
                        ok = ?BDB:c_put(State#state.bdbPortHandle, 
                                        Cursor, 
                                        State#state.dbKey, 
                                        <<(State#state.persistEveryNConsumedValues):64>>, 
                                        bdb_DB_KEYFIRST),

                        State#state{nextValue           = 0,
                                    startNextBatchValue = State#state.persistEveryNConsumedValues}
                end

        end,    

    ok = ?BDB:c_close(NewState#state.bdbPortHandle, Cursor),
    ok = ?BDB:txn_commit(NewState#state.bdbPortHandle, Txn, []),
    NewState.
