%%%-------------------------------------------------------------------
%%% File    : berkeley_db_utilities.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%%   Generic functions for lists, proplists etc that are missing from stdlib.
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_utilities).

%% API
-export([lists_multikeyreplace/3,
         lists_transpose/1, 
         lists_find_and_remove_first/2,
         is_proplist/1, 
         normalize_sub_proplists_for_key/3,
         get_dominant_proplist_values/1, 
         safe_resolve_config_proplist/2,
         pmap/2,
         now_diff_milliseconds/2
        ]).

%% Test suite
-export([test/0]).

%%====================================================================
%% API
%%====================================================================
%%--------------------------------------------------------------------
%% Function: 
%% Description:
%%--------------------------------------------------------------------

%%
%% Lists
%%

lists_transpose([])           -> [];
lists_transpose([[] | _])     -> [];
lists_transpose(ListOfLists)  -> [lists:map(fun(List) -> hd(List) end, ListOfLists)
                                  | lists_transpose(lists:map(fun(List) -> tl(List) end, ListOfLists))].



%% multi_keyreplace(KeysNewTuples, KeyPos, TupleList1) -> TupleList2 
%%  
%% Types: 
%%  
%% KeysNewTuples           = [KeyNewTuple]
%% KeyNewTuple             = {Key, NewTuple}
%% Key                     = term{}
%% NewTuple   = Tuple      = tuple()
%% TupleList1 = TupleList2 = [Tuple]
%% KeyPos                  = 1..size(Tuple)
%%  
%% Returns a copy of TupleList1, where ALL occurences of tupes 
%% whose Nth element is a Key in KeysNewTuples is replaced with 
%% the associated NewTuple from KeysNewTuples
%%
%% If a Key appears multiple times in KeysNewTuples then the NewTuple corresponding to the first occurence is used.
%%
%% i.e. this is like lists:keyreplace/4 but 
%%   a) replaces ALL occurences not just the first
%%   b) can perform multiple substitutions in a single pass
%t
%% Complexity is of course O(N*M)

lists_multikeyreplace(ReplaceKeysNewValues, KeyPos, [Tuple|MoreTuples]) ->
    case lists:keysearch(element(KeyPos, Tuple), 1, ReplaceKeysNewValues) of
        {value, {_, NewValue}} ->
            [NewValue | lists_multikeyreplace(ReplaceKeysNewValues, KeyPos, MoreTuples)];
        _ ->
            [Tuple | lists_multikeyreplace(ReplaceKeysNewValues, KeyPos, MoreTuples)]
    end;

lists_multikeyreplace(_, _, []) -> [].


%% lists_find_and_remove_first
%%
%% This is useful if you want to both discover whether an element is present in a list 
%% and remove it.  (It is assumed to occur only once.)
%%
%% returns
%%     not_found
%% or 
%%     {found, ListAfterRemoval}

lists_find_and_remove_first(_, []) ->
    not_found;

lists_find_and_remove_first(FindElement, [FindElement | T]) ->
    {found, T};
    
lists_find_and_remove_first(FindElement, [OtherElement | T]) ->
    case lists_find_and_remove_first(FindElement, T) of
        {found, ListBeyondFind} ->  {found, [OtherElement | ListBeyondFind]};
        not_found               ->  not_found
    end.
    

%%
%% PropLists
%% 
%%

%% This checks that a list is a well-formed property list 
%% (a list of 2-tuples -- we don't allow single atoms)
is_proplist(List) when is_list(List) ->
    lists:foldl(fun(Elem, TrueSoFar) when is_tuple(Elem), size(Elem) == 2 ->
                        TrueSoFar;
                   (_Elem, _TrueSoFar) -> 
                        false
                end,
                true, % empty list is OK
                List);
is_proplist(_) ->
    false.
                                     

%% When using proplists for config parameters, we often want options to have lists of sub-options.
%% We also want to allow multiple occurences of those options, and we want to treat those multiple occurences
%% as cummulative (comcatenated) proplists in their own right.  
%%
%%  eg.  [{multi_option_key1, [{key1, value1}]}, {multi_option_key2, [{key1, value2}]}, ... ]
%% becomes
%%       [{multi_option_key1, [{key1, value1}, {key1, value2}]}]
%%
%% Also, we cannot mix 'disabled' parameters and active config (either all occurences are 'disabled', or none of them are).
%%
%% This function 
%%    extracts all occurences for a give key
%%    concatenates the lists of sub-options for that key
%%    runs get_dominant_proplist_values() on that concatenated list
%%    returns a new version of PropList that has exactly zero or one occurences of the key
%%      (if there is an occurence it is placed at the START of the proplist, and it's value is the processed list described above).
%%
%% This does NOT process any other key/value in PropList; they are all returned in the same order as on entry.
%% In particular, this does NOT run get_dominant_proplist_values() on the PropList, as there may be other 
%% keys that the caller wants to process in this fashion.

normalize_sub_proplists_for_key(Key, PropList, DefaultSubListIfNotDisabled) ->
    
    case lists:foldl(fun({K, _} = OtherKeyValue, {RevSubLists, RevOthers}) when K /= Key ->
                             {RevSubLists, [OtherKeyValue|RevOthers]};

                        %% In all patterns below, we know we are dealing with Key
                        ({_, disabled}, {[], RevOthers}) -> 
                             {disabled, RevOthers};

                        ({_, disabled}, {disabled, RevOthers}) -> 
                             {disabled, RevOthers};

                        ({_, BadArg} = Elem, _) when not is_list(BadArg) -> 
                             %% TODO: alternatively we could make them lists; would that be useful or dangerous?
                             throw({value_for_key_is_not_a_list_or_disabled, Elem, PropList});

                        (_,  {disabled, _}) -> 
                             throw({cannot_mix_disabled_and_active_config, Key, PropList});

                        ({_, SubList}, {RevSubLists, RevOthers}) -> 
                             {[SubList|RevSubLists], RevOthers}

                     end,
                     {[], []},
                     PropList) of

        {[], _} ->
            %% TODO: there is a case for throwing here, to avoid suprise behaviour (caller can always append to PropList before calling to prevent it)
            PropList;    % no change (Key not found at all is *implicitly disabled*)

        {disabled, RevOthers} ->
            [{Key, disabled} | lists:reverse(RevOthers)];    % no real change, but move Key to front for consistency (in the documented contract)
        
        {RevSubLists, RevOthers} ->
            [{Key, get_dominant_proplist_values( lists:reverse([DefaultSubListIfNotDisabled | RevSubLists]) )}
             | lists:reverse(RevOthers)]
    end.


%% This returns a list of all of the first occurences of 
    
%% Special case optimization for safe_resolve_config_proplist/2
%% in the common case when the first argument is an empty list.
get_dominant_proplist_values([[], List]) ->
    List;

get_dominant_proplist_values(ListOfProplists) ->
    lists:foldl(fun({Key, _} = Item, DominantProplist) ->
                        case proplists:lookup(Key, DominantProplist) of
                            {Key, _} -> DominantProplist;
                            none     -> [Item | DominantProplist]
                        end
                end,
                [],
                lists:append(ListOfProplists)).


%% Optimization for the usual case; SpecifiedConfigProplist is empty
safe_resolve_config_proplist([], Default) ->
    {ok, Default};

safe_resolve_config_proplist(Specified, Default) ->
    case is_proplist(Specified) of 
        true -> 
            %% This is in a try-catch so that bad SpecifiedConfigProplist cannot kill the server.
            try
                {ok, get_dominant_proplist_values([Specified, Default])}
            catch
                Exception ->
                    {error, Exception}
            end;
        false ->
            {error, malformed_proplist}
    end.


%% From Luke Gorrie.  Like rpc:pmap but takes a Fun
pmap(Fun, List) ->
    Parent = self(),
    [receive {Pid, Result} -> Result end
     || Pid <- [spawn(fun() -> Parent ! {self(), Fun(Elem)} end) || Elem <- List]].


%% This is faster than timer:now_diff() because it does not use bignums.
%% But it returns *milliseconds*  (timer:now_diff returns microseconds.)
%% From http://www.erlang.org/ml-archive/erlang-questions/200205/msg00027.html
now_diff_milliseconds({M,S,U}, {M,S1,U1}) ->
    ((S-S1) * 1000) + ((U-U1) div 1000);
now_diff_milliseconds({M,S,U}, {M1,S1,U1}) ->
    ((M-M1)*1000000+(S-S1))*1000 + ((U-U1) div 1000).


%%====================================================================
%% Internal functions
%%====================================================================


test() ->
    
    io:format("lists_transpose~n"),
    []            = lists_transpose([]),
    []            = lists_transpose([[]]),
    [[a],[b]]     = lists_transpose([[a,b]]),
    [[a,c],[b,d]] = lists_transpose([[a,b],[c,d]]),

    io:format("lists_multikeyreplace~n"),
    %% no substitutions found
    [{key1, oldval1}, {key2, oldval2}, {key3, oldval3}]
        = lists_multikeyreplace([{key10, {key10, newval10}}, {key30, {key30, newval30}}], 
                                  1,
                                  [{key1, oldval1}, {key2, oldval2}, {key3, oldval3}]),

    [{key1, oldval1}, {key2, oldval2}, {key3, oldval3}]
        = lists_multikeyreplace([], 
                                1, 
                                [{key1, oldval1}, {key2, oldval2}, {key3, oldval3}]),

    %% some substitutions used, but some substitutions not used
    [{key1, newval1}]
        = lists_multikeyreplace([{key1, {key1, newval1}}], 
                                 1, 
                                 [{key1, oldval1}]),

    [{key1, newval1}]
        = lists_multikeyreplace([{key1, {key1, newval1}}, {key3, {key3, newval3}}], 
                                  1, 
                                  [{key1, oldval1}]),

    [{key1, newval1}]
        = lists_multikeyreplace([{key3, {key3, newval3}}, {key1, {key1, newval1}}], 
                                  1, 
                                  [{key1, oldval1}]),

    %% some values substituted, but some not substituted (and order is preserved)
    [{key1, newval1}, {key2, oldval2}]
        = lists_multikeyreplace([{key1, {key1, newval1}}], 
                                 1, 
                                 [{key1, oldval1}, {key2, oldval2}]),

    [{key2, oldval2}, {key1, newval1}]
        = lists_multikeyreplace([{key1, {key1, newval1}}], 
                                1, 
                                [{key2, oldval2}, {key1, oldval1}]),

    %% multiple substitutions
    [{key1, newval1}, {key2, oldval2}, {key3, newval3}]
        = lists_multikeyreplace([{key1, {key1, newval1}}, {key3, {key3, newval3}}], 
                                1,
                                [{key1, oldval1}, {key2, oldval2}, {key3, oldval3}]),

    [{key1, newval1}, {key2, oldval2}, {key3, newval3}, {key2, oldval2}, {key1, newval1}, {key3, newval3}]
        = lists_multikeyreplace([{key1, {key1, newval1}}, {key3, {key3, newval3}}], 
                                1,
                                [{key1, oldval1}, {key2, oldval2}, {key3, oldval3}, {key2, oldval2}, {key1, oldval1},{key3, oldval3}]),

    [{key0, oldval0}, {key1, newval1}, {key2, oldval2}, {key3, newval3}, {key4, oldval4}]
        = lists_multikeyreplace([{key1, {key1, newval1}}, {key3, {key3, newval3}}], 
                                1, 
                                [{key0, oldval0}, {key1, oldval1}, {key2, oldval2}, {key3, oldval3}, {key4, oldval4}]),


    io:format("lists_find_and_remove_first~n"),
    not_found = lists_find_and_remove_first(elem, []),
    not_found = lists_find_and_remove_first(elem, [not_elem]),
    not_found = lists_find_and_remove_first(elem, [not_elem, not_elem2]),
    {found, []} = lists_find_and_remove_first(elem, [elem]),
    {found, [not_elem]} = lists_find_and_remove_first(elem, [elem, not_elem]),
    {found, [not_elem1, not_elem2]} = lists_find_and_remove_first(elem, [elem, not_elem1, not_elem2]),
    {found, [not_elem]} = lists_find_and_remove_first(elem, [not_elem, elem]),
    {found, [not_elem1, not_elem2]} = lists_find_and_remove_first(elem, [not_elem1, not_elem2, elem]),
    {found, [not_elem1, not_elem2]} = lists_find_and_remove_first(elem, [not_elem1, elem, not_elem2]),
    {found, [not_elem1, not_elem2, not_elem3, not_elem4]} = lists_find_and_remove_first(elem, [not_elem1, not_elem2, elem, not_elem3, not_elem4]),
    %% multiple occurencess...
    {found, [elem]} = lists_find_and_remove_first(elem, [elem, elem]),
    {found, [elem, not_elem]} = lists_find_and_remove_first(elem, [elem, elem, not_elem]),
    {found, [elem, not_elem1, not_elem2]} = lists_find_and_remove_first(elem, [elem, elem, not_elem1, not_elem2]),
    {found, [not_elem, elem]} = lists_find_and_remove_first(elem, [not_elem, elem, elem]),
    {found, [not_elem, elem]} = lists_find_and_remove_first(elem, [elem, not_elem, elem]),
    {found, [not_elem1, not_elem2, elem]} = lists_find_and_remove_first(elem, [not_elem1, not_elem2, elem, elem]),
    {found, [not_elem1, elem, not_elem2]} = lists_find_and_remove_first(elem, [not_elem1, elem, elem, not_elem2]),
    {found, [not_elem1, not_elem2, elem, not_elem3, not_elem4]} = lists_find_and_remove_first(elem, [not_elem1, not_elem2, elem, elem, not_elem3, not_elem4]),


    io:format("is_proplist~n"),
    true  = is_proplist([]),
    true  = is_proplist([{key, value}]),
    false = is_proplist([{key_no_value}]),
    false = is_proplist([{key, value, extra}]),
    false = is_proplist(atom),
    false = is_proplist([atom]),

    io:format("normalize_sub_proplists_for_key~n"),
    [] = normalize_sub_proplists_for_key(key, [], []),

    [{key2, [{subkey2, subvalue2}]}] = normalize_sub_proplists_for_key(key, [{key2, [{subkey2, subvalue2}]}], []), % doesn't matter if key is not present

    [{key, [{subkey2, subvalue2}]}] = normalize_sub_proplists_for_key(key, [{key, [{subkey2, subvalue2}]}], []),

    [{key, [{default_is_applied, subvalue3}, {subkey2, subvalue2a}]}] 
        = normalize_sub_proplists_for_key(key, 
                                          [{key, [{subkey2, subvalue2a}]}], 
                                          [{default_is_applied, subvalue3}, {subkey2, subvalue2b}]),

    [{key2, disabled}, {key3, val3a}, {key3, val3b}, {key2, val2b}] 
        = normalize_sub_proplists_for_key(key, 
                                          [{key2, disabled}, {key3, val3a}, {key3, val3b}, {key2, val2b}],
                                          [{default_not_applied, as_missing_means_disabled}]), % others pass through in same order (and dupes are not removed)

    [{key, disabled}, {key_2, val2}] 
        = normalize_sub_proplists_for_key(key, 
                                          [{key_2, val2}, {key, disabled}], 
                                          [{default_not_applied, as_missing_means_disabled}]), % disabled is allowed, and key always moves to front

    [{key, [{default_is_applied, subvalue3}, {subkey3, subvalue3}, {subkey1, subvalue1a}, {subkey2, subvalue2a}]}, {key__2, val2}] 
        = normalize_sub_proplists_for_key(key, 
                                          [{key, [{subkey2, subvalue2a}, {subkey1, subvalue1a}]}, {key__2, val2}, {key, [{subkey1, subvalue1b}, {subkey3, subvalue3}]}], 
                                          [{default_is_applied, subvalue3}, {subkey2, subvalue2default}]), % sublists are found and first version of multiple occurences within sublists wins

    {value_for_key_is_not_a_list_or_disabled, {key, not_list}, [{key___2, val2}, {key, disabled}, {key, not_list}]}
        = (catch normalize_sub_proplists_for_key(key, 
                                                 [{key___2, val2}, {key, disabled}, {key, not_list}], 
                                                 [])), % all values must be sub-lists (default is irrelevant)

    {cannot_mix_disabled_and_active_config, key, [{key____2, val2}, {key, disabled}, {key, [{subkey1, subvalue1}]}]}
        = (catch normalize_sub_proplists_for_key(key, 
                                                 [{key____2, val2}, {key, disabled}, {key, [{subkey1, subvalue1}]}], 
                                                 [])), % cannot mix 'disabled' with active config (default is irrelevant)


    io:format("get_dominant_proplist_values~n"),
    [] = get_dominant_proplist_values([]),
    [] = get_dominant_proplist_values([[]]),
    [] = get_dominant_proplist_values([[], []]),
    [{key, value}] = get_dominant_proplist_values([[], [{key, value}]]),
    [{key, value}] = get_dominant_proplist_values([[{key, value}], []]),
    [{key, value1}] = get_dominant_proplist_values([[{key, value1}], [{key, value2}]]),
    [{key2, value2}, {key, value1}] = get_dominant_proplist_values([[{key, value1}], [{key2, value2}]]),
    [{key2, value2}, {key, value1}] = get_dominant_proplist_values([[{key, value1}, {key2, value2}], [{key2, value3}]]),

    io:format("safe_resolve_config_proplist~n"),
    {ok, []} = safe_resolve_config_proplist([], []),
    {ok, [{key, value}]}  = safe_resolve_config_proplist([], [{key, value}]),
    {ok, [{key, value}]}  = safe_resolve_config_proplist([{key, value}], []),
    {ok, [{key, value1}]} = safe_resolve_config_proplist([{key, value1}], [{key, value2}]),
    {ok, [{key2, value2}, {key, value1}]} = safe_resolve_config_proplist([{key, value1}], [{key2, value2}]),
    {ok, [{key2, value2}, {key, value1}]} = safe_resolve_config_proplist([{key, value1}, {key2, value2}], [{key2, value3}]),
    {error, malformed_proplist} = safe_resolve_config_proplist(atom, []),

    io:format("now_diff_milliseconds~n"),
    0 = now_diff_milliseconds({0,0,0}, {0,0,0}),
    0 = now_diff_milliseconds({0,0,999}, {0,0,0}),        %        999 microseconds
    1 = now_diff_milliseconds({0,0,1000}, {0,0,0}),       %       1 millisecond
    1000 = now_diff_milliseconds({0,1,0}, {0,0,0}),       %       1 millisecond
    1000 = now_diff_milliseconds({0,1,0}, {0,0,999}),     %       1 millisecond
    0 = now_diff_milliseconds({0,1,0}, {0,0,1000000}),    % equivalent times
    1000 = now_diff_milliseconds({0,1,999999}, {0,0,999999}),
    0 = now_diff_milliseconds({0,1,999999}, {0,1,999999}),
    1999 = now_diff_milliseconds({0,1,999999}, {0,0,0}),
    1999 = now_diff_milliseconds({0,1,1000000}, {0,0,1}),
    1000000000 = now_diff_milliseconds({1,0,0}, {0,0,0}), % 1 million seconds
    0 = now_diff_milliseconds({1,0,0}, {0,1000000,0}),    % equivalent times
    -1 = now_diff_milliseconds({1,0,0}, {0,1000000,1000}),

    ok.
