
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>

#include <erl_driver.h>
#include <erl_driver_tk.h>

#include "my-berkeley_db.h"

#ifdef	DRIVER_USING_PTHREADS
#include <pthread.h>
static pthread_mutex_t errcall_output_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif	/* DRIVER_USING_PTHREADS */

#ifdef WRITE_CORE_DUMP_IN_ERRCALL
#include <google/coredumper.h>
#include <time.h>
#include <dlfcn.h>
static int    library_opened   = 0;
static void * library_handle   = NULL;
static int (*pfnWriteCoreDump)(const char *);
static time_t lastCoreDumpTime = 0;
#define MIN_TIME_BETWEEN_COREDUMPS_SECS  30
#endif

static int make_dbt(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount, DBT *dbt, char *orig_buf);
static int enqueue_untagged_event_notification(descriptor_t *desc, int event_command_code, u_int32_t payload_uint32, long payload_long);

void
null(void)
{
    fprintf(stderr, "\r\n%s: should never be called. Aborting.\r\n", __FUNCTION__);
    exit(EXIT_FAILURE);
}

/*
** We flip around the return value of db_env_create so that we normally
** return the new DB_ENV *, which makes EDTK's simple valmap mechanism
** happy.
*/

DB_ENV *
my_db_env_create(u_int32_t flags, int *my_error_value)
{
    DB_ENV	*dbenv;
    int		err;

    if ((err = db_env_create(&dbenv, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    return dbenv;
}

int
my_db_env_close(DB_ENV *dbenv, u_int32_t flags)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    edtk_debug("%s: %s: dbenv = 0x%p", desc->port_label, __FUNCTION__, (void*)dbenv);
    return dbenv->close(dbenv, flags);
}

int
my_db_env_close_no_flags(DB_ENV *dbenv)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    edtk_debug("%s: %s: dbenv = 0x%p", desc->port_label, __FUNCTION__, (void*)dbenv);
    return dbenv->close(dbenv, 0);
}


void 
my_msgcall(const DB_ENV *dbenv, const char *msg)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);

    time_t            now;
    char              time_buf[30];             /* docs say 'at least 26 bytes' */
    time(&now);
    ctime_r(&now, time_buf);
    time_buf[sizeof(time_buf) - 1] = 0;   /* ensure null-termination */
    time_buf[strlen(time_buf) - 1] = 0;   /* the asctime functions append a \n for no good reason */
    
#ifdef DRIVER_USING_PTHREADS
    pthread_mutex_lock(& errcall_output_mutex);
#endif
    
    fprintf(stderr, "%s: %s: MSG: %s\r\n", desc->port_label, time_buf, msg);

#ifdef DRIVER_USING_PTHREADS
    pthread_mutex_unlock(& errcall_output_mutex);
#endif
}

void 
my_errcall(const DB_ENV *dbenv, const char *errpfx, const char *msg)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);

    time_t            now;
    char              time_buf[30];             /* docs say 'at least 26 bytes' */
    time(&now);
    ctime_r(&now, time_buf);
    time_buf[sizeof(time_buf) - 1] = 0;   /* ensure null-termination */
    time_buf[strlen(time_buf) - 1] = 0;   /* the asctime functions append a \n for no good reason */
    
#ifdef DRIVER_USING_PTHREADS
    pthread_mutex_lock(& errcall_output_mutex);
#endif

    fprintf(stderr, "%s: %s: ERROR: %s: %s\r\n", desc->port_label, time_buf, errpfx ? errpfx : "(no errpfx)", msg);

#ifdef WRITE_CORE_DUMP_IN_ERRCALL
    if (  ! strstr(msg, "EOF on connection")
       && ! strstr(msg, "redundant incoming connection will be ignored") )
    {
        /* For some reason, I can't get automatic loading of Google's libcoredumper.so
         * to work.  So to avoid "undefined symbol: WriteCoreDump" errors, I load it
         * and find the symbol myself.  It works, at least. */
        if ( ! library_opened ) {
            library_opened = 1;
            if ( (library_handle = dlopen("libcoredumper.so", RTLD_LOCAL | RTLD_LAZY)) == NULL ) {
                fprintf(stderr, "%s: %s: Failed to dlopen libcoredumper.so : %s\r\n", desc->port_label, __FUNCTION__, dlerror());
            }
            if ( (pfnWriteCoreDump = dlsym(library_handle, "WriteCoreDump")) == NULL ) {
                fprintf(stderr, "%s: %s: Failed to find symbol WriteCoreDump in libcoredumper.so : %s\r\n", desc->port_label, __FUNCTION__, dlerror());
            }
        }

        if ( pfnWriteCoreDump != NULL )
        {
            time_t now;
            time(&now);
            if ( difftime(now, lastCoreDumpTime) >= MIN_TIME_BETWEEN_COREDUMPS_SECS )
            {
                char filename[128];
                snprintf(filename, sizeof(filename), "core_bdb_errcall.%ld", now);
                filename[sizeof(filename)-1] = 0;
            
                fprintf(stderr, "%s: %s: Writing coredump to %s\r\n", desc->port_label, __FUNCTION__, filename);
                if ( (*pfnWriteCoreDump)(filename) == 0 ) {
                    fprintf(stderr, "%s: %s: Sucessfylly written coredump to %s\r\n", desc->port_label, __FUNCTION__, filename);
                } else {
                    fprintf(stderr, "%s: %s: Failed to write coredump %s : %d\r\n", desc->port_label, __FUNCTION__, filename, errno);
                }
                lastCoreDumpTime = now;
           }
        }
    }
#endif                 
   
#ifdef DRIVER_USING_PTHREADS
    pthread_mutex_unlock(& errcall_output_mutex);
#endif
}

int
my_env_set_errfile(DB_ENV *dbenv, char *file, char *mode)
{
    FILE	*fp;

    if ((fp = fopen(file, mode)) == NULL) {
	return errno;
    }
    dbenv->set_errfile(dbenv, fp);
    return 0;
}

int
my_db_set_errfile(DB *db, char *file, char *mode)
{
    FILE	*fp;

    if ((fp = fopen(file, mode)) == NULL) {
	return errno;
    }
    db->set_errfile(db, fp);
    return 0;
}

DB *
my_db_create(DB_ENV *dbenv, u_int32_t flags, int *my_error_value)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    DB	*db;
    int	err;

    if ((err = db_create(&db, dbenv, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    db->app_private = desc;
    return db;
}

int
my_db_close_no_flags(DB *db)
{
    descriptor_t * desc = (descriptor_t *)(db->app_private);
    edtk_debug("%s: %s: db = 0x%p", desc->port_label, __FUNCTION__, (void*)db);
    return db->close(db, 0);
}

int
my_db_close(DB *db, u_int32_t flags)
{
    descriptor_t * desc = (descriptor_t *)(db->app_private);
    edtk_debug("%s: %s: db = 0x%p", desc->port_label, __FUNCTION__, (void*)db);
    return db->close(db, flags);
}

DBC *
my_cursor_create(DB *db, DB_TXN *txnid, u_int32_t flags, int *my_error_value)
{
    DBC	*cursorp;
    int	err;

    if ((err = db->cursor(db, txnid, &cursorp, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    /* TODO: if/when DBC gets app_private field, set it here from db->app_private */
    return cursorp;
}

DBC *
my_cursor_dup(DBC *cursor, u_int32_t flags, int *my_error_value)
{
    DBC	*cursorp;
    int	err;

    if ((err = cursor->c_dup(cursor, &cursorp, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    /* TODO: if/when DBC gets app_private field, copy it here */
    return cursorp;
}

int
my_cursor_close(DBC *cursor)
{
    /* TODO: as DBC does not currently have an app_private field we do not have access to desc->port_label here */
    edtk_debug("%s: cursor = 0x%p", __FUNCTION__, (void*)cursor);
    return cursor->c_close(cursor);
}


/* transaction wrappers */

DB_TXN *
my_txn_begin(DB_ENV *dbenv, DB_TXN *parent_txnid, u_int32_t flags,
             int *my_error_value)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    DB_TXN * new_txn;
    int	err;

    if ((err = dbenv->txn_begin(dbenv, parent_txnid, & new_txn, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }

    /* TODO: if/when DB_TXN gets app_private field, set it here from db->app_private */
    edtk_debug("%s: %s: db_txn = 0x%p", desc->port_label, __FUNCTION__, (void*)new_txn);
    return new_txn;
}

/* my_txn_abort is only used as a valmap cleanup function */
int
my_txn_abort(DB_TXN *db_txn)
{
    /* TODO: as DB_TXN does not currently have an app_private field we do not have access to desc->port_label here */
    edtk_debug("%s: db_txn = 0x%p", __FUNCTION__, (void*)db_txn);
    return db_txn->abort(db_txn);
}


int
my_db_verify(DB *db, char *file, char *database, char *outfilename,
	     u_int32_t flags)
{
    FILE	*outfile = NULL;

    if (outfilename != NULL && (outfile = fopen(outfilename, "w")) == NULL) {
	return errno;
    }
    return db->verify(db, file, database, outfile, flags);
}

/* These are called from xtra_return functions, from inside _ready_async.
 * They must increment *members by 1.
 * They currently ALWAYS return *members (if 1 this indicates success, but it's a number, not 'true').
 * Failure is handled by **hard-aborting** the OS process
 *
 * They also adjust *members and *msgcount as appropriate.
 * IMPORTANT: The caller will ASSERT that *members == 1 (but *msgcount can be more than that) */

int
make_dbt_key(descriptor_t *desc, callstate_t *c,
             ErlDrvTermData *msg, int *members, int *msgcount)
{

    return make_dbt(desc, c, msg, members, msgcount, &c->i.key_in, c->i.ekey_in);
}

int
make_dbt_data(descriptor_t *desc, callstate_t *c,
              ErlDrvTermData *msg, int *members, int *msgcount)
{

    return make_dbt(desc, c, msg, members, msgcount, &c->i.data_in, c->i.edata_in);
}

static int
make_dbt(descriptor_t *desc,
         callstate_t *      c,
         ErlDrvTermData *   msg,
	 int *              members,
         int *              msgcount,
         DBT *              dbt,
         char *             orig_buf)
{
    char		*bin_mem;
    ErlDrvBinary	*bin = NULL;

    /* TODO: We currently always copy the data. But in many cases (e.g. the key argument to db_get)
     * the data we are returning might already be an Erlang binary that was part of the input, and
     * we could just use the same binary (before free_callstate calls driver_free_binary on it).
     * We can detect those cases because dbt->data == orig_buf.
     * An alternative is to add a new db_get_data function that returns <<ok, data>> only. */
    
    if ((bin_mem = edtk_driver_alloc_wrapper(dbt->size)) == NULL) {
        fprintf(stderr, "\r\n%s: %s: edtk_driver_alloc_wrapper failed! Aborting.\r\n", desc->port_label, __FUNCTION__);
        exit(EXIT_FAILURE);
    }
    memcpy(bin_mem, dbt->data, dbt->size);
    bin = edtk_alloced_ptr2ErlDrvBinary(bin_mem);

    /* Ensure that we decrement it's refcount at the end of _ready_async
     * (*AFTER* being send to Erlang -- i.e. Erlang will have incremented
     * it's refcount, so it won't really be freed) */
    assert(desc->num_tofree < MAX_RETURN_TERMS);
    desc->tofree[desc->num_tofree++] = bin;

    /* If BDB malloced data for this object then we must free it, now that we have copied it to a driver binary.
     * However, we free it in free_callstate not here, because this function (make_dbt) might not be called.
     * If driver shutdown has begun after invoke_XX has called BDB but before ready_async is called then ready_async
     * won't actually be called (and therefore cannot call us). Only free_callstate will be called. */

    *msgcount = LOAD_BINARY(msg, *msgcount, bin, 0, dbt->size);
    *members = 1;
    return 1;
}


/* This is called when txn_prepare is successful (by a hack in berkeley_db.xml).
 * It is called for each immediate child of the prepared txn, NOT the prepared txn itself
 * (as it clears the slots of each txn it is called on, *after* recursing on it's children. */
void discard_descendents_of_prepared_txn(descriptor_t * desc, int idx)
{
    if ( desc->valmap_db_txn[idx].child_count != 0 )
    {
        int child_idx;
        edtk_debug("%s: %s: threadid=%lx, prepared parent=%d has child_count %d", desc->port_label, __FUNCTION__, pthread_self(), idx, desc->valmap_db_txn[idx].child_count);
        for (child_idx = 0; child_idx < desc->num_valmap_db_txn_slots; child_idx++)
        {
            if ( desc->valmap_db_txn[child_idx].parent_index == idx )
            {
                edtk_debug("%s: %s: threadid=%lx, found child %d (of parent %d)", desc->port_label, __FUNCTION__, pthread_self(), child_idx, idx);

                /* Recurse to the leaf nodes and then cleanup on the way back out */
                discard_descendents_of_prepared_txn(desc, child_idx);

                /* Most parents will have few children, and they will be found early.  
                 * So it's probably worth stopping the scan of the array when we've done them all.*/
                if ( desc->valmap_db_txn[idx].child_count == 0 ) {
                   break;
                }
            }
        }
    }
    
    desc->valmap_db_txn[idx].value = NULL;

    /* IMPORTANT: in cleanup_$(.name)_valmap_indes there is a hack section here for "cleanup_additional_valmap_struct_members".
     * berkeley_db.xml has a hack that frees desc->valmap_db_txn[i].distributed_txn_guid.
     * We KNOW that cannot be set for children of a prepared txn.
     * But any other such hacks for db_txn must be included here!!
     * (Maintenance headache of course; berkeley_db.xml does have documentation for this. */         
    assert( desc->valmap_db_txn[idx].distributed_txn_guid == NULL );

    desc->valmap_db_txn[idx].flags = 0;

    ++(desc->valmap_db_txn[idx].generation_id); /* to prevent reuse of expired valmap tuples that the Erlang application might still have */
    /* edtk_debug("%s: %s: threadid=%lx, i=%d, made generation-id=%d", desc->port_label, __FUNCTION__, pthread_self(), idx, desc->valmap_db_txn[idx].generation_id); */

    if ( desc->valmap_db_txn[idx].parent_index >= 0 ) {
        /* This was a child (the only time it's not a child is the top-level prepared txn).
         * Decrement parent's child count. */
        --(desc->valmap_db_txn[desc->valmap_db_txn[idx].parent_index].child_count);
        edtk_debug("%s: %s: threadid=%lx, i=%d is a child with parent=%d. Have reduced parent's child count to %d", desc->port_label, __FUNCTION__, pthread_self(), idx, desc->valmap_db_txn[idx].parent_index, desc->valmap_db_txn[desc->valmap_db_txn[idx].parent_index].child_count);
    }

    desc->valmap_db_txn[idx].parent_index = -1;

    /* desc->valmap_db_txn[idx].group_id = -1; */    /* *not* a special value, but handy for debugging */
    edtk_debug("%s: %s: threadid=%lx, finished discarding children of prepared parent=%d", desc->port_label, __FUNCTION__, pthread_self(), idx);
}


/* This is only called after a successful call to txn_recover.
 * c->o.num_unresolved_txns is the length of the list.
 * Any unresolved transactions have been installed in
 * i.e. valmap_db_txn[*].value              is the DB_TXN *,
 * and  valmap_db_txn[*].distributed_txn_id is the GUID
 *
 * Note that c->o.num_unresolved_txns has already been capped to
 * avoid overflowing the valmap_db_txn array
 * AND avoid overflowing MAX_RETURN_TERMS when constructing this list.
 * See code in berkeley_db.xml (hack section for txn_recover).
 * If this ever creates more than (7*N) + 1 stack entries then that code needs to change.
 */
int
my_make_list_unresolved_txns(descriptor_t *desc,
                             callstate_t *c,
                             ErlDrvTermData *msg,
                             int *members,
                             int *msgcount)
{
    int txn_idx;

    edtk_debug("%s: %s: num unresolved txns = %d", desc->port_label, __FUNCTION__, (int)(c->o.num_unresolved_txns));
    for (txn_idx = 0; txn_idx < c->o.num_unresolved_txns; ++txn_idx)
    {
        /* Append <<DistTxnGUID>>, #bdb_valmap_db_txn{}.
         * This allocates memory, so may fail. */
        int inner_members = 0;
        /* This consumes 6 stack slots, and creates 2 members. */
        /* edtk_debug("%s: %s: before my_LOAD_unresolved_DistTxnId_and_valmap_record *msgcount=%d", desc->port_label, __FUNCTION__, *msgcount); */
        if ( 0 == my_LOAD_unresolved_DistTxnId_and_valmap_record(desc, txn_idx, msg, &inner_members, msgcount) )
        {
            edtk_debug("%s: %s : after (failed) my_LOAD_unresolved_DistTxnId_and_valmap_record() *msgcount=%d inner_members=%d. Calling txn_discard on remaining %d unresolved txns and returning list of %d txns", desc->port_label, __FUNCTION__, *msgcount, inner_members, (int)(c->o.num_unresolved_txns - txn_idx), txn_idx);
            /* Failed.  We explicitly txn_discard all of the remaing transactions
             * and return what we have.
             * This function is only called by read_async, so doing this calls
             * txn_discard from the Erlang VM thread. That's OK as it does not acquire
             * locks, and this should never happen anyway.
             *
             * Note: if we fail to allocate on the first txn then we CANNOT simplj 
             * return an empty list, as that would tell the port_coordinator that there are NO
             * unresolved txns remaining.  That is not safe -- BDB will NOT currently reject attempts
             * to create new transactions if there are unresolved txns, it will just CORRUPT the log.
             * So if we would return an empty list we append an atom to it, which will 
             * cause a badmatch in the port_coordinator */
            
            for ( ; txn_idx < c->o.num_unresolved_txns; ++txn_idx )
            {
                edtk_debug("%s: %s: ************ discarding txn #%d", desc->port_label, __FUNCTION__, txn_idx);
                desc->valmap_db_txn[txn_idx].value->discard(desc->valmap_db_txn[txn_idx].value, 0);

                free(desc->valmap_db_txn[txn_idx].distributed_txn_guid);
                desc->valmap_db_txn[txn_idx].distributed_txn_guid = NULL;
                desc->valmap_db_txn[txn_idx].value                = NULL;
            }
            if ( txn_idx == 0 ) {
                *msgcount = LOAD_ATOM(msg, *msgcount, am_enomem);
                ++(*members);
            }
            break;
        } else {
            /* edtk_debug("%s: %s: after (success) my_LOAD_unresolved_DistTxnId_and_valmap_record *msgcount=%d inner_members=%d", desc->port_label, __FUNCTION__, *msgcount, inner_members); */
        }
        /* make it into a 2-tuple */
        *msgcount = LOAD_TUPLE(msg, *msgcount, inner_members);
        ++(*members);
    }

    /* The list has to end with NIL */
    *msgcount = LOAD_NIL(msg, *msgcount);
    ++(*members);
    
    edtk_debug("%s: %s: added %d out of %ld (*msgcount=%d, *members=%d)", desc->port_label, __FUNCTION__, txn_idx, c->o.num_unresolved_txns, *msgcount, *members);

    /* In berkeley_db.xml, this xtra-return type is declared as a 'list'.
     * So reply_xtra will do the final "*msgcount = LOAD_LIST(msg, *msgcount, members);" */

    return *members;
}


/* Helper which adds these two parts to a message.
 * Note, these are NOT wrapped in an enclosing tuple.
 *
 *    DistTxnGUIDBinary, #bdb_valmap_db_txn{}
 *
 * This consumes 6 stack slots, and creates 2 members.
 *
 * Returns 0 if fails, N (number of added members) if successfull.
 * Can only fail due to out-of-memory, and will not change anything if it does fail.
 *
 * If this ever adds more than 6 slots then code in berkeley_db.xml
 * (hack section for txn_recover) must change.
 */
int my_LOAD_unresolved_DistTxnId_and_valmap_record(descriptor_t *desc,
                                                   int txn_idx,
                                                   ErlDrvTermData *msg,
                                                   int *members,
                                                   int *msgcount)
{
    char *         memory_in_binary;
    ErlDrvBinary * binary;

    /* First the binary */
    if ((memory_in_binary = edtk_driver_alloc_wrapper(DB_XIDDATASIZE)) == NULL) {
        edtk_debug("%s : edtk_driver_alloc_wrapper failed", __FUNCTION__);
        *members = 0;
        return 0;
    }
    memcpy(memory_in_binary, desc->valmap_db_txn[txn_idx].distributed_txn_guid, DB_XIDDATASIZE);
    binary = edtk_alloced_ptr2ErlDrvBinary(memory_in_binary);

    /* Ensure that we decrement it's refcount at the end of _ready_async
     * (*AFTER* being send to Erlang -- i.e. Erlang will have incremented
     * it's refcount, so it won't really be freed) */
    assert(desc->num_tofree < MAX_RETURN_TERMS);
    desc->tofree[desc->num_tofree++] = binary;

    *msgcount = LOAD_BINARY(msg, *msgcount, binary, 0, DB_XIDDATASIZE);

    /* Now the valmap_db_txn.
     * This MUST match the record definition generated by hrl_template.gsl.
     * Similar code is in reply_ok_valmap(). */
    *msgcount = LOAD_ATOM(msg, *msgcount, am_valmap_db_txn);
    *msgcount = LOAD_PORT(msg, *msgcount, driver_mk_port(desc->port));
    *msgcount = LOAD_INT(msg, *msgcount, txn_idx);
    *msgcount = LOAD_INT(msg, *msgcount, desc->valmap_db_txn[txn_idx].generation_id);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 4);
    /* If this ever adds more than 6 slots then code in berkeley_db.xml
     * (hack section for txn_recover) must change. */

    *members = 2;
    return 2;  /* two new Erlang terms added */
}


/*
** This is called when there is a prepared transaction that must be resolved incrementally.
** It enqueues a command that will send an the event to the ports connected process (the port_coordinator)
** as an untagged_event message.
** The command includes the txn_idx, AND the generation_id from the txn_id slot.
** The generation_id allows the 'send event' code (in berkeley_db.xml, injected into ready_async)
** to determine if the transaction has been commited or aborted while the 'send event' command
** was waiting to be processed. (If that happens the event is not sent.)
**
** This ASSUMES that the caller owns desc->mutex.
**
** As the transaction is holding locks, this event is critical to availability.
** So we HARD-ABORT the OS process if we fail to allocage memory for the command.
*/
void my_send_need_to_resolve_prepared_txn_notification(descriptor_t * desc, int txn_idx)
{
    if ( 0 == enqueue_untagged_event_notification(desc,
                                                  BDB_NEED_TO_RESOLVE_PREPARED_TXN_NOTIFICATION,
                                                  (u_int32_t)txn_idx,
                                                  desc->valmap_db_txn[txn_idx].generation_id) )
    {
        fprintf(stderr, "\r\n%s: enqueue_untagged_event_notification failled. ABORTING.\r\n", __FUNCTION__);
        exit(EXIT_FAILURE);
    }
}


/*
** This is a callback, installed with DbEnv->set_event_notify().
** It sends the event to the ports connected process (the port_coordinator)
** as an 'untagged' message.
**
** As some of the events are critical to availability/data-integrity
** (DUPMASTER, NEWMASTER, MASTER, DOWNGRADED_TO_CLIENT etc)
** we **hard-abort** the OS process if we fail to allocate memory for the event.
** 
*/
void my_handle_sleepycat_event_notification(DB_ENV *dbenv, u_int32_t event, void *event_info)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    if ( 0 == enqueue_untagged_event_notification(desc, BDB_SLEEPYCAT_EVENT_NOTIFICATION, event, 0) )
    {
        fprintf(stderr, "\r\n%s: %s: enqueue_untagged_event_notification failled. ABORTING.\r\n", desc->port_label, __FUNCTION__);
        exit(EXIT_FAILURE);
    }
}


/* Helper for sending untagged events to ports connected process.  See code in berkeley_db.xml.
 *
 * Returns 0 if FAILED, 1 if successful.
 *
 * This does NOT attempt to acquire desc->mutex.
 * It DOES acquire and release desc->async_done_list_mutex.
 */
static int enqueue_untagged_event_notification(descriptor_t *desc,
                                               int           event_command_code,
                                               u_int32_t     payload_uint32,
                                               long          payload_long)
{
    async_queue_entry_t * entry;
    callstate_t *         c;
    int                   res;
    
    if ( (entry = malloc(sizeof(async_queue_entry_t))) == NULL ) {
        edtk_debug("%s: %s: failed to allocate memory for async_queue_entry", desc->port_label, __FUNCTION__);
        return 0;
    }

    if ( (c = bdb_alloc_callstate(event_command_code, desc)) == NULL) {
        edtk_debug("%s: %s: failed to allocate memory for callstate", desc->port_label, __FUNCTION__);
        return 0;
    }

    c->payload_uint32   = payload_uint32;
    c->payload_long     = payload_long;

    entry->next         = NULL;
    entry->prev         = NULL;
    entry->async_data   = c;
    entry->async_invoke = (void(*)(void *))-1;    /* won't be used (but not NULL as NULL is poison pill -- we want accidental use to cause a crash) */
    entry->async_free   = bdb_free_callstate;

    pthread_mutex_lock(&desc->async_done_list_mutex);
    edtk_debug("%s: %s: threadid=%lx Adding bdb_untagged_event operation %p to done_list", desc->port_label, __FUNCTION__, pthread_self(), (void*)(entry->async_data));
    entry->next = desc->async_done_list;
    desc->async_done_list = entry;
    pthread_mutex_unlock(&desc->async_done_list_mutex);

    /* Tell the Erlang VM to call bdb_ready_input (which will call bdb_ready_async to send the message).
     *
     * We DON'T test the return value of write() because I can't think of anything reasonable to do if it fails.
     * Aborting the process seems too aggressive, as the next normal command will cause the message to be sent
     * i.e. the message won't be lost, just delayed, unless no more messages can possibly be sent at all,
     * in which case the Erlang stubs for normal commands will start to get timeouts.
     * It coild be argued that a delay in receiving an event like DUPMASTER is dangerous, but if the entire driver's
     * reply channel is broken, then no other commands (e.g. txn_commit) will be returning anyway. */
     
    res = write(desc->async_done_signal_pipe_fds[1], "0", 1);
    edtk_debug("%s: %s: threadid=%lx write() to done_signal fd returned %d", desc->port_label, __FUNCTION__, pthread_self(), res);

    return 1;  /* success */
}

/* Returns 1 on FAILURE (any txn still in use), 0 on success (and *pout is set to NumUnresolvedTxns) */
int my_make_db_close_safe_if_there_are_unresolved_transactions(DB_ENV *dbenv, long *pout)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int i;
    assert( dbenv );
    *pout = 0;
    pthread_mutex_lock(&desc->mutex);
    for (i = 0; i < desc->num_valmap_db_txn_slots; i++) {
        if ( desc->valmap_db_txn[i].flags & VALMAP_EXCL_LOCK ) {
            edtk_debug("%s: %s: txn #%d is in-use", desc->port_label, __FUNCTION__, i);
            return 1; /* failure */
        }
        if ( desc->valmap_db_txn[i].flags & VALMAP_CANNOT_AUTO_CLEANUP ) {
            assert(desc->valmap_db_txn[i].value != NULL);
            assert(desc->valmap_db_txn[i].parent_index == -1);
            edtk_debug("%s: %s: found prepared txn #%d", desc->port_label, __FUNCTION__, i);
            ++(*pout);
        }
    }
    
    if (*pout == 0) {
        edtk_debug("%s: %s: no unresolved txns found", desc->port_label, __FUNCTION__);
    }
    else {
        edtk_debug("%s: %s: there are %ld unresolved txns, so setting dbenv_shutdown (to prevent logging in DB::close)", desc->port_label, __FUNCTION__, *pout);
        dbenv->dbenv_shutdown = 1;
    }
    pthread_mutex_unlock(&desc->mutex);
    return 0; /* success */
}


/* The caller is responsible for calling erl_free_compound() on the returned value.
 * Also this function does not  free(c->o.rep_stat_p)  -- free_callstate will do that later. */
ETERM * make_rep_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    edtk_debug("%s: %s: calling erl_format", desc->port_label, __FUNCTION__);

    return erl_format(
        "[{st_status,~i},"                    /* u_int32_t  Current replication status. */
        "{st_next_lsn,{~i,~i}},"              /* DB_LSN  Next LSN to use or expect. */
        "{st_waiting_lsn,{~i,~i}},"           /* DB_LSN  LSN we're awaiting, if any. */
        "{st_next_pg,~i},"                    /* db_pgno_t  Next pg we expect. */
        "{st_waiting_pg,~i},"                 /* db_pgno_t  pg we're awaiting, if any. */
        "{st_dupmasters,~i},"                 /* u_int32_t  # of times a duplicate master condition was detected.+ */
        "{st_env_id,~i},"                     /* int  Current environment ID. */
        "{st_env_priority,~i},"               /* int  Current environment priority. */
        "{st_bulk_fills,~i},"                 /* u_int32_t  Bulk buffer fills. */
        "{st_bulk_overflows,~i},"             /* u_int32_t  Bulk buffer overflows. */
        "{st_bulk_records,~i},"               /* u_int32_t  Bulk records stored. */
        "{st_bulk_transfers,~i},"             /* u_int32_t  Transfers of bulk buffers. */
        "{st_client_rerequests,~i},"          /* u_int32_t  Number of forced rerequests. */
        "{st_client_svc_req,~i},"             /* u_int32_t  Number of client service requests received by this client. */
        "{st_client_svc_miss,~i},"            /* u_int32_t  Number of client service requests missing on this client. */
        "{st_gen,~i},"                        /* u_int32_t  Current generation number. */
        "{st_egen,~i},"                       /* u_int32_t  Current election gen number. */
        "{st_log_duplicated,~i},"             /* u_int32_t  Log records received multiply.+ */
        "{st_log_queued,~i},"                 /* u_int32_t  Log records currently queued.+ */
        "{st_log_queued_max,~i},"             /* u_int32_t  Max. log records queued at once.+ */
        "{st_log_queued_total,~i},"           /* u_int32_t  Total # of log recs. ever queued.+ */
        "{st_log_records,~i},"                /* u_int32_t  Log records received and put.+ */
        "{st_log_requested,~i},"              /* u_int32_t  Log recs. missed and requested.+ */
        "{st_master,~i},"                     /* int  Env. ID of the current master. */
        "{st_master_changes,~i},"             /* u_int32_t  # of times we've switched masters. */
        "{st_msgs_badgen,~i},"                /* u_int32_t  Messages with a bad generation #.+ */
        "{st_msgs_processed,~i},"             /* u_int32_t  Messages received and processed.+ */
        "{st_msgs_recover,~i},"               /* u_int32_t  Messages ignored because this site was a client in recovery.+ */
        "{st_msgs_send_failures,~i},"         /* u_int32_t  # of failed message sends.+ */
        "{st_msgs_sent,~i},"                  /* u_int32_t  # of successful message sends.+ */
        "{st_newsites,~i},"                   /* u_int32_t  # of NEWSITE msgs. received.+ */
        "{st_nsites,~i},"                     /* int  Current number of sites we will assume during elections. */
        "{st_nthrottles,~i},"                 /* u_int32_t  # of times we were throttled. */
        "{st_outdated,~i},"                   /* u_int32_t  # of times we detected and returned an OUTDATED condition.+ */
        "{st_pg_duplicated,~i},"              /* u_int32_t  Pages received multiply.+ */
        "{st_pg_records,~i},"                 /* u_int32_t  Pages received and stored.+ */
        "{st_pg_requested,~i},"               /* u_int32_t  Pages missed and requested.+ */
        "{st_startup_complete,~i},"           /* u_int32_t  Site completed client sync-up. */
        "{st_txns_applied,~i},"               /* u_int32_t  # of transactions applied.+ */
        "{st_elections,~i},"                  /* u_int32_t  # of elections held.+ */
        "{st_elections_won,~i},"              /* u_int32_t  # of elections won by this site.+ */
        "{st_election_cur_winner,~i},"        /* int  Current front-runner. */
        "{st_election_gen,~i},"               /* u_int32_t  Election generation number. */
        "{st_election_lsn,{~i,~i}},"          /* DB_LSN  Max. LSN of current winner. */
        "{st_election_nsites,~i},"            /* int  # of "registered voters". */
        "{st_election_nvotes,~i},"            /* int  # of "registered voters" needed. */
        "{st_election_priority,~i},"          /* int  Current election priority. */
        "{st_election_status,~i},"            /* int  Current election status. */
        "{st_election_tiebreaker,~i},"        /* u_int32_t  Election tiebreaker value. */
        "{st_election_votes,~i},"             /* int  Votes received in this round. */
        "{st_election_sec,~i},"               /* u_int32_t  Last election time seconds. */
        "{st_election_usec,~i}]",             /* u_int32_t  Last election time useconds. */
        
        (int)(c->o.rep_stat_p->st_status),
        (int)(c->o.rep_stat_p->st_next_lsn.file), (int)(c->o.rep_stat_p->st_next_lsn.offset),
        (int)(c->o.rep_stat_p->st_waiting_lsn.file), (int)(c->o.rep_stat_p->st_waiting_lsn.offset),
        (int)(c->o.rep_stat_p->st_next_pg),
        (int)(c->o.rep_stat_p->st_waiting_pg),
        (int)(c->o.rep_stat_p->st_dupmasters),
        (int)(c->o.rep_stat_p->st_env_id),
        (int)(c->o.rep_stat_p->st_env_priority),
        (int)(c->o.rep_stat_p->st_bulk_fills),
        (int)(c->o.rep_stat_p->st_bulk_overflows),
        (int)(c->o.rep_stat_p->st_bulk_records),
        (int)(c->o.rep_stat_p->st_bulk_transfers),
        (int)(c->o.rep_stat_p->st_client_rerequests),
        (int)(c->o.rep_stat_p->st_client_svc_req),
        (int)(c->o.rep_stat_p->st_client_svc_miss),
        (int)(c->o.rep_stat_p->st_gen),
        (int)(c->o.rep_stat_p->st_egen),
        (int)(c->o.rep_stat_p->st_log_duplicated),
        (int)(c->o.rep_stat_p->st_log_queued),
        (int)(c->o.rep_stat_p->st_log_queued_max),
        (int)(c->o.rep_stat_p->st_log_queued_total),
        (int)(c->o.rep_stat_p->st_log_records),
        (int)(c->o.rep_stat_p->st_log_requested),
        (int)(c->o.rep_stat_p->st_master),
        (int)(c->o.rep_stat_p->st_master_changes),
        (int)(c->o.rep_stat_p->st_msgs_badgen),
        (int)(c->o.rep_stat_p->st_msgs_processed),
        (int)(c->o.rep_stat_p->st_msgs_recover),
        (int)(c->o.rep_stat_p->st_msgs_send_failures),
        (int)(c->o.rep_stat_p->st_msgs_sent),
        (int)(c->o.rep_stat_p->st_newsites),
        (int)(c->o.rep_stat_p->st_nsites),
        (int)(c->o.rep_stat_p->st_nthrottles),
        (int)(c->o.rep_stat_p->st_outdated),
        (int)(c->o.rep_stat_p->st_pg_duplicated),
        (int)(c->o.rep_stat_p->st_pg_records),
        (int)(c->o.rep_stat_p->st_pg_requested),
        (int)(c->o.rep_stat_p->st_startup_complete),
        (int)(c->o.rep_stat_p->st_txns_applied),
        (int)(c->o.rep_stat_p->st_elections),
        (int)(c->o.rep_stat_p->st_elections_won),
        (int)(c->o.rep_stat_p->st_election_cur_winner),
        (int)(c->o.rep_stat_p->st_election_gen),
        (int)(c->o.rep_stat_p->st_election_lsn.file), (int)(c->o.rep_stat_p->st_election_lsn.offset),
        (int)(c->o.rep_stat_p->st_election_nsites),
        (int)(c->o.rep_stat_p->st_election_nvotes),
        (int)(c->o.rep_stat_p->st_election_priority),
        (int)(c->o.rep_stat_p->st_election_status),
        (int)(c->o.rep_stat_p->st_election_tiebreaker),
        (int)(c->o.rep_stat_p->st_election_votes),
        (int)(c->o.rep_stat_p->st_election_sec),
        (int)(c->o.rep_stat_p->st_election_usec)
        );
}


/* The caller is responsible for calling erl_free_compound() on the returned value.
 * Also this function does not  free(c->o.lock_stat_p)  -- free_callstate will do that later. */
ETERM * make_lock_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    edtk_debug("%s: %s: calling erl_format", desc->port_label, __FUNCTION__);

    return erl_format(
        "[{st_id,~i},"            /* u_int32_t  Last allocated locker ID. */
        "{st_cur_maxid,~i},"      /* u_int32_t  Current maximum unused ID. */
        "{st_maxlocks,~i},"       /* u_int32_t  Maximum number of locks in table. */
        "{st_maxlockers,~i},"     /* u_int32_t  Maximum num of lockers in table. */
        "{st_maxobjects,~i},"     /* u_int32_t  Maximum num of objects in table. */
        "{st_nmodes,~i},"         /* int  Number of lock modes. */
        "{st_nlocks,~i},"         /* u_int32_t  Current number of locks. */
        "{st_maxnlocks,~i},"      /* u_int32_t  Maximum number of locks so far. */
        "{st_nlockers,~i},"       /* u_int32_t  Current number of lockers. */
        "{st_maxnlockers,~i},"    /* u_int32_t  Maximum number of lockers so far. */
        "{st_nobjects,~i},"       /* u_int32_t  Current number of objects. */
        "{st_maxnobjects,~i},"    /* u_int32_t  Maximum number of objects so far. */
        "{st_nrequests,~i},"      /* u_int32_t  Number of lock gets. */
        "{st_nreleases,~i},"      /* u_int32_t  Number of lock puts. */
        "{st_nupgrade,~i},"       /* u_int32_t  Number of lock upgrades. */
        "{st_ndowngrade,~i},"     /* u_int32_t  Number of lock downgrades. */
        "{st_lock_wait,~i},"      /* u_int32_t  Lock conflicts w/ subsequent wait */
        "{st_lock_nowait,~i},"    /* u_int32_t  Lock conflicts w/o subsequent wait */
        "{st_ndeadlocks,~i},"     /* u_int32_t  Number of lock deadlocks. */
        "{st_locktimeout,~i},"    /* db_timeout_t  Lock timeout. */
        "{st_nlocktimeouts,~i},"  /* u_int32_t  Number of lock timeouts. */
        "{st_txntimeout,~i},"     /* db_timeout_t  Transaction timeout. */
        "{st_ntxntimeouts,~i},"   /* u_int32_t  Number of transaction timeouts. */
        "{st_region_wait,~i},"    /* u_int32_t  Region lock granted after wait. */
        "{st_region_nowait,~i},"  /* u_int32_t  Region lock granted without wait. */
        "{st_regsize,{~i, ~i}}]",  /* roff_t  Region size. {GB, B} as it won't fit in an int on a 64-bit machine */

        (int)(c->o.lock_stat_p->st_id),
        (int)(c->o.lock_stat_p->st_cur_maxid),
        (int)(c->o.lock_stat_p->st_maxlocks),
        (int)(c->o.lock_stat_p->st_maxlockers),
        (int)(c->o.lock_stat_p->st_maxobjects),
        (int)(c->o.lock_stat_p->st_nmodes),
        (int)(c->o.lock_stat_p->st_nlocks),
        (int)(c->o.lock_stat_p->st_maxnlocks),
        (int)(c->o.lock_stat_p->st_nlockers),
        (int)(c->o.lock_stat_p->st_maxnlockers),
        (int)(c->o.lock_stat_p->st_nobjects),
        (int)(c->o.lock_stat_p->st_maxnobjects),
        (int)(c->o.lock_stat_p->st_nrequests),
        (int)(c->o.lock_stat_p->st_nreleases),
        (int)(c->o.lock_stat_p->st_nupgrade),
        (int)(c->o.lock_stat_p->st_ndowngrade),
        (int)(c->o.lock_stat_p->st_lock_wait),
        (int)(c->o.lock_stat_p->st_lock_nowait),
        (int)(c->o.lock_stat_p->st_ndeadlocks),
        (int)(c->o.lock_stat_p->st_locktimeout),
        (int)(c->o.lock_stat_p->st_nlocktimeouts),
        (int)(c->o.lock_stat_p->st_txntimeout),
        (int)(c->o.lock_stat_p->st_ntxntimeouts),
        (int)(c->o.lock_stat_p->st_region_wait),
        (int)(c->o.lock_stat_p->st_region_nowait),
        (int)(c->o.lock_stat_p->st_regsize >> 30), (int)(c->o.lock_stat_p->st_regsize & ((1<<30)-1))); /* 2^30 = 1GB */
}

ETERM * make_mutex_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    edtk_debug("%s: %s: calling erl_format", desc->port_label, __FUNCTION__);

    return erl_format(
        "[{st_mutex_align,~i},"            /* u_int32_t  Mutex alignment */
        "{st_mutex_tas_spins,~i},"        /* u_int32_t  Mutex test-and-set spins */
        "{st_mutex_cnt,~i},"              /* u_int32_t  Mutex count */
        "{st_mutex_free,~i},"             /* u_int32_t  Available mutexes */
        "{st_mutex_inuse,~i},"            /* u_int32_t  Mutexes in use */
        "{st_mutex_inuse_max,~i},"        /* u_int32_t  Maximum mutexes ever in use */
        "{st_region_wait,~i},"            /* u_int32_t  Region lock granted after wait. */
        "{st_region_nowait,~i},"          /* u_int32_t  Region lock granted without wait. */
        "{st_regsize,{~i,~i}}]",          /*roff_t  Region size. */

        (int)(c->o.mutex_stat_p->st_mutex_align),
        (int)(c->o.mutex_stat_p->st_mutex_tas_spins),
        (int)(c->o.mutex_stat_p->st_mutex_cnt),
        (int)(c->o.mutex_stat_p->st_mutex_free),
        (int)(c->o.mutex_stat_p->st_mutex_inuse),
        (int)(c->o.mutex_stat_p->st_mutex_inuse_max),
        (int)(c->o.mutex_stat_p->st_region_wait),
        (int)(c->o.mutex_stat_p->st_region_nowait),
        (int)(c->o.mutex_stat_p->st_regsize >> 30), (int)(c->o.mutex_stat_p->st_regsize & ((1<<30)-1))); /* 2^30 = 1GB */
}

/* private helper for make_txn_stat_external_term() */
ETERM * make_active_txn_term(DB_TXN_ACTIVE * active_txn_p)
{
    ETERM * active_txn_term;
    ETERM * xid_term = (active_txn_p->status == TXN_PREPARED)
                       ? erl_mk_binary((char *)(active_txn_p->xid), DB_XIDDATASIZE)
                       : erl_mk_atom("not_prepared");
    if ( xid_term == NULL ) {
        return NULL;
    }

    if ( (active_txn_term = erl_format(
              "{~i,"                            /* u_int32_t  Transaction ID (as returned by DB_TXN->id()) */
              "[{parent_txn_id,~i},"            /* u_int32_t  Transaction ID of parent */
              "{proces_id,~i},"                 /* pid_t  Process owning txn ID */
              "{thread_id,~i},"                 /* db_threadid_t Thread owning txn ID */
              "{lsn,{~i,~i}},"                  /* DB_LSN   LSN when transaction began */
              "{read_lsn,{~i,~i}},"             /* DB_LSN   Read LSN for MVCC */
              "{mvcc_ref,~i},"                  /* u_int32_t  MVCC reference count */
              "{status,~a},"                    /* u_int32_t  status */
              "{xid,~w},"                       /* u_int8_t[DB_XIDDATASIZE]  global transaction ID (if prepared) */
              "{name,~s}]}",                    /* char[51]  50 bytes of name, nul termination */

              (int)(active_txn_p->txnid),
              (int)(active_txn_p->parentid),
              (int)(active_txn_p->pid),
              (int)(active_txn_p->tid),
              (int)(active_txn_p->lsn.file),      (int)(active_txn_p->lsn.offset),
              (int)(active_txn_p->read_lsn.file), (int)(active_txn_p->read_lsn.offset),
              (int)(active_txn_p->mvcc_ref),
              bdb_TXN_status_code_atoms[active_txn_p->status - 1].atom_text,
              xid_term,
              (int)(active_txn_p->name)))
         == NULL )
    {
        erl_free_compound(xid_term);
        return NULL;
    }
    return active_txn_term;
}

ETERM * make_txn_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    int               i;
    ETERM *           active_txn_term_list;
    ETERM *           txn_stat_term;
    
    edtk_debug("%s: %s: calling erl_format", desc->port_label, __FUNCTION__);
    
    if ( (active_txn_term_list = erl_mk_empty_list()) == NULL ) {  /* we will cons onto the front of this */
        return NULL;
    }

    for (i = 0; i < c->o.txn_stat_p->st_nactive; ++i)
    {
        ETERM * active_txn_term;
        ETERM * new_cons;

        if ( (active_txn_term = make_active_txn_term(& c->o.txn_stat_p->st_txnarray[i])) == NULL ) {
            erl_free_compound(active_txn_term_list);
            return NULL;
        }
        if ( (new_cons = erl_cons(active_txn_term, active_txn_term_list)) == NULL ) {
            erl_free_compound(active_txn_term);
            erl_free_compound(active_txn_term_list);
            return NULL;
        }
        active_txn_term_list = new_cons;
    }

    if ( (txn_stat_term = erl_format(
              "[{active_txns,~w},"            /* list of active transactions */
              "{st_last_ckp,{~i,~i}},"          /* DB_LSN   lsn of the last checkpoint */
              "{st_time_ckp,{~i,~i,~i}},"       /* time_t  time of last checkpoint   (a tuple compatible with erlang:now()) */
              "{st_last_txnid,~i},"             /* u_int32_t  last transaction id given out */
              "{st_maxtxns,~i},"                /* u_int32_t  maximum txns possible */
              "{st_naborts,~i},"                /* u_int32_t  number of aborted transactions */
              "{st_nbegins,~i},"                /* u_int32_t  number of begun transactions */
              "{st_ncommits,~i},"               /* u_int32_t  number of committed transactions */
              "{st_nactive,~i},"                /* u_int32_t  number of active transactions */
              "{st_nsnapshot,~i},"              /* u_int32_t  number of snapshot transactions */
              "{st_nrestores,~i},"              /* u_int32_t  number of restored transactions after recovery. */
              "{st_maxnactive,~i},"             /* u_int32_t  maximum active transactions */
              "{st_maxnsnapshot,~i},"           /* u_int32_t  maximum snapshot transactions */
              "{st_region_wait,~i},"            /* u_int32_t  Region lock granted after wait. */
              "{st_region_nowait,~i}"           /* u_int32_t  Region lock granted without wait. */
              "{st_regsize,{~i,~i}}]",          /*roff_t  Region size. */

              active_txn_term_list,
              (int)(c->o.txn_stat_p->st_last_ckp.file), (int)(c->o.txn_stat_p->st_last_ckp.offset),
              (int)(c->o.txn_stat_p->st_time_ckp / 1000000), (int)(c->o.txn_stat_p->st_time_ckp % 1000000), 0, /* time_t is seconds since 00:00:00 on January 1, 1970 UTC.  erlang:now is same but microseconds. */
              (int)(c->o.txn_stat_p->st_last_txnid),
              (int)(c->o.txn_stat_p->st_maxtxns),
              (int)(c->o.txn_stat_p->st_naborts),
              (int)(c->o.txn_stat_p->st_nbegins),
              (int)(c->o.txn_stat_p->st_ncommits),
              (int)(c->o.txn_stat_p->st_nactive),
              (int)(c->o.txn_stat_p->st_nsnapshot),
              (int)(c->o.txn_stat_p->st_nrestores),
              (int)(c->o.txn_stat_p->st_maxnactive),
              (int)(c->o.txn_stat_p->st_maxnsnapshot),
              (int)(c->o.txn_stat_p->st_region_wait),
              (int)(c->o.txn_stat_p->st_region_nowait),
              (int)(c->o.txn_stat_p->st_regsize >> 30), (int)(c->o.txn_stat_p->st_regsize & ((1<<30)-1)))) /* 2^30 = 1GB */
         == NULL )
    {
        erl_free_compound(active_txn_term_list);
        return NULL;
    }
    return txn_stat_term;
}

/* private helper for make_memp_stat_external_term */
ETERM * make_memp_fstat_term(DB_MPOOL_FSTAT * mpool_fstat_p)
{
    return erl_format(
        "{~s,"                            /* char[51] File name. */
        "[{st_pagesize,~i},"              /* u_int32_t  Page size. */
        "{st_map,~i},"                    /* u_int32_t  Pages from mapped files. */
        "{st_cache_hit,~i},"              /* u_int32_t  Pages found in the cache. */
        "{st_cache_miss,~i},"             /* u_int32_t  Pages not found in the cache. */
        "{st_page_create,~i},"            /* u_int32_t  Pages created in the cache. */
        "{st_page_in,~i},"                /* u_int32_t  Pages read in. */
        "{st_page_out,~i}]}",             /* u_int32_t  Pages written out. */

        mpool_fstat_p->file_name,
        (int)(mpool_fstat_p->st_pagesize),
        (int)(mpool_fstat_p->st_map),
        (int)(mpool_fstat_p->st_cache_hit),
        (int)(mpool_fstat_p->st_cache_miss),
        (int)(mpool_fstat_p->st_page_create),
        (int)(mpool_fstat_p->st_page_in),
        (int)(mpool_fstat_p->st_page_out));
}

ETERM * make_memp_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    ETERM *           fstat_term_list;
    DB_MPOOL_FSTAT ** fstat_pp;
    ETERM *           memp_stat_term;
    
    edtk_debug("%s: %s: calling erl_format", desc->port_label, __FUNCTION__);

    if ( (fstat_term_list = erl_mk_empty_list()) == NULL ) {  /* we will cons onto the front of this */
        return NULL;
    }

    /* fprintf(stderr, "***** fstat array is at %p\n", (void*)(c->o.mpool_fstat_ary_p)); */
            
    if ( (fstat_pp = c->o.mpool_fstat_ary_p) != NULL )   /* might be no array at all if there are no files */
    {
        while ( *fstat_pp != NULL )
        {
            ETERM * fstat_term;
            ETERM * new_cons;

            /* fprintf(stderr, "***** at %p (idx %d), adding %s\n", (void*)(fstat_pp), fstat_pp - c->o.mpool_fstat_ary_p, (*fstat_pp)->file_name); */
            
            if ( (fstat_term = make_memp_fstat_term(*fstat_pp)) == NULL ) {
                erl_free_compound(fstat_term_list);
                return NULL;
            }
            if ( (new_cons = erl_cons(fstat_term, fstat_term_list)) == NULL ) {
                erl_free_compound(fstat_term);
                erl_free_compound(fstat_term_list);
                return NULL;
            }
            fstat_term_list = new_cons;
            ++fstat_pp;
        }
    }
    /* fprintf(stderr, "***** end of fstat array is at %p (array idx %d)\n", (void*)(fstat_pp), fstat_pp - c->o.mpool_fstat_ary_p); */

    if ( (memp_stat_term = erl_format(
              "[{file_stats,~w},"          /* fstat array */
              "{st_cache_size,{~i,~i}},"   /* 2*u_int32_t  Total cache size: {GB, B} */
              "{st_ncache,~i},"            /* u_int32_t  Number of caches. */
              "{st_regsize,{~i,~i}},"      /* roff_t Region size.           {GB, B} */
              "{st_mmapsize,{~i, ~i}},"    /* size_t Maximum file size for mmap. {GB, B} */
              "{st_maxopenfd,~i},"         /* int Maximum number of open fd's. */
              "{st_maxwrite,~i},"          /* int Maximum buffers to write. */
              "{st_maxwrite_sleep,~i},"    /* int Sleep after writing max buffers. */
              "{st_map,~i},"               /* u_int32_t  Pages from mapped files. */
              "{st_cache_hit,~i},"         /* u_int32_t  Pages found in the cache. */
              "{st_cache_miss,~i},"        /* u_int32_t  Pages not found in the cache. */
              "{st_page_create,~i},"       /* u_int32_t  Pages created in the cache. */
              "{st_page_in,~i},"           /* u_int32_t  Pages read in. */
              "{st_page_out,~i},"          /* u_int32_t  Pages written out. */
              "{st_ro_evict,~i},"          /* u_int32_t  Clean pages forced from the cache. */
              "{st_rw_evict,~i},"          /* u_int32_t  Dirty pages forced from the cache. */
              "{st_page_trickle,~i},"      /* u_int32_t  Pages written by memp_trickle. */
              "{st_pages,~i},"             /* u_int32_t  Total number of pages. */
              "{st_page_clean,~i},"        /* u_int32_t  Clean pages. */
              "{st_page_dirty,~i},"        /* u_int32_t  Dirty pages. */
              "{st_hash_buckets,~i},"      /* u_int32_t  Number of hash buckets. */
              "{st_hash_searches,~i},"     /* u_int32_t  Total hash chain searches. */
              "{st_hash_longest,~i},"      /* u_int32_t  Longest hash chain searched. */
              "{st_hash_examined,~i},"     /* u_int32_t  Total hash entries searched. */
              "{st_hash_nowait,~i},"       /* u_int32_t  Hash lock granted with nowait. */
              "{st_hash_wait,~i},"         /* u_int32_t  Hash lock granted after wait. */
              "{st_hash_max_wait,~i},"     /* u_int32_t  Max hash lock granted after wait. */
              "{st_region_nowait,~i},"     /* u_int32_t  Region lock granted with nowait. */
              "{st_region_wait,~i},"       /* u_int32_t  Region lock granted after wait. */
              "{st_mvcc_frozen,~i},"       /* u_int32_t  Buffers frozen. */
              "{st_mvcc_thawed,~i},"       /* u_int32_t  Buffers thawed. */
              "{st_mvcc_freed,~i},"        /* u_int32_t  Frozen buffers freed. */
              "{st_alloc,~i},"             /* u_int32_t  Number of page allocations. */
              "{st_alloc_buckets,~i},"     /* u_int32_t  Buckets checked during allocation. */
              "{st_alloc_max_buckets,~i}," /* u_int32_t  Max checked during allocation. */
              "{st_alloc_pages,~i},"       /* u_int32_t  Pages checked during allocation. */
              "{st_alloc_max_pages,~i},"   /* u_int32_t  Max checked during allocation. */
              "{st_io_wait,~i}]",          /* u_int32_t  Thread waited on buffer I/O. */

              fstat_term_list,
              (int)(c->o.mpool_stat_p->st_gbytes), (int)(c->o.mpool_stat_p->st_bytes),
              (int)(c->o.mpool_stat_p->st_ncache),
              (int)(c->o.mpool_stat_p->st_regsize >> 30),  (int)(c->o.mpool_stat_p->st_regsize  & ((1<<30)-1)), /* 2^30 = 1GB */
              (int)(c->o.mpool_stat_p->st_mmapsize >> 30), (int)(c->o.mpool_stat_p->st_mmapsize & ((1<<30)-1)),
              (int)(c->o.mpool_stat_p->st_maxopenfd),
              (int)(c->o.mpool_stat_p->st_maxwrite),
              (int)(c->o.mpool_stat_p->st_maxwrite_sleep),
              (int)(c->o.mpool_stat_p->st_map),
              (int)(c->o.mpool_stat_p->st_cache_hit),
              (int)(c->o.mpool_stat_p->st_cache_miss),
              (int)(c->o.mpool_stat_p->st_page_create),
              (int)(c->o.mpool_stat_p->st_page_in),
              (int)(c->o.mpool_stat_p->st_page_out),
              (int)(c->o.mpool_stat_p->st_ro_evict),
              (int)(c->o.mpool_stat_p->st_rw_evict),
              (int)(c->o.mpool_stat_p->st_page_trickle),
              (int)(c->o.mpool_stat_p->st_pages),
              (int)(c->o.mpool_stat_p->st_page_clean),
              (int)(c->o.mpool_stat_p->st_page_dirty),
              (int)(c->o.mpool_stat_p->st_hash_buckets),
              (int)(c->o.mpool_stat_p->st_hash_searches),
              (int)(c->o.mpool_stat_p->st_hash_longest),
              (int)(c->o.mpool_stat_p->st_hash_examined),
              (int)(c->o.mpool_stat_p->st_hash_nowait),
              (int)(c->o.mpool_stat_p->st_hash_wait),
              (int)(c->o.mpool_stat_p->st_hash_max_wait),
              (int)(c->o.mpool_stat_p->st_region_nowait),
              (int)(c->o.mpool_stat_p->st_region_wait),
              (int)(c->o.mpool_stat_p->st_mvcc_frozen),
              (int)(c->o.mpool_stat_p->st_mvcc_thawed),
              (int)(c->o.mpool_stat_p->st_mvcc_freed),
              (int)(c->o.mpool_stat_p->st_alloc),
              (int)(c->o.mpool_stat_p->st_alloc_buckets),
              (int)(c->o.mpool_stat_p->st_alloc_max_buckets),
              (int)(c->o.mpool_stat_p->st_alloc_pages),
              (int)(c->o.mpool_stat_p->st_alloc_max_pages),
              (int)(c->o.mpool_stat_p->st_io_wait)) )
         == NULL )
    {
        erl_free_compound(fstat_term_list);
        return NULL;
    }
    return memp_stat_term;
}

ETERM * make_log_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    edtk_debug("%s: %s: calling erl_format", desc->port_label, __FUNCTION__);

    return erl_format(
        "[{st_magic,~i},"             /* u_int32_t  Log file magic number. */
        "{st_version,~i},"            /* u_int32_t  Log file version number. */
        "{st_mode,~i},"               /* int Log file permissions mode. */
        "{st_lg_bsize,~i},"           /* u_int32_t  Log buffer size. */
        "{st_lg_size,~i},"            /* u_int32_t  Log file size. */
        "{st_record,~i},"             /* u_int32_t  Records entered into the log. */
        "{st_w_bytes,~i},"            /* u_int32_t  Bytes to log. */
        "{st_w_mbytes,~i},"           /* u_int32_t  Megabytes to log. */
        "{st_wc_bytes,~i},"           /* u_int32_t  Bytes to log since checkpoint. */
        "{st_wc_mbytes,~i},"          /* u_int32_t  Megabytes to log since checkpoint. */
        "{st_wcount,~i},"             /* u_int32_t  Total I/O writes to the log. */
        "{st_wcount_fill,~i},"        /* u_int32_t  Overflow writes to the log. */
        "{st_rcount,~i},"             /* u_int32_t  Total I/O reads from the log. */
        "{st_scount,~i},"             /* u_int32_t  Total syncs to the log. */
        "{st_region_wait,~i},"        /* u_int32_t  Region lock granted after wait. */
        "{st_region_nowait,~i},"      /* u_int32_t  Region lock granted without wait. */
        "{st_cur_file,~i},"           /* u_int32_t  Current log file number. */
        "{st_cur_offset,~i},"         /* u_int32_t  Current log file offset. */
        "{st_disk_file,~i},"          /* u_int32_t  Known on disk log file number. */
        "{st_disk_offset,~i},"        /* u_int32_t  Known on disk log file offset. */
        "{st_regsize,{~i,~i}},"       /* roff_t     Region size {GB, B} */
        "{st_maxcommitperflush,~i},"  /* u_int32_t  Max number of commits in a flush. */
        "{st_mincommitperflush,~i}]", /* u_int32_t  Min number of commits in a flush. */

        (int)(c->o.log_stat_p->st_magic),
        (int)(c->o.log_stat_p->st_version),
        (int)(c->o.log_stat_p->st_mode),
        (int)(c->o.log_stat_p->st_lg_bsize),
        (int)(c->o.log_stat_p->st_lg_size),
        (int)(c->o.log_stat_p->st_record),
        (int)(c->o.log_stat_p->st_w_bytes),
        (int)(c->o.log_stat_p->st_w_mbytes),
        (int)(c->o.log_stat_p->st_wc_bytes),
        (int)(c->o.log_stat_p->st_wc_mbytes),
        (int)(c->o.log_stat_p->st_wcount),
        (int)(c->o.log_stat_p->st_wcount_fill),
        (int)(c->o.log_stat_p->st_rcount),
        (int)(c->o.log_stat_p->st_scount),
        (int)(c->o.log_stat_p->st_region_wait),
        (int)(c->o.log_stat_p->st_region_nowait),
        (int)(c->o.log_stat_p->st_cur_file),
        (int)(c->o.log_stat_p->st_cur_offset),
        (int)(c->o.log_stat_p->st_disk_file),
        (int)(c->o.log_stat_p->st_disk_offset),
        (int)(c->o.log_stat_p->st_regsize >> 30), (int)(c->o.log_stat_p->st_regsize & ((1<<30)-1)), /* 2^30 = 1GB */
        (int)(c->o.log_stat_p->st_maxcommitperflush),
        (int)(c->o.log_stat_p->st_mincommitperflush));
}

/* We only need this helper because we need to call DB->get_type in order to decode the output */
int my_db_stat(DB *db, DB_TXN *txnid, DBTYPE * db_type_p, void ** db_stat_p, u_int32_t flags)
{
    int res;
    if ( (res = db->get_type(db, db_type_p)) != 0 )
        return res;
    
    return db->stat(db, txnid, db_stat_p, flags);
}

ETERM * make_db_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    int idx = c->o.db_type - 1;

    edtk_debug("%s: %s: calling erl_format", desc->port_label, __FUNCTION__);

    assert(idx >= 0 && idx < NUM_ARY_ELEMS(bdb_DBTYPE_code_atoms) );
    
    switch ( c->o.db_type )
    {
    case DB_BTREE:
        return erl_format(
            "[{db_type,~a},"                  /* we add this so caller doesn't have to call db_get_type */
            "{bt_magic,~i},"                  /* u_int32_t  Magic number. */
            "{bt_version,~i},"                /* u_int32_t  Version number. */
            "{bt_metaflags,~i},"              /* u_int32_t  Metadata flags. */
            "{bt_nkeys,~i},"                  /* u_int32_t  Number of unique keys. */
            "{bt_ndata,~i},"                  /* u_int32_t  Number of data items. */
            "{bt_pagesize,~i},"               /* u_int32_t  Page size. */
            "{bt_minkey,~i},"                 /* u_int32_t  Minkey value. */
            "{bt_re_len,~i},"                 /* u_int32_t  Fixed-length record length. */
            "{bt_re_pad,~i},"                 /* u_int32_t  Fixed-length record pad. */
            "{bt_levels,~i},"                 /* u_int32_t  Tree levels. */
            "{bt_int_pg,~i},"                 /* u_int32_t  Internal pages. */
            "{bt_leaf_pg,~i},"                /* u_int32_t  Leaf pages. */
            "{bt_dup_pg,~i},"                 /* u_int32_t  Duplicate pages. */
            "{bt_over_pg,~i},"                /* u_int32_t  Overflow pages. */
            "{bt_empty_pg,~i},"               /* u_int32_t  Empty pages. */
            "{bt_free,~i},"                   /* u_int32_t  Pages on the free list. */
            "{bt_int_pgfree,~i},"             /* u_int32_t  Bytes free in internal pages. */
            "{bt_leaf_pgfree,~i},"            /* u_int32_t  Bytes free in leaf pages. */
            "{bt_dup_pgfree,~i},"             /* u_int32_t  Bytes free in duplicate pages. */
            "{bt_over_pgfree,~i}]",           /* u_int32_t  Bytes free in overflow pages. */

            bdb_DBTYPE_code_atoms[idx].atom_text,    /* text to convert to atom -- i.e. NOT .atom !! */
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_magic),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_version),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_metaflags),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_nkeys),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_ndata),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_pagesize),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_minkey),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_re_len),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_re_pad),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_levels),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_int_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_leaf_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_dup_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_over_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_empty_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_free),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_int_pgfree),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_leaf_pgfree),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_dup_pgfree),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_over_pgfree));

    case DB_HASH:
        return erl_format(
            "[{db_type,~a},"                  /* we add this so caller doesn't have to call db_get_type */
            "{hash_magic,~i},"                /* u_int32_t  Magic number. */
            "{hash_version,~i},"              /* u_int32_t  Version number. */
            "{hash_metaflags,~i},"            /* u_int32_t  Metadata flags. */
            "{hash_nkeys,~i},"                /* u_int32_t  Number of unique keys. */
            "{hash_ndata,~i},"                /* u_int32_t  Number of data items. */
            "{hash_pagesize,~i},"             /* u_int32_t  Page size. */
            "{hash_ffactor,~i},"              /* u_int32_t  Fill factor specified at create. */
            "{hash_buckets,~i},"              /* u_int32_t  Number of hash buckets. */
            "{hash_free,~i},"                 /* u_int32_t  Pages on the free list. */
            "{hash_bfree,~i},"                /* u_int32_t  Bytes free on bucket pages. */
            "{hash_bigpages,~i},"             /* u_int32_t  Number of big key/data pages. */
            "{hash_big_bfree,~i},"            /* u_int32_t  Bytes free on big item pages. */
            "{hash_overflows,~i},"            /* u_int32_t  Number of overflow pages. */
            "{hash_ovfl_free,~i},"            /* u_int32_t  Bytes free on ovfl pages. */
            "{hash_dup,~i},"                  /* u_int32_t  Number of dup pages. */
            "{hash_dup_free,~i}]",            /* u_int32_t  Bytes free on duplicate pages. */

            bdb_DBTYPE_code_atoms[idx].atom_text,    /* text to convert to atom -- i.e. NOT .atom !! */
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_magic),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_version),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_metaflags),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_nkeys),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_ndata),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_pagesize),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_ffactor),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_buckets),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_free),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_bfree),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_bigpages),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_big_bfree),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_overflows),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_ovfl_free),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_dup),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_dup_free));
        
    case DB_QUEUE:
        return erl_format(
            "[{db_type,~a},"                  /* we add this so caller doesn't have to call db_get_type */
            "{qs_magic,~i},"                  /* u_int32_t  Magic number. */
            "{qs_version,~i},"                /* u_int32_t  Version number. */
            "{qs_metaflags,~i},"              /* u_int32_t  Metadata flags. */
            "{qs_nkeys,~i},"                  /* u_int32_t  Number of unique keys. */
            "{qs_ndata,~i},"                  /* u_int32_t  Number of data items. */
            "{qs_pagesize,~i},"               /* u_int32_t  Page size. */
            "{qs_extentsize,~i},"             /* u_int32_t  Pages per extent. */
            "{qs_pages,~i},"                  /* u_int32_t  Data pages. */
            "{qs_re_len,~i},"                 /* u_int32_t  Fixed-length record length. */
            "{qs_re_pad,~i},"                 /* u_int32_t  Fixed-length record pad. */
            "{qs_pgfree,~i},"                 /* u_int32_t  Bytes free in data pages. */
            "{qs_first_recno,~i},"            /* u_int32_t  First not deleted record. */
            "{qs_cur_recno,~i}]",             /* u_int32_t  Next available record number. */

            bdb_DBTYPE_code_atoms[idx].atom_text,    /* text to convert to atom -- i.e. NOT .atom !! */
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_magic),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_version),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_metaflags),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_nkeys),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_ndata),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_pagesize),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_extentsize),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_pages),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_re_len),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_re_pad),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_pgfree),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_first_recno),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_cur_recno));
        
    default:
        return NULL;
    }

}


int make_db_compact_xreturn(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount)
{
    char		*bin_mem;
    ErlDrvBinary	*bin = NULL;

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_pages_free);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_pages_free);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_pages_examine);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_pages_examine);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_levels);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_levels);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_deadlock);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_deadlock);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_pages_truncated);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_pages_truncated);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_end_key);
   
    /* TODO: refactor the following to call make_dbt */    
    if ((bin_mem = edtk_driver_alloc_wrapper(c->o.end_key.size)) == NULL) {
        fprintf(stderr, "\r\n%s: %s: edtk_driver_alloc_wrapper failed! Aborting.\r\n", desc->port_label, __FUNCTION__);
        exit(EXIT_FAILURE);
    }
    memcpy(bin_mem, c->o.end_key.data, c->o.end_key.size);
    bin = edtk_alloced_ptr2ErlDrvBinary(bin_mem);

    assert(desc->num_tofree < MAX_RETURN_TERMS);
    desc->tofree[desc->num_tofree++] = bin;

    /* As BDB malloced data for this object then we must free it, now that we have copied it to a driver binary.
     * However, we free it in free_callstate not here, because this function might not be called.
     * If driver shutdown has begun after invoke_XX has called BDB but before ready_async is called then ready_async
     * won't actually be called (and therefore cannot call us). Only free_callstate will be called. */
    *msgcount = LOAD_BINARY(msg, *msgcount, bin, 0, c->o.end_key.size);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);
   
    /* The list has to end with NIL */
    *msgcount = LOAD_NIL(msg, *msgcount);

    *members = 7; /* including NIL l */
    return 7;
}


/* This must advance *members by 1.
 *
 * This is used to convert
 *    0 to 'ok'
 *  +ve to posix (e.g. 'enomem', 'einval')
 *  -ve to BDB   (e.g. 'bdb_DB_NOTFOUND', 'bdb_DB_LOCK_DEADLOCK')
 *
 * Unrecognized codes are converted to
 *         {unknown_error_code, int()}
 */
void convert_bdb_return_code_to_driver_term(descriptor_t *desc,
                                            int error_code,
                                            ErlDrvTermData *msg,
                                            int *members,
                                            int *msgcount)
{
    if ( error_code == 0 )
    {
        *msgcount = LOAD_ATOM(msg, *msgcount, am_ok);
        *members += 1;
        return;
    }
    else if ( error_code > 0 )
    {
        /* POSIX (error_codeno) */
        int idx = error_code - 1;
        if (  idx < NUM_ARY_ELEMS(errno_atoms)
           && errno_atoms[idx].code != 0 )
        {
            *msgcount = LOAD_ATOM(msg, *msgcount, errno_atoms[idx].atom);
            *members += 1;
            return;
        }
        else {
            /* unknown POSIX (positive) error_codeor */
        }
    }
    else if ( error_code >= -30999  &&  error_code <= -30800 )
    {
        /* In documented range of a BDB public error_codeor code */
        int idx = 0 - (-30999 - error_code);
        if (  idx < NUM_ARY_ELEMS(bdb_public_error_code_atoms)
           && errno_atoms[idx].code != 0 )
        {
            *msgcount = LOAD_ATOM(msg, *msgcount, bdb_public_error_code_atoms[idx].atom);
            *members += 1;
            return;
        }
        else {
            /* We don't have a definition for this code */
        }
    }
    else {
        /* negative but not in BDB public range */
    }

    *msgcount = LOAD_ATOM(msg, *msgcount, am_unknown_error_code);
    *msgcount = LOAD_INT(msg, *msgcount, error_code);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);
    *members += 1;
}

int make_dbtype(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount)
{
    int idx = c->o.db_type - 1;
    assert(idx >= 0 && idx < NUM_ARY_ELEMS(bdb_DBTYPE_code_atoms) );
    *msgcount = LOAD_ATOM(msg, *msgcount, bdb_DBTYPE_code_atoms[idx].atom);
    *members += 1;
    return 1;
}


int
my_make_list_log_filenames(descriptor_t *desc,
                           callstate_t *c,
                           ErlDrvTermData *msg,
                           int *members,
                           int *msgcount)
{
    char ** filename_p;

    if (c->o.log_filenames_p != NULL)
    {
        for (filename_p = c->o.log_filenames_p; *filename_p != NULL; ++filename_p)
        {
            *msgcount = LOAD_STRING(msg, *msgcount, *filename_p, strlen(*filename_p));
            ++(*members);
        }
        /* c->o.log_filenames_p will be freed in free_callstate() */
    }
    
    /* The list has to end with NIL */
    *msgcount = LOAD_NIL(msg, *msgcount);
    ++(*members);
    
    /* In berkeley_db.xml, this xtra-return type is declared as a 'list'.
     * So reply_xtra will do the final "*msgcount = LOAD_LIST(msg, *msgcount, members);" */
    return *members;
}


